/*
 * cp40.h --
 *
 *	Declarations of externally visible resources of the CP40 library.
 *	The numerous comments throughout this file are intended to provide
 *	a basic documentation for the use of the CP40 library.
 *
 * Copyright (c) 1995-1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id: cp40.h,v 2.2 2005/06/29 20:25:21 eric Exp $
 *	$Log: cp40.h,v $
 *	Revision 2.2  2005/06/29 20:25:21  eric
 *	 - Saved prior to making CP40 as a standalone Yorick package.
 *
 *	Revision 2.1  2003/06/25 14:03:19  eric
 *	 - Last save before new major release.
 *
 *	Revision 2.0  2000/01/20 08:13:57  eric
 *	*** empty log message ***
 *
 *-----------------------------------------------------------------------------
 */

#ifndef CP40_H
#define CP40_H 1

/* Standard header files needed. */
#ifndef BUFSIZ
# include <stdio.h>
#endif
#include "gpl_base.h"
#include "gpl_string.h"
#include "gpl_hash.h"
#include "gpl_xdr.h"

/* CP40 version.  */
#define CP40_VERSION_MAJOR  5
#define CP40_VERSION_MINOR  0
#define CP40_VERSION_STRING "5.0"

GPL_BEGIN_DECLS
/*---------------------------------------------------------------------------*/
/* PREDEFINED DATA TYPES */

typedef gpl_xdr_uint_t              cp40_raw_photon_t;
typedef struct _cp40_raw_frame_t  cp40_raw_frame_t;
typedef struct _cp40_stream_t     cp40_stream_t;
typedef struct _cp40_stream_ops_t cp40_stream_ops_t;
typedef struct _cp40_io_file_t    cp40_io_file_t;
typedef struct _cp40_io_socket_t  cp40_io_socket_t;
typedef struct _cp40_instance_t   cp40_instance_t;
typedef struct _cp40_range_t      cp40_range_t;
typedef struct _cp40_window_t     cp40_window_t;
typedef struct _cp40_attr_entry_t cp40_attr_entry_t;
typedef struct _cp40_avg_stat_t   cp40_avg_stat_t;


/* A structure cp40_task_t is used to store all resources needed to perform
   a given photon-counting integration (e.g., long exposure, intercorrelation,
   autocorrelation, ..).  Each task is effectively performed onto the current
   pre-processed data by a routine with prototype cp40_task_op_t. */
typedef struct _cp40_task_t cp40_task_t;
typedef int cp40_task_op_t(cp40_task_t *task);

/* A structure cp40_stat_t is used to store raw-photons and statistics
   for input frames. */
typedef struct _cp40_stat_t cp40_stat_t;

/*---------------------------------------------------------------------------*/
/* CP40 HARDWARE SETTINGS (CCD sizes, etc). */

#define CP40_NCHANNELS       4	/* Number of CP40 channels. */
#define CP40_CCD_WIDTH     384
#define CP40_CCD_HEIGHT	   288
#define CP40_WIDTH(pixel)  (CP40_CCD_WIDTH  << (2 - (pixel)))
#define CP40_HEIGHT(pixel) (CP40_CCD_HEIGHT << (2 - (pixel)))
/*	Return the number of raw pixels for one CP40 channel along the
	vertical/horizontal direction; the argument is one of the
	CP40_PIXEL_... values (see below). */

/*---------------------------------------------------------------------------*/
/* Window settings (if one of the bounds is equal to CP40_WINDOW_WILD, then
   the corresponding extreme value should be used instead). */
struct _cp40_window_t {
# define CP40_WINDOW_WILD (0x80000000L)
  int xmin, xmax, ymin, ymax;
};

/* Interval of values (the flags indicate if a particular bound is set).  */
struct _cp40_range_t {
# define CP40_MIN_SET  ((unsigned int)0x1)
# define CP40_MAX_SET  ((unsigned int)0x2)
  unsigned int flags;
  double min, max;
};

/* Constants for header sizes.  */
#define CP40_HEADER_SIZE     0x4000	/* 16 Kbytes header. */
#define CP40_TEXT_OFFSET     0x0400	/*  1 Kbytes binary header. */

/*---------------------------------------------------------------------------*/
			 /**************************
			  * PRE-PROCESSING OF DATA *
			  **************************/

/*---------------------------------------------------------------------------*/
/* SELECTION OF FRAME PARITY */

/* Frame selection based on the parity of the frame number (date). */
#define CP40_PARITY_EVEN  1
#define CP40_PARITY_ODD	  2
#define CP40_PARITY_ANY   (CP40_PARITY_EVEN|CP40_PARITY_ODD)


/* Routines to query/set parity of selected frames. */
extern int cp40_get_parity(cp40_instance_t *instance);
extern int cp40_set_parity(cp40_instance_t *instance, int parity);


/*---------------------------------------------------------------------------*/
/* RAW PIXEL RESOLUTION */

typedef enum {
  /* Raw-pixel  sampling (IMPORTANT:  The following values  are assumed  to be
     equal to the number of right  shift bits needed to convert data raw pixel
     coordinates,  i.e. logical,  to  a particular  sampling.   DO NOT  CHANGE
     THEM). */
  CP40_PIXEL_LOGICAL=       0,	/* Raw pixel resolution given by the CP40
				   camera after centroiding (MUST BE 0). */
  CP40_PIXEL_HALF_PHYSICAL= 1,	/* 2 logical pixels (MUST BE 1). */
  CP40_PIXEL_PHYSICAL=      2	/* 4 logical pixels, also the CCD resolution
				   of the CP40 (*MUST* be 2). */
} cp40_pixel_t;

/* Routines to query/set raw-pixel sampling. */
extern cp40_pixel_t cp40_get_pixel(cp40_instance_t *instance);
extern int cp40_set_pixel(cp40_instance_t *instance, cp40_pixel_t value);

/*---------------------------------------------------------------------------*/
/* SPURIOUS PHOTONS */

/* In principle,  the centroiding  device  forbid  the  detection of  multiple
   events in  the same  frame within  the same small  region (3x3  CCD pixel).
   This produce a bias known  as the "trou du centreur" (photon-counting hole)
   in autocorrelations of  frames.  The first step of  the data pre-processing
   consist in dealing with such "spurious" events. */

typedef enum {
  /* Possible methods to deal with spurious photo-events: */
  CP40_SPURIOUS_LEAVE=   1, /* Do nothing special about spurious (i.e.
			       assume there is no spurious events). */
  CP40_SPURIOUS_DELETE=  2, /* Delete all spurious photo-events. */
  CP40_SPURIOUS_CENTRE=  3, /* Replace group of spurious events by a single
			       one with average position. */
  CP40_SPURIOUS_SINGLE=  4  /* Replace group of spurious events by first
			       of them. */
} cp40_spurious_t;

/* Routines to query/set method which deals with spurious photo-events. */
extern cp40_spurious_t cp40_get_spurious(cp40_instance_t *instance);
extern int cp40_set_spurious(cp40_instance_t *instance, cp40_spurious_t value);

/*---------------------------------------------------------------------------*/
/* DISTORTION CORRECTION */

/* The image intensifier (mainly the 1st generation tube) of the CP40 is
   responsible for a geometrical distortion that can be fixed thanks to a 2-D
   polynomial correction.  A 2rd order correction is sufficient:

     X' = CX[0] + CX[1]*X + CX[2]*Y + CX[3]*X*X + CX[4]*X*Y + CX[5]*Y*Y
        + CX[6]*X*X*X +  CX[7]*X*X*Y + CX[8]*X*Y*Y + CX[9]*Y*Y*Y
     Y' = CY[0] + CY[1]*X + CY[2]*Y + CY[3]*X*X + CY[4]*X*Y + CY[5]*Y*Y
        + CY[6]*X*X*X +  CY[7]*X*X*Y + CY[8]*X*Y*Y + CY[9]*Y*Y*Y

   For  efficiency  factorized  expressions   should  be  used  instead.   The
   following macros give  the coordinates (X',Y') of the  pixel corrected from
   the distortion,  C is  the array of  coefficients of  distortion correction
   (double C[2*CP40_NCOEFS], i.e., CY=CX+CP40_NCOEFS), `X' and `Y' are the raw
   *PHYSICAL*  pixel  coordinates (eventually  floating  values for  sub-pixel
   position).  */
#define CP40_XCOR(c,x,y) \
    ((c)[0]+((c)[1]+(c)[4]*(y)+((c)[3]+(c)[6]*(x)+(c)[7]*(y))*(x))*(x)\
     +((c)[2]+((c)[5]+(c)[8]*(x)+(c)[9]*(y))*(y))*(y))
#define CP40_YCOR(c,x,y) \
    ((c)[10]+((c)[11]+(c)[14]*(y)+((c)[13]+(c)[16]*(x)+(c)[17]*(y))*(x))*(x)\
     +((c)[12]+((c)[15]+(c)[18]*(x)+(c)[19]*(y))*(y))*(y))

#define CP40_NCOEFS      10 /* number of distortion correction coefficients
			       per polynomial */
#define CP40_COEF_COUNT (2*CP40_NCHANNELS*CP40_NCOEFS)
#define CP40_COEF_SIZE  (sizeof(double) * CP40_COEF_COUNT)

typedef enum {
  /* Possible distortion correction methods: */
  CP40_CORRECTION_NONE=     1, /* No distortion correction, the pixel sampling
				  is defined by the "pixel" attribute. */
  CP40_CORRECTION_SIMPLE=   2, /* Transformed positions get rounded to pixel
				  precision. */
  CP40_CORRECTION_RANDOM=   3, /* A sub-raw-pixel position is randomly
				  generated, the transformed positions get
				  rounded to pixel precision. */
  CP40_CORRECTION_WEIGHTED= 4  /* Each photon gives several positions with
				  different weights, those weights and
				  positions are tabulated. */
} cp40_correction_t;

/* Routines to query/set method used for distortion correction. */
extern cp40_correction_t cp40_get_correction(cp40_instance_t *instance);
extern int cp40_set_correction(cp40_instance_t *instance,
			       cp40_correction_t value);

extern double cp40_get_angle(cp40_instance_t *instance);
extern int    cp40_set_angle(cp40_instance_t *instance, double angle);
/* Query or set rotation angle (in degrees anti-clockwise) by
   distortion correction.  The "set" routine return standard result:
   GPL_OK or GPL_ERROR. */

extern double cp40_get_scale(cp40_instance_t *instance);
extern int    cp40_set_scale(cp40_instance_t *instance, double scale);
/* Query or set magnification scale (in pixels per grid step) by distortion
   correction.  The "set" routine return standard result: GPL_OK or GPL_ERROR.
   A value SCALE=0 means no distortion correction. */

extern int cp40_set_raw_coefs(cp40_instance_t *instance, double *raw_coefs);
extern void cp40_get_raw_coefs(cp40_instance_t *instance, double *raw_coefs);
extern void cp40_get_coefs(cp40_instance_t *instance, double *coefs);
/* Query/set raw (unscaled) or effective (scaled) distortion correction
   coefficients.  There are CP40_COEF_COUNT coefficients (CP40_NCOEFS per
   axis per channel). */

extern int cp40_shift(cp40_instance_t *instance, int channel,
		      double dx, double dy);
/* Modify raw coefficients of INSTANCE so that undistorted image of CHANNEL
   gets shifted by offset (DX,DY).  The offsets are given in undistorted
   pixel units (i.e.  the routine takes care of the current ANGLE and SCALE
   attributes of INSTANCE). */

extern int cp40_reweight_table(cp40_instance_t *this, int channel,
			       const double weight[],
			       int width, int height);
/* Apply flat field correction to weighted distortion correction.  WEIGHT
   is a WIDTH by HEIGHT array of weights (= 1/flat) computed _without_
   distortion correction. The dimensions must match those of the CCD
   for the raw pixel resolution as set in THIS. */

/*---------------------------------------------------------------------------*/
/* PRE-PROCESSED PHOTON FRAME */

typedef struct _cp40_frame_t   cp40_frame_t;
typedef struct _cp40_photon_t  cp40_photon_t;

#define CP40_FORCE_REPROCESSING 0x8
extern cp40_frame_t *cp40_get_frame(cp40_instance_t *instance, int flags);
/* Return a frame of preprocessed events from current raw (super-)frame of
   INSTANCE.  FLAGS is the channel number, possibly OR'ed with
   CP40_FORCE_REPROCESSING (to skip setp 1 below). The following steps are
   involved:
     1. If INSTANCE already has a frame for CHANNEL synchronized with
        current raw-frame date, the address of this frame is returned.
     2. INSTANCE->frame[CHANNEL] is created if it does not exists.
     3. Selection based on "parity": if INSTANCE->parity does not match the
        parity of INSTANCE->frame_date, INSTANCE->frame[CHANNEL] is
        returned with INSTANCE->frame[CHANNEL]->status=0 (frame was not
        selected) and INSTANCE->frame[CHANNEL]->count=0 (no events).
     4. Select raw-events (cp40_raw_photon_t) from CHANNEL and convert
        them to event (cp40_photon_t), deal with "spurious" events and
        raw-pixel "pixel" resolution.
     5. Apply distortion "correction", and "window" rules.
     6. Compute frame statistics.
     7. Apply other selection rules ("flux", "spread", "distance", "ratio").

   Return value is INSTANCE->frame[CHANNEL] (see FRAME->status for different
   selection cases) or NULL on error (error message can be recovered by
   gpl_error_info). */

struct _cp40_frame_t {
  /* Settings for photons selection: after a call to cp40_preprocess,
     FRAME->buffer is filled with photons from FRAME->channel and with
     positions in [FRAME->xmin:FRAME->xmax, FRAME->ymin:FRAME->ymax]. */

  cp40_instance_t *owner;     /* where events come from */
  int channel;                /* selected channel */
  int xmin, xmax, ymin, ymax; /* coordinate boundaries */
  int status;                 /* after a call to cp40_preprocess():
                                  0 if the current frame was rejected (check
                                    FRAME->count to see if the selection occurs
				    before/after the preprocessing of events);
				  1 if the, possibly empty, current frame was
				    effectively selected;
				 -1 on error */
  cp40_photon_t *buffer;      /* buffer of selected and preprocessed photons */
  int size;	              /* buffer size (maximum count) */
  int date;                   /* frame number */
  int count;                  /* actual number of photons in buffer */

  /* Frame statistics. */
  double flux;	              /* total of photon weights (set to -1 when
				 the frame has not yet been pre-processed). */
  double x, y;                /* photo-centre coordinates */
  double spread;              /* spread of photons (in pixels) */
  double distance;            /* distance of photo-centre to the closest
				 edge */
  double ratio;               /* ratio FRAME->distance / FRAME->spread */
};

struct _cp40_photon_t {
  /* Photon-event as stored in cp40_frame_t. */
  double w;	/* weight */
  int    x;	/* abscissa */
  int    y;	/* ordinate */
};


/*---------------------------------------------------------------------------*/
/* DATA STREAM */

extern void cp40_init_stream(cp40_stream_t *stream);
/*	Initialize CP40 stream access informations.  Must be called before
	using the stream. */

extern int cp40_assoc_file(cp40_instance_t *instance, const char *name,
			   int mode);
extern int cp40_assoc_socket(cp40_instance_t *instance, const char *host,
			      int port);
extern int cp40_close_stream(cp40_instance_t *instance, int update);

extern int cp40_read_header(cp40_instance_t *instance);
extern int cp40_write_header(cp40_instance_t *instance);
extern int cp40_set_exposure(cp40_instance_t *instance, int exposure);

extern const char *cp40_get_io_filename(cp40_instance_t *instance);
/*	Return stream filename of INSTANCE or NULL if instance is NULL or
	have no associated file stream. */

extern const char *cp40_get_io_hostname(cp40_instance_t *instance);
/*	Return server hostname of INSTANCE or NULL if instance is NULL or
	have no associated socket stream. */

extern int cp40_get_io_port(cp40_instance_t *instance);
/*	Return server port number of INSTANCE or -1 if instance is NULL or
	have no associated socket stream. */

extern int cp40_get_io_mode(cp40_instance_t *instance);
/*	Return data stream access mode of INSTANCE: CP40_IO_MODE_NONE (no
	stream or INSTANCE is NULL), CP40_IO_MODE_READ, CP40_IO_MODE_WRITE or
	CP40_IO_MODE_UPDATE. */

extern int cp40_get_io_type(cp40_instance_t *instance);
/*	Return data stream type of INSTANCE: CP40_IO_NONE (no stream
	or INSTANCE is NULL), CP40_IO_FILE, CP40_IO_SOCKET. */

extern int cp40_readable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream that can be readed in the sense
	that cp40_read_frame can be acted on it. */

extern int cp40_rewindable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream that can be rewinded in the sense
	that cp40_rewind_stream can be acted on it. */

extern int cp40_updatable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream whose header can be updated in the
	sense that cp40_write_header can be acted on it. */

extern int cp40_writable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream that can be written in the sense
	that cp40_write_frame can be acted on it. */

/* Prototype of data stream routines. */
typedef int cp40_read_frame_op_t(cp40_instance_t *instance);
typedef int cp40_write_frame_op_t(cp40_instance_t *instance);
typedef int cp40_read_header_op_t(cp40_instance_t *instance);
typedef int cp40_write_header_op_t(cp40_instance_t *instance);
typedef int cp40_close_stream_op_t(cp40_instance_t *instance);
typedef int cp40_rewind_stream_op_t(cp40_instance_t *instance);
typedef int cp40_set_exposure_op_t(cp40_instance_t *instance, int exposure);

struct _cp40_io_file_t {
  FILE *handle;
  char *name;  /* malloc'ed file name */
};

struct _cp40_io_socket_t {
  /* CP40 socket (only used for reading raw photons and raw frames
     send by, e.g. OS9 photon server. */
  int   handle;  /* socket file descriptor */
  char *host;    /* malloc'ed host name */
  int   port;    /* port number */
  char *buffer;  /* dynamically allocated input data buffer */
  int   size;    /* size of photon buffer */
  int   pending; /* number of bytes stored in input buffer */
};

struct _cp40_stream_t {
  const cp40_stream_ops_t *ops; /* Table of operations on stream (must be the
				   first member). */
  int error; /* Non-zero on read/write error. Each successful attempt to
		read/write stream clear this flag. */
  int end;   /* Non-zero if no new frame could be read (i.g. end of input
		stream or not enough pending input data). Each successful
		attempt to read a new frame clear this flag. */
  union {
    cp40_io_file_t   file;
    cp40_io_socket_t socket;
  } io;
};

/* Structure to hold all stream-dependant operations. */
struct _cp40_stream_ops_t {
  int type; /* Data stream type; one of the following values: */
#define CP40_IO_NONE   0
#define CP40_IO_FILE   1
#define CP40_IO_SOCKET 2

  int mode; /* Data stream access mode; one of the following values: */
#define CP40_IO_MODE_NONE   0 /* No associated stream (must be 0). */
#define CP40_IO_MODE_READ   1 /* Data can be read, parameters can be modified,
				 file header cannot be updated (must be 1). */
#define CP40_IO_MODE_WRITE  2 /* Can write data into a new file, can update
				 file header (must be 2). */
#define CP40_IO_MODE_UPDATE 3 /* Can modify file header and read data. */

  cp40_close_stream_op_t  *close;        /* close data stream */
  cp40_rewind_stream_op_t *rewind;       /* rewind stream to the beginning
					    of data */
  cp40_set_exposure_op_t  *set_exposure; /* set number of frames per
					    super-frame (only effective
					    in read/update modes) */
  cp40_read_frame_op_t    *read_frame;   /* read next raw data frame */
  cp40_write_frame_op_t   *write_frame;  /* write current raw data frame */
  cp40_read_header_op_t   *read_header;  /* (re)read header */
  cp40_write_header_op_t  *write_header; /* (over)write header */
};

/*---------------------------------------------------------------------------*/
/* CP40 DATA INSTANCE */

/* Prototype of photon selection routine. */
typedef int cp40_select_op_t(cp40_frame_t *frame_ptr);

typedef struct _cp40_extra_t cp40_extra_t;

struct _cp40_instance_t {

  /* INFORMATION ABOUT DATA FILE (no incidence on processing) */
  int nphotons[CP40_NCHANNELS];         /* Total photons per channel. */
  int nframes;                          /* Total number of frames. */
  int nsaturations;                     /* Total number of saturations. */
  
  /* PHOTON BUFFERS FOR THE CURRENT (SUPER-)FRAME */
  cp40_stream_t stream;                 /* Associated data stream and access
					   functions */
  cp40_raw_frame_t *raw_frame;          /* (Super-) frame of raw-photons.
					   Note: super-frames can only be used
					   in read mode of regular data files
					   (i.e., not raw data files). */
  int exposure;                         /* Number of elementary frames per
                                           super-frame, should be >= 1,
                                           irrelevant for write mode. */
  int *count_list;                      /* Counts per elementary raw frames,
                                           has at least EXPOSURE elements
                                           (irrelevant for WRITE mode). */
  cp40_frame_t frame[CP40_NCHANNELS];   /* Cached pre-processed photon-
					   counting frame for every
					   channels. */

  /* TASKS AND REAL-TIME STATISTICS */
  cp40_task_t *task;                    /* Chained list of tasks to perform. */
  int stat_size;                        /* Number of frames to remember. */
  int stat_index;                       /* Current position >=0 and
					   <STAT_SIZE */
  cp40_stat_t *stat;                    /* Array of structures for
					   statistics. */
#define CP40_FOV_REDUCTION 4 /* Shift for real-time FOV task. */
#define CP40_FOV_WIDTH     (2 * (4*CP40_CCD_WIDTH  >> CP40_FOV_REDUCTION))
#define CP40_FOV_HEIGHT    (2 * (4*CP40_CCD_HEIGHT >> CP40_FOV_REDUCTION))
#define CP40_FOV_NBYTES    (CP40_FOV_WIDTH * CP40_FOV_HEIGHT \
                            * sizeof(cp40_task_pixel_t))

  /* SETTINGS FOR DATA PRE-PROCESSING */
  int parity;                           /* Select even/odd numbered frames? */
  cp40_select_op_t *spurious;           /* Deal with spurious photons. */
  int pixel;                            /* Raw-pixel sampling. */
  cp40_select_op_t *correction;         /* Distortion correction method. */
  int notify;                           /* Non-zero if distortion correction
					   parameters (i.e. COEFS, EXTREME,
					   WGHT) must be recomputed due to a
					   change in one of CORRECTION,
					   RAWCOEF, SCALE, ANGLE or PIXEL. */
  double angle;                         /* Rotation angle (in degrees anti-
					   clockwise) after distortion
					   correction. */
  double scale;                         /* Scale after distortion correction
			 		   in pixels per grid step. SCALE=0
					   means no distortion correction. */
  double raw_coefs[CP40_COEF_COUNT];    /* Public (unscaled) set of coefs. */
  double coefs[CP40_COEF_COUNT];        /* Private set of coefficients. */
  cp40_window_t window[CP40_NCHANNELS]; /* Window boundaries as set by user. */
  cp40_window_t extreme[CP40_NCHANNELS];/* Extreme bounds which depend on the
					   distortion correction settings or
					   on the raw-pixel sampling. */
  void *wght[CP40_NCHANNELS];           /* Table for weighted distortion
					   correction. */
  cp40_range_t distance[CP40_NCHANNELS];/* Min/max distance to the edges. */
  cp40_range_t flux[CP40_NCHANNELS];    /* Min/max photon counts. */
  cp40_range_t ratio[CP40_NCHANNELS];   /* Min/max ratio distance/spread. */
  cp40_range_t spread[CP40_NCHANNELS];  /* Min/max image spread. */
  
  /* EXTRA ATTRIBUTES */
  gpl_hash_table_t    extra_table;            /* Hash table for extra-attributes. */
  cp40_extra_t *extra_list;             /* List of ordered entries in
					   EXTRA_TABLE. */
  int           extra_count;            /* Number of extra attributes entries
					   in EXTRA_LIST. */
  int           extra_size;             /* Maximum Number of extra attributes
					   entries in EXTRA_LIST. */
  
  /* HOOK FOR EXTERNAL CLIENT PROGRAMS THAT WOULD IMPORT CP40 FACILITIES */
  void  *client_data;                 /* Anything needed by the client. */
  void (*delete_client_data)(void *); /* If non-NULL, client routine used to
                                         free INSTANCE->client_data when
                                         INSTANCE gets deleted. */
};

/*---------------------------------------------------------------------------*/
/* RAW PHOTON FRAME */

/* Raw  (super-)frame is buffered into cp40_instance_t  structure all channels
   mixed together.   The photons  from a particular  channel are  selected and
   pre-processed   by  cp40_preprocess_photons  and   stored  in   a  struture
   cp40_frame_t. */

struct _cp40_raw_frame_t {
  cp40_raw_photon_t *buffer; /* Dynamically allocated photon buffer. */
  int date;                  /* Frame number. */
  int count;                 /* Actual number of photons in buffer. */
  int size;                  /* Maximum number of photons in buffer. */
#define CP40_RAW_FRAME_STATIC_BUFFER_SIZE 1020
  cp40_raw_photon_t static_buffer[CP40_RAW_FRAME_STATIC_BUFFER_SIZE];
};

extern cp40_raw_frame_t *cp40_new_raw_frame(void);
/*	Allocate and instanciate a new FRAME.  Return NULL in case of error. */

extern void cp40_delete_raw_frame(cp40_raw_frame_t *frame);
/*	Free FRAME and its resources. */

void cp40_reset_raw_frame(cp40_raw_frame_t *frame);
/*	Reset raw photon buffer of FRAME to be empty. */

extern int cp40_resize_raw_frame(cp40_raw_frame_t *frame, int new_size);
/*	Set size of raw photon buffer of FRAME to be NEW_SIZE.  The operation
	should be protected against interrupts unless compiler optimizations
	change the order of operations.  A standard result is returned: GPL_OK
	or GPL_ERROR.  Note: if NEW_SIZE is smaller or equal the size of the
	static buffer embedded into FRAME, then this static buffer gets used
        and FRAME->size becomes the size of the static buffer. */

/* PROTECTION AGAINST INTERRUPTS
     A routine protected against interrupts can be safely interrupted (e.g. by
     a signal) because it takes care to keep everything in a consistent state.
     The only damage an interrupt can cause is the loss of some dyamically
     allocated buffer. */
/*---------------------------------------------------------------------------*/

/*_____________________________________________________________________________
 *
 * Basic operations
 * ================
 *
 *	You must first create a new CP40 data instance (i.e. allocate and
 *	instanciate with defaults a new cp40_instance_t structure) with
 *	cp40_new_instance(); e.g.:
 *
 *	    cp40_instance_t *instance = cp40_new_instance();
 *	    if (instance == NULL) {
 *		fprintf(stderr, "%s\n", gpl_error_info(1));
 *		exit(3);
 *	    }
 *
 *	At this stage, you can only get or change attributes of the new
 *	CP40 data instance (see section "Data attributes").  In order to
 *	perform input/output operation (that is reading or writing photons),
 *	you must associate the CP40 data instance with a stream with
 *	functions: cp40_assoc_file() or cp40_assoc_socket().  The input/output
 *	operations are described in section "Data input/output and
 *	pre-processing".
 *
 *	When a CP40 data instance has an associated stream, you can use
 *	cp40_read_header() to read or read again CP40 data file header, e.g.
 *	to revert attribute changes.  cp40_read_header() is automatically
 *	called when a data stream is open in read or update mode to initialize
 *	attribute values.  Besides, cp40_write_header() (re)write CP40 data
 *	stream header, e.g. to update attribute changes.  cp40_write_header()
 *	may be automatically called when closing a data stream which was open
 *	in write or update mode.
 *
 *	When the CP40 data instance is no more needed, you just call
 *	cp40_delete_instance(instance, 1/0) to release all related resources
 *	(if second argument is non-zero, cp40_delete_instance takes care of
 *	updating header if this is relevant).
 */
extern cp40_instance_t *cp40_new_instance(void);
extern int cp40_delete_instance(cp40_instance_t *instance, int update);
extern int cp40_get_file_version(const char *path);

/*---------------------------------------------------------------------------*/
/* HIGH-LEVEL INTERFACE FOR INSTANCE ATTRIBUTES */

/* CP40 library provides you with two levels of interface routines to manage
   CP40 attributes.

   The  low-level  interface  routines  query  or  set  *STANDARD*  attributes
   (i.e. those in cp40_instance_t  structure) by value.  Since some attributes
   have side effects (e.g.  modifying the distortion correction method trigger
   the  updating of  distortion  correction parameters)  and  some others  are
   read-only,  the user  should use  these routines  preferably  than directly
   setting/getting  values in  the cp40_instance_t  structure.   The interface
   routines handle these underlying side effects.
   
   By  contrast, the  high-level  interface routines  handle *ALL*  attributes
   (i.e. the standard ones plus the extra ones) by human-readable keywords and
   string values.  In effect, they reference all attributes by their names and
   they convert their values to/from human readable strings.  Extra-attributes
   all  have string  values  (i.e. no  binary  counterpart in  cp40_instance_t
   structure)  and  are  stored the  private  hash  table  of each  CP40  data
   instance.  Only  two routines are  sufficient to implement this  high level
   interface (plus a further one to get the name and value of extra attributes
   in the same order as they were created). */

extern int cp40_get(cp40_instance_t *instance, const char *name, 
		    gpl_dstr_t *value);
/* Get  human-readable string value of  attribute NAME for  CP40 data INSTANCE
   and store  it into  dynamic string VALUE.   VALUE must have  been correctly
   initialized by gpl_dstr_init.  Return standard result: GPL_OK or GPL_ERROR (error
   message can be retrieved by gpl_error_info). */

extern int cp40_set(cp40_instance_t *instance, const char *name,
		    const char *value);
/* Use human-readable string VALUE to set the value of attribute NAME for CP40
   data INSTANCE.  Return standard result: GPL_OK or GPL_ERROR (error message
   can be retrieved by gpl_error_info). */

extern int cp40_get_next_extra(cp40_instance_t *instance, int *index_ptr,
			       const char **name_ptr, const char **value_ptr);
/*  Get  name and  value  of  next extra  attribute  for  CP40 data  INSTANCE.
   INDEX_PTR is the  address where is stored index  of current extra attribute
   (contents  should be initialized  to a  negative value  to get  first extra
   attribute) also used to store  index of next extra attribute.  NAME_PTR and
   VALUE_PTR  are addresses  where to  store  next attribute  name and  value.
   Return value  is 1 upon success,  0 if no more  extra attributes (*NAME_PTR
   and *VALUE_PTR set to NULL) or -1 in case of error. */

/*---------------------------------------------------------------------------*/
struct _cp40_extra_t {
  /* Structure to store extra (i.e.  non standard) attribute value (to be used
     in  conjonction with a  hash table  so as  to combine  the features  of a
     reference-by-keyword table and reference-by-order list).  */
  gpl_hash_entry_t *entry;     /* Corresponding entry in hash table. */
  const char *string;	 /* Address of dynamic memory to store the extra
			    attribute value. */
};

struct _cp40_attr_entry_t {
  /* An attribute table is a  gpl_hash_table indexed by label strings.  The value of
     each entry  in this  ``label table'' is  a malloc'ed gpl_hash_table  indexed by
     attribute names.  The  value of each entry in these  ``name tables'' is a
     malloc'ed cp40_attr_entry structure and the corresponding attribute value
     is a malloc'ed string. */
  gpl_hash_entry_t *label;	/* Attribute label. */
  gpl_hash_entry_t *name;	/* Attribute name. */
  const char *value;	/* Attribute value (malloc'ed string). */
};

/* The following macros take a pointer to a valid cp40_attr_entry. */
#define cp40_get_attr_value(ptr)  ((ptr)->value)
#define cp40_get_attr_name(ptr)   ((ptr)->name->key.string)
#define cp40_get_attr_label(ptr)  ((ptr)->label->key.string)
#define cp40_init_attr_table(ptr) gpl_hash_init_table(ptr, GPL_HASH_STRING_KEYS)

extern cp40_attr_entry_t *cp40_find_attr_entry(gpl_hash_table_t *label_table,
					       const char *label,
					       const char *name);
extern cp40_attr_entry_t *cp40_create_attr_entry(gpl_hash_table_t *label_table,
						 const char *label,
						 const char *name,
						 int *new_ptr);
extern void cp40_delete_attr_entry(cp40_attr_entry_t *attr_entry);
extern void cp40_delete_attr_table(gpl_hash_table_t *label_table);
extern int  cp40_set_attr_value(cp40_attr_entry_t *attr_entry,
				const char *value);
extern int  cp40_create_attr_entry_with_value(gpl_hash_table_t *lab_table,
					      const char *label,
					      const char *name,
					      const char *value);

/* Macro version of cp40_create_attr_entry_with_value. */
#define cp40_create_attr_entry_with_value(table, label, name, value) \
    cp40_set_attr_value(cp40_create_attr_entry(table, label, name, NULL), value)

/* The following is to save/restore attribute settings. */
#define CP40_ATTR_STANDARD        (1<<0) /* Save/restore only standard
					    attributes (Note: read only
					    attributes are ignored or get
					    commented). */
#define CP40_ATTR_EXTRA           (1<<1) /* Save/restore only extra (non-
					    standard) attributes. */
#define CP40_ATTR_PREPROCESSING   (1<<2) /* Save/restore only attributes
					    involved in pre-processing. */
#define CP40_ATTR_ALL             (CP40_ATTR_STANDARD|CP40_ATTR_EXTRA)

extern int cp40_save_attr(cp40_instance_t *instance, const char *filename,
			  int flags);
extern int cp40_restore_attr(cp40_instance_t *instance, const char *filename,
			     int flags);
extern int cp40_copy_attr(cp40_instance_t *dst_ptr,
	      /* const */ cp40_instance_t *src_ptr, int flags);

/*---------------------------------------------------------------------------*/

#define CP40_UPDATE_CACHE_MAYBE(ip) if ((ip)->notify) cp40_update_cache(ip)

extern void cp40_update_cache(cp40_instance_t *instance);
/* (Re)compute distortion correction settings and frame boundaries.  Must be
   called before any pre-processing operation as soon as INSTANCE->notify
   is non-zero.  The routine computes the values of:
     INSTANCE->coefs
     INSTANCE->extreme
     INSTANCE->frame[0..3].xmin (as well as xmax, ymin and ymax)
   according to the values of:
     INSTANCE->raw_coefs
     INSTANCE->pixel
     INSTANCE->correction
     INSTANCE->angle
     INSTANCE->scale
     INSTANCE->window[0..3]
   and set INSTANCE->notify to zero.   If you want to force computation
   of settings do:
     INSTANCE->notify= 3;
     cp40_update_cache(INSTANCE);
*/

extern int cp40_get_nframes(cp40_instance_t *instance);
/* Get total number of frames in a CP40 data file. Return INSTANCE->nframes
   or -1 for invalid argument. */

extern int cp40_get_nsaturations(cp40_instance_t *instance);
/* Get total number of saturations in a CP40 data file. Return
   INSTANCE->nsaturations or -1 for invalid argument. */

extern int cp40_get_nphotons(cp40_instance_t *instance, int channel);
/* Get total number of photons in given channel of a CP40 data file.
   Return INSTANCE->nphotons[CHANNEL] or -1 for invalid argument. */

extern int cp40_get_frame_count(cp40_instance_t *instance);
/* Get total number of photons in current (super-)frame buffer of a CP40
   data file.  Return INSTANCE->raw_frame->count or -1 for invalid argument. */

extern int cp40_get_frame_date(cp40_instance_t *instance);
/* Get total date (number) of current (super-)frame buffer of a CP40
   data file.  Return INSTANCE->raw_frame->date or -1 for invalid argument. */

extern int cp40_get_exposure(cp40_instance_t *instance);
/* Get current number of short-exposure frames per super-frame.  Return
   INSTANCE->exposure or -1 for invalid argument. */

/*---------------------------------------------------------------------------*/

extern int cp40_get_limits(cp40_instance_t *instance, int channel,
			   int *xmin, int *xmax, int *ymin, int *ymax);
/*  Get  coordinate boundaries  after  pre-processing.  Distortion  correction
   settings get updated if needed. */

extern int cp40_get_limits2(cp40_instance_t *instance, int chn1, int chn2,
			    int *xmin, int *xmax, int *ymin, int *ymax);
/* Get coordinate boundaries after pre-processing for an auto/intercorrelation
   between CHN1 and CHN2.  Positions in  the result are taken as: POSITION_2 -
   POSITION_1 (i.e. CHN1 is the origin).  Return standard result. */


extern int cp40_set_window(cp40_instance_t *instance, int channel,
			   cp40_window_t *window_ptr);
extern int cp40_get_window(cp40_instance_t *instance, int channel,
			   cp40_window_t *window_ptr);
extern int cp40_get_extreme(cp40_instance_t *instance, int channel,
			    cp40_window_t *window_ptr);
extern int cp40_set_distance(cp40_instance_t *instance, int channel,
			     cp40_range_t *range_ptr);
extern int cp40_get_distance(cp40_instance_t *instance, int channel,
			     cp40_range_t *range_ptr);
extern int cp40_set_flux(cp40_instance_t *instance, int channel,
			 cp40_range_t *range_ptr);
extern int cp40_get_flux(cp40_instance_t *instance, int channel,
			 cp40_range_t *range_ptr);
extern int cp40_set_ratio(cp40_instance_t *instance, int channel,
			  cp40_range_t *range_ptr);
extern int cp40_get_ratio(cp40_instance_t *instance, int channel,
			  cp40_range_t *range_ptr);
extern int cp40_set_spread(cp40_instance_t *instance, int channel,
			   cp40_range_t *range_ptr);
extern int cp40_get_spread(cp40_instance_t *instance, int channel,
			   cp40_range_t *range_ptr);


/*_____________________________________________________________________________
 *
 * Data input/output and pre-processing
 * ====================================
 *
 *	The raw-photons from the current (super-)frame are directly stored
 *	into the corresponding cp40_instance_t structure.  In this internal
 *	buffer raw-photons from all channels are mixed together.  The macro
 *	cp40_read_frame() read next raw photon frame from data stream and
 *	store it into the internal super-frame buffer (cp40_instance_t must be
 *	`readable'; use CP40_Readable() function to check this).  The macro
 *	cp40_write_frame() triggers the writting of the internal raw-photon
 *	buffer into the data stream as a single raw frame (i.e. no
 *	super-frame support is provided in write/update mode).  In order to
 *	ensure that the current raw-photon buffer is large enough to store
 *	your raw-photons, use cp40_set_size with the number of events you
 *	intend to store in instance->buffer.  */
#define cp40_write_frame(ptr)    ((ptr)->stream.ops->write_frame(ptr))
extern int cp40_read_frame(cp40_instance_t *instance);
/*	Read next raw photon (super-)frame from CP40 data stream of INSTANCE.
	Possible return values are:
	   1 if a new frame was effectively read
	   0 if no new frame could be read (either because normal
	     end-of-stream or not enough pending input data).
	  -1 on error.
	You can also check the flags INSTANCE->io_error and INSTANCE->io_end
	for a diagnostic. */


/*	Processed photons (i.e. no more "raw" ones) are stored in the dynamic
 *	photon buffer of a cp40_frame_t structure.  A cp40_frame_t structure is
 *	created by a call to cp40_new_frame().  Before any use of a
 *	cp40_frame_t structure, you must configure it by a call to
 *	cp40_configure_frame() THIS IS MANDATORY to insure proper settings.
 *	You should not modify preprocessing options after
 *	cp40_configure_frame() since they affects the limits stored in the
 *	cp40_frame_t structure nor should you modify contents of frame_ptr.
 *	In order to select photons from the channel of interest and to
 *	apply pre-processing (as specified by attributes) you just call the
 *	function cp40_preprocess_photons() (which also instanciates statistics in
 *	the argument cp40_frame_t structure).  When you no more need the
 *	cp40_frame_t structure, you must free its internal buffer(s) with
 *	cp40_delete_frame().  You don't need to free your cp40_frame_t
 *	structure between calls to cp40_preprocess_photons() (even if you use it
 *	to get photons from a different channel) nor you need to set the
 *	size of its internal buffer with the function cp40_resize_frame()
 *	this is automatically done by cp40_preprocess_photons().
 *
 *	It is guaranted that photons stored in a cp40_frame_t structure
 *	by cp40_preprocess_photons() have been properly selected; in particular
 *	their positions are within the bounds returned by cp40_get_limits()
 *	(inclusive).  */

extern int cp40_import(cp40_instance_t *instance,
		       const cp40_raw_photon_t *buffer,
		       long number, int flags);
/*	Import `number` raw photons from `buffer` into writable CP40 stream
	`instance`.  Photon with highest bit cleared (set to 0) are used
	between frames: every positive photon marks the end/beginning of a
	frame if bit CP40_IMPORT_LAST is non-zero/zero in `flags`.  If bit
	CP40_IMPORT_WFS of `flags` is set, the imported frames get numbered
	according to the wavefront sensor number.   Frames can span
	across input buffer boundaries.  The routine returns ETL_OK or
	ETL_ERROR. */
#define CP40_IMPORT_LAST 0x1
#define CP40_IMPORT_WFS  0x2


/*_____________________________________________________________________________
 *
 * Integration
 * ===========
 *
 *	cp40_stat	Integrate frame statistics.
 *	cp40_it		Integrate (long exposure) image.
 *	cp40_ac		Integrate autocorrelation.
 *	cp40_ic		Integrate intercorrelation.
 *
 *	For all these functions: COUNT is the number of frames to read (if
 *	COUNT=0, photons from current super-frame get used);  NSELPTR is the
 *	address to store number of selected frames (even in case of error,
 *	you can trust this number);  WIDTH and HEIGHT are the dimensions of
 *	output array ARRAYPTR (for the cp40_stat function ARRAYPTR must have
 *	at least 7 x MAX(COUNT,1) elements); other arguments have obvious
 *	meaning.  The returned value is: 0 on end-of-data, -1 on error, and
 *	1 otherwise.  Pre-processing and frame selection are applied as
 *	indicated by the attributes of DATAPTR (with one exception: if
 *	COUNT=0 and if you have changed these attributes after the last
 *	previous frame selection).  These functions make use of 1 or 2
 *	private cp40_frame_t structures to gather photons (you don't have to
 *	worry about that).
 */
#define CP40_ALL_FRAMES  (2147483647L)	/* A hopefully large value. */

extern int cp40_stat(cp40_instance_t *instance, double *array_ptr,
		     long count, int chn, long *n_sel_ptr);

extern int cp40_it(cp40_instance_t *instance, long *array_ptr, long width,
		   long height, long count, int chn, long *n_sel_ptr);

extern int cp40_ac(cp40_instance_t *instance, long *array_ptr, long width,
		   long height, long count, int chn, long *n_sel_ptr,
		   int flags);

extern int cp40_ic(cp40_instance_t *instance, long *array_ptr, long width,
		   long height, long count, int chn1, int chn2,
		   long *n_sel_ptr);

/*---------------------------------------------------------------------------*/
/* TASKS AND REAL-TIME STATISTICS */
typedef double cp40_task_pixel_t; /* Data type for integration of CP40
				     estimators. */
struct _cp40_task_t {
  cp40_instance_t *owner;      /* This is needed for backward reference. */
  cp40_task_t *next;           /* Point to next task or NULL. */
#define CP40_TOKEN_SIZE 8      /* Maximum size (length+1) for task token
                                  string (8 should be enough). */ 
  char token[CP40_TOKEN_SIZE]; /* Task identifier. */
  cp40_task_op_t *op;          /* The operation to execute. */
  int run;                     /* Run task? */
  int chn1, chn2;              /* Photon-counting channels (<0 if
                                  irrelevant). */
  int width, height;           /* Dimensions of result array. */
  int xmin, xmax, ymin, ymax;  /* Boundaries of result array. */
  cp40_task_pixel_t **pixel;   /* Array of row pointers for fast indexing
                                  of 2-D array RESULT, e.g., just do:
                                    int **ptr= task->pixel;
                                    ptr[y][x]= ...; */
  cp40_task_pixel_t data[1];   /* Image to store the result. _MUST_ be the
				  last member: actual size will be sufficient
				  to store the result. */
};

struct _cp40_stat_t {
  cp40_raw_frame_t *raw_frame;
  double flux[CP40_NCHANNELS];  /* Flux for each channel after pre-processing. */
  double x[CP40_NCHANNELS];     /* Mean abscissae for ... */
  double y[CP40_NCHANNELS];     /* Mean ordinate for ... */
  double spread[CP40_NCHANNELS];/* Mean image spread for ... */
};

extern cp40_task_op_t *cp40_parse_task_token(const char *token,
					     int *chn1, int *chn2);
extern char           *cp40_print_task_token(char *token, const char *task,
					     int chn1, int chn2);
/* Parse/format task token from/into TOKEN (at least 5 bytes).
   cp40_parse_task_token() return address of task operation routine and store
   channel number(s) (or -1) in CHN1 and CHN2 (if not NULL).
   cp40_print_task_token() return TOKEN; both routines return NULL in case of
   error.*/

extern cp40_task_t *cp40_schedule_task(cp40_instance_t *instance,
				       const char *token);
/* Schedule a photon-counting task to run on data from INSTANCE.  The task
   type and related channels is given by TOKEN (see cp40_parse_task_token).
   If the task is already scheduled, its structure address is returned;
   otherwise resources for a new task are allocated and the address of the new
   stack structure is returned.  In case of error NULL is returned.  Note:
   INSTANCE keep track of its tasks so there is no need to remember the task
   structure address. */

extern cp40_task_t *cp40_get_task(cp40_instance_t *instance,
				  const char *token);
/* Return structure address (NULL in case of error) of task scheduled in
   INSTANCE and identified by TOKEN (see cp40_parse_task_token). */

extern void cp40_stop_task(cp40_task_t *task);
/* Stop a running task (its result is still visible).  Nothing appends if TASK
   is NULL or is already stopped. */

extern void cp40_start_task(cp40_task_t *task);
/* Start a stopped photon-counting task.  Nothing appends if TASK is NULL or
   already running. */

extern void cp40_cancel_task(cp40_task_t *task);
/* Cancel a photon-counting task. Free all related resources and the task is
   no longer existing.  Nothing appends if TASK is NULL. */

extern void cp40_clear_task(cp40_task_t *task);
/* Set all TASK pixels to zero. */

extern void cp40_start_all_tasks(cp40_instance_t *instance);
extern void cp40_stop_all_tasks(cp40_instance_t *instance);
extern void cp40_cancel_all_tasks(cp40_instance_t *instance);
extern void cp40_clear_all_tasks(cp40_instance_t *instance);
/*  Apply global task operation for all scheduled tasks of INSTANCE. */

extern int cp40_set_stat_duration(cp40_instance_t *instance, int size);
/* Set statistics duration to be SIZE frames for INSTANCE.  If SIZE<=0, free
   all resources related to integration of statistics.  This is intended for
   real-time operation. */

extern int cp40_run_tasks(cp40_instance_t *instance, int count);
/* If COUNT>=0, read and process as explained below at most that number of
   frames (not accounting for discarded frames); otherwise, i.e.  COUNT<0,
   read and process all remaining frames.  The number of processed frames (not
   accounting for discarded frames) is returned.  The caller can check
   INSTANCE->io_error for possible errors.  Real-time processes can safely
   ignore such errors.
	
   For each frame, the processing consists in:
     1. reading next raw (super-)frame;
     2. apply preprocessing;
     3. update statistics for real-time;
     4. update real-time field-of-view image;
     5. stole raw (super-)frame (this is needed for circular averaging
        involved in real-time operation). */

#define CP40_USE_NAME_AS_PREFIX (1<<0)
#define CP40_MAY_OVERWRITE      (1<<1)
extern int cp40_save_task(cp40_task_t *task, const char *name, int flags);
/* Set TASK image as FITS file NAME.  If bit CP40_USE_NAME_AS_PREFIX of FLAGS
   is set, the output file name will be: "NAME-TOKEN.fits"; where TOKEN is the
   task name.  Return standard result: GPL_OK or GPL_ERROR. */

struct _cp40_avg_stat_t {
  int nframes; /* sample size */
  double mean_flux, mean_x, mean_y, mean_spread;
};

extern int cp40_average_stat(cp40_instance_t *instance,
			     cp40_avg_stat_t avg[CP40_NCHANNELS]);
/* Fill AVG with average statistics from real-time information collected in
   INSTANCE (see cp40_run_tasks).  Return sample size (indication of the
   number of frames accounted for). */

/*---------------------------------------------------------------------------*/
/* MISCELLANEOUS */

extern int cp40_check_channel(int channel);
/* Check CP40 channel number. Return standard result: GPL_OK or GPL_ERROR.  In
   case of error, left an error message (see gpl_error_info). */

extern void *cp40_alloc(unsigned int size, const char *what);
extern void *cp40_alloc_clear(unsigned int size, const char *what);
/* Allocate and return address of SIZE bytes dynamic memory.  In case of
   error, return NULL and left an error message that can be recovered with
   gpl_error_info.  If not NULL, WHAT is used to produce an the error message
   in the form: "insufficient memory for WHAT".  In addition, cp40_alloc_clear
   fill the allocated memory with zeroes. */

/*---------------------------------------------------------------------------*/
/* PRIVATE RESOURCES/ROUTINES */

extern void _cp40_invalidate_cache(cp40_instance_t *instance);
/* Invalidate computed distortion correction settings to force them to be
   recomputed the next time they are needed.  All related resources
   (e.g. tables for weighted correction) get freed.  This affect: effective
   distortion correction coefficients, weighted correction tables, limits.
   WARNING: before calling this routine, make sure that there are no pending
   integration, in particular no scheduled tasks (i.e. INSTANCE->task must
   be NULL). */

extern void _cp40_init_frame(cp40_frame_t *frame,
			     cp40_instance_t *owner, int channel);
/* Instanciate photon-counting frame.  Must be called before first use.
   Must not be called again before internal buffers get freed by
   _cp40_resize_frame(FRAME, 0). */

extern int _cp40_resize_frame(cp40_frame_t *frame, int size);
/* (Re)alloc photon buffer of FRAME to be at least SIZE photons (SIZE be
   greater or equal actual FRAME->count).  Return standard result: GPL_OK or
   GPL_ERROR -- in case of error, a message is left that can be retrieved by a
   call to gpl_error_info().  If SIZE is 0, intenal buffers get freed. */

/*---------------------------------------------------------------------------*/
GPL_END_DECLS
#endif /* CP40_H */
