/*
 * cp40cmd_attr.c --
 *
 *      Source code for command to copy/extract CP40 atributes.
 *
 * Copyright (c) 1998-1999, Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *      $Id$
 *      $Log$
 *-----------------------------------------------------------------------------
 */

#include "cp40.h"

static int is_data(const char *name);

const char *help = "\n\
Save / restore CP40 attributes.\n\
\n\
Usage: cp40attr [OPTIONS] SRC DST[...]\n\
\n\
  SRC                    Source: either a CP40 data file or an editable\n\
                         attribute file.\n\
  DST [...]              Destination file(s): if existing, CP40 data file;\n\
                         otherwise editable attribute file.\n\
\n\
Options:\n\
  -h, --help             Print this help.\n\
  -a, --all              Save / restore all attributes.\n\
  -p, --preprocessing    Save/restore preprocessing attributes(this is the\n\
                         default).\n\
  -s, --standard         Save/restore standard attributes.\n\
  -x, --extra            Save/restore extra(i.e.non - standard) attributes.\n\n";

int main(int argc, const char *argv[])
{
  int i, j, c;
  int flags = 0;
  cp40_instance_t *src_ptr = NULL;
  cp40_instance_t *dst_ptr = NULL;

  /* Parse options. */
  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-') {
      break;
    }
    if (argv[i][1] == '-') {
      /* Long name attribute. */
      if (strcmp(argv[i], "--help") == 0) {
	fprintf(stderr, help);
	exit(0);
      } else if (strcmp(argv[i], "--all") == 0) {
	flags|= CP40_ATTR_ALL;
      } else if (strcmp(argv[i], "--preprocessing") == 0) {
	flags|= CP40_ATTR_PREPROCESSING;
      } else if (strcmp(argv[i], "--standard") == 0) {
	flags|= CP40_ATTR_STANDARD;
      } else if (strcmp(argv[i], "--extra") == 0) {
	flags|= CP40_ATTR_EXTRA;
      } else {
	goto syntax;
      }
    } else {
      /* Short name attribute. */
      if (argv[i][1] == 0)
	goto syntax;
      for (j = 1; (c = argv[i][j]) != 0; j++) {
	if (c == 'h') {
	  fprintf(stderr, help);
	  exit(0);
	} else if (c == 'a') {
	  flags|= CP40_ATTR_ALL;
	} else if (c == 'p') {
	  flags|= CP40_ATTR_PREPROCESSING;
	} else if (c == 's') {
	  flags|= CP40_ATTR_STANDARD;
	} else if (c == 'x') {
	  flags|= CP40_ATTR_EXTRA;
	} else {
	  goto syntax;
	}
      }
    }
  }
  if (flags == 0) flags= CP40_ATTR_PREPROCESSING;

  if (argc - i < 2) {
  syntax:
    fprintf(stderr, help);
    exit(1);
  }

  /* Parse and process source. */
  if (is_data(argv[i])) {
    /* Source is a CP40 data file. */
    if ((src_ptr= cp40_new_instance()) == NULL
	|| cp40_assoc_file(src_ptr, argv[i], CP40_IO_MODE_READ) != GPL_OK) {
      goto error;
    }
  } else {
    /* Source is an editable attribute file. */
    if ((src_ptr= cp40_new_instance()) == NULL
	|| cp40_restore_attr(src_ptr, argv[i], flags) != GPL_OK) {
      goto error;
    }
  }

  /* Parse and process destination(s). */
  while (++i < argc) {
    if (is_data(argv[i])) {
      /* Destination is a CP40 data file. */
      if ((dst_ptr= cp40_new_instance()) == NULL
	  || cp40_assoc_file(dst_ptr, argv[i], CP40_IO_MODE_UPDATE) != GPL_OK
	  || cp40_copy_attr(dst_ptr, src_ptr, flags) != GPL_OK
	  || cp40_delete_instance(dst_ptr, 1) != GPL_OK) {
	goto error;
      }
      dst_ptr= NULL;
    } else {
      /* Destination is an editable attribute file. */
      if (cp40_save_attr(src_ptr, argv[i], flags) != GPL_OK) {
	goto error;
      }
    }
  }
  if (src_ptr && cp40_delete_instance(src_ptr, 0) != GPL_OK) {
    goto error;
  }
  return 0;
 error:
  fprintf(stderr, "ERROR: %s\n", gpl_error_info(1));
  return 1;
}

static int is_data(const char *name)
{
  int version= cp40_get_file_version(name);
  if (version < 0) {
    (void)gpl_error_info(1);	/* Clear error. */
    return 0;
  }
  return version > 0;
}
