#include <stdio.h>
#include <string.h>
#include <stdlib.h>


#define UInt32 unsigned long    /* unsigned 32-bit integer */


/* Definitions for byte order, according to significance of bytes, from low
   addresses to high addresses.  The value is what you get by putting '4' in
   the most significant byte, '3' in the second most significant byte, '2' in
   the second least significant byte, and '1' in the least significant byte. */

#undef LITTLE_ENDIAN
#define	LITTLE_ENDIAN 1234

#undef BIG_ENDIAN
#define	BIG_ENDIAN    4321

#undef PDP_ENDIAN
#define	PDP_ENDIAN    3412

static int ByteOrder(void);

typedef int ReadProc(void *ptr, int number, FILE *stream);
static ReadProc ReadUInt32_NoSwap, ReadUInt32_Swap;

static void Count(long counts[], const UInt32 buffer[], int number);

static void AllocBuffer(UInt32 **buffer, int number);

static const char *help= "\n\
USAGE: cp40count [OPTIONS] FILE1 [[OPTIONS] FILE2 [...]]\n\
\n\
OPTIONS:\n\
-?, --help   Print this help.\n\
-r, --raw    Next files are raw ones produced by OS9 (16kb header).\n\
-o, --old    Next files are old raw ones (512b header, 1Mb data, ...).\n\
-n, --new    Next files are imported ones.\n\n";

#define OLD_DATA_CHUNK ((2048*512)/4)

int main(int argc, const char *argv[])
{
  FILE *file;
  int raw= 0, old= 0;
  ReadProc *ReadUInt32;
  UInt32 *buffer= NULL, hdr[2];
  int i, j, nread, flux= 0, size= 0;
  long counts[4];
    
  if (argc<2) {
    fputs(help, stderr);
    return 1;
  }

  switch (ByteOrder()) {
  case BIG_ENDIAN:
    ReadUInt32= ReadUInt32_NoSwap;
    break;
  case LITTLE_ENDIAN:
    ReadUInt32= ReadUInt32_Swap;
    break;
  default:
    fputs("unsupported byte order\n", stderr);
    exit(1);
  }

  /* Pre-allocate buffer. */
  size= OLD_DATA_CHUNK;
  AllocBuffer(&buffer, size);

#undef SAME
#define SAME(s1,s2) (strcmp(s1,s2)==0)
  for (i=1 ; i<argc ; i++) {
    if (SAME(argv[i], "-?") || SAME(argv[i], "--help")) {
      fputs(help, stdout);
      return 0;
    }
    if (SAME(argv[i], "-o") || SAME(argv[i], "--old")) {
      old= raw= 1;
      continue;
    }
    if (SAME(argv[i], "-r") || SAME(argv[i], "--raw")) {
      old= 0;
      raw= 1;
      continue;
    }
    if (SAME(argv[i], "-n") || SAME(argv[i], "--new")) {
      raw= old= 0;
      continue;
    }
    if ((file= fopen(argv[i], "r"))==NULL) {
      fprintf(stderr, "unable to open '%s'\n", argv[i]);
      continue;
    }
    counts[0]= counts[1]= counts[2]= counts[3]= 0;
    if (old) {
      for (j=0; fseek(file, 512*(1+2049*j), SEEK_SET)==0
	     && (flux= ReadUInt32(buffer,OLD_DATA_CHUNK,file))>0;
	   j++) {
	Count(counts, buffer, flux);
      }
    } else if (fseek(file, 16*1024, SEEK_SET)==0) {
      if (raw) {
	while ((nread= ReadUInt32(buffer,size,file))>0) {
	  Count(counts, buffer, nread);
	}
      } else {
	while (ReadUInt32(hdr,2,file)==2) {
	  flux= hdr[1];
	  if (flux==0) continue;
	  if (size<flux) {
	    size= 2*flux;
	    AllocBuffer(&buffer, size);
	  }
	  if ((nread= ReadUInt32(buffer,flux,file))>0) {
	    Count(counts, buffer, nread);
	  }
	  if (nread<flux) {
	    fprintf(stderr, "%s: unexpected end of file\n",
		    argv[i]);
	    break;
	  }
	}
      }
    }
    fclose(file);
    printf("%s: %ld %ld %ld %ld [%ld]\n", argv[i],
	   counts[0], counts[1], counts[2], counts[3],
	   counts[0]+counts[1]+counts[2]+counts[3]);
  }
  return 0;
}

static int ReadUInt32_NoSwap(void *ptr, int number, FILE *stream)
{
  return fread(ptr, 4, number, stream);
}

static int ReadUInt32_Swap(void *ptr, int number, FILE *stream)
{
  if ((number= fread(ptr, 4, number, stream))>0) {
    register int i, n= number*4;
    register char c, *data= (char *)ptr;	
    for (i=0 ; i<n ; i+=4) {
      c= data[i];   data[i]=   data[i+3]; data[i+3]= c;
      c= data[i+1]; data[i+1]= data[i+2]; data[i+2]= c;
    }
  }
  return number;
}

static void Count(long counts[], const UInt32 buffer[], int number)
{
  UInt32 p;
  int i;

  for (i=0 ; i<number ; i++) {
    p= buffer[i];
    if ((p&0x7ff)<(384<<2) && (p&0x7ff000)<(288<<14)) {
      switch (p & 0x800800) {
      case 0x000000: counts[0]+= 1; break;
      case 0x000800: counts[1]+= 1; break;
      case 0x800000: counts[2]+= 1; break;
      default:       counts[3]+= 1; break;
      }
    }
  }
}

static void AllocBuffer(UInt32 **buffer, int number)
{
  int i;
  UInt32 *new, *old= buffer[0];

  if ((new= malloc(number*4))==NULL) {
    fputs("unable to (re)allocate photon buffer\n", stderr);
    exit(1);
  }
  if (old) {
    for (i=0 ; i<number ; i++)
      new[i]= old[i];
    free(old);
  }
  *buffer= new;
}

static int ByteOrder(void)
{
  int i;
  UInt32 value= 0;
  char *ptr= (char*)&value;
    
  if (sizeof(UInt32)!=4) {
    fputs("failed assumption about data type size(s)\n", stderr);
    exit(1);
  }

  /* Byte ordering for integers is directly testable -- just need to
     compare set values in successive bytes by the shift left operator,
     then re-interpret them as characters to find the byte order.  */
  for (i=0 ; i<sizeof(value) ; i++) {
    value|= (sizeof(value)-i)<<(8*i);
  }
    
  for (i=0 ; i<sizeof(value) ; i++) {
    if (ptr[i]==1 && ptr[i+1]==2 && ptr[i+2]==3 && ptr[i+3]==4) {
      return BIG_ENDIAN;
    }
    if (ptr[i]==4 && ptr[i+1]==3 && ptr[i+2]==2 && ptr[i+3]==1) {
      return LITTLE_ENDIAN;
    }
    if (ptr[i]==2 && ptr[i+1]==1 && ptr[i+2]==4 && ptr[i+3]==3) {
      return PDP_ENDIAN;
    }
  }
  fputs("could not determine byte order\n", stderr);
  exit(1);
}
