/*
 * cp40cmd_import.c --
 *
 *	Import raw CP40 data files (those produced by the OS9 acquisition
 *	system).
 *
 * Copyright and warranty:
 *	Copyright (c) 1998 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Log: cp40import.c,v $
 *	Revision 1.2  1999/04/09 14:42:42  eric
 *	 - Added option: --offset.
 *
 *	Revision 1.1  1999/04/09 14:21:54  eric
 *	Initial revision
 *
 *_____________________________________________________________________________
 */

const char *version= "$Id: cp40import.c,v 1.2 1999/04/09 14:42:42 eric Exp $";

#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include "gpl_xdr.h"
#include "gpl_option.h"
#include "gpl_string.h"
#include "cp40.h"

/*---------------------------------------------------------------------------*/
/* Private routines. */

static void Warning(const char *message);
static void FatalError(const char *message);
static void FatalErrorInElib(void);
static int YesOrNo(const char *question, const char *help, int def);

static void CloseOutput(cp40_instance_t **instance);
/*	Close output data file and set it to NULL. */

static void OpenOutput(cp40_instance_t **instance, const char *path);
/*	(Re)open output data file (takes care of closing it if needed). */


static long inputOffset= 0;
static FILE *OpenInput(const char *name);
/*	Open file NAME for reading only, (exit in case of error). */

static const char *time_stamp(unsigned char date[]);

static void StripString(char *dst, const char *src);
/*	Remove leading and trailing spaces and replace single/multiple spaces
	by a single ordinary space.  The operation can be performed
	`in-place'. */

typedef void ImportProc(const char *name, gpl_xdr_uint_t buffer[], int flags);
/*	Prototype of routine called to import data from file NAME.  BUFFER is
	the raw photon buffer (1Mb) or NULL if the routine should only list
	the contents of NAME.  FLAGS are for cp40_import. */
static ImportProc ImportVersion1p0;
static ImportProc ImportVersion2p0, ImportVersion2p1;
static ImportProc ImportVersion3p0, ImportVersion3p1;

#undef KILO_BYTE
#define KILO_BYTE    1024

#undef MEGA_BYTE
#define MEGA_BYTE    0x100000

/* Raw photon buffer. */
#undef BUFFER_SIZE
#define BUFFER_SIZE  0x100000        /* Must be one Mb */
#undef BUFFER_COUNT
#define BUFFER_COUNT (BUFFER_SIZE/4) /* Number of raw photons to read
					at a time (not adjustable). */

#ifndef PATH_MAX
# define PATH_MAX 1024 /* max # chars in a path name */
#endif

#undef FUNCTION
#define FUNCTION(NAME, XDR_TYPE, XDR_READ) \
static void NAME(FILE *stream, long offset, XDR_TYPE *ptr, int number);	      \
static void NAME(FILE *stream, long offset, XDR_TYPE *ptr, int number)	      \
{									      \
  if (fseek(stream, offset, SEEK_SET) != 0				      \
      || XDR_READ(ptr, number, stream) != number) {			      \
    gpl_sys_error("unable to read header");				      \
    exit(1);								      \
  }									      \
}
FUNCTION(read_b, gpl_xdr_uchar_t, gpl_xdr_read_byte)
     FUNCTION(read_c, gpl_xdr_char_t,  gpl_xdr_read_char)
     FUNCTION(read_i, gpl_xdr_int_t,   gpl_xdr_read_int)
     FUNCTION(read_f, gpl_xdr_float_t, gpl_xdr_read_float)
#undef FUNCTION

/*---------------------------------------------------------------------------*/
/* Print wavelength BandWidth. */
     static const char *PrintBW(double lambda, double delta_lambda);
     static const char *PrintBW(double lambda, double delta_lambda)
{
  static char string[100];
  sprintf(string, "%g / %g", lambda, delta_lambda);
  return string;
}

/*---------------------------------------------------------------------------*/
static void SetAttrQ(cp40_instance_t *instance, const char *name,
		     const char *value);
static void SetAttrQ(cp40_instance_t *instance, const char *name,
		     const char *value)
{
  if (cp40_set(instance, name, value) != GPL_OK)
    fprintf(stderr, "*** WARNING cannot set attribute `%s'\n", name);
}

static void SetAttrD(cp40_instance_t *instance, const char *name,
		     double value);
static void SetAttrD(cp40_instance_t *instance, const char *name,
		     double value)
{
  char valueStr[100];
  sprintf(valueStr, "%g", value);
  SetAttrQ(instance, name, valueStr);
}

#if 0
static void SetAttrL(cp40_instance_t *instance, const char *name, long value);
static void SetAttrL(cp40_instance_t *instance, const char *name, long value)
{
  char valueStr[100];
  sprintf(valueStr, "%ld", value);
  SetAttrQ(instance, name, valueStr);
}
#endif
/*---------------------------------------------------------------------------*/
static const char *GetPath(const char *message, const char *defaultPath,
			   int flags);
#define FILE_EXISTING           (1<<0)
#define FILE_NEW                (1<<1)
#if 0
#define FILE_READABLE           (1<<2)
#define FILE_WRITABLE           (1<<3)
#define FILE_EXECUTABLE         (1<<4)
#define FILE_BLOCK_DEV          (1<<5)
#define FILE_CHARACTER_DEV      (1<<6)
#define FILE_DIRECTORY          (1<<7)
#define FILE_REGULAR            (1<<8)
#define FILE_GID                (1<<9)
#define FILE_UID                (1<<10)
#define FILE_STICKY             (1<<11)
#define FILE_SYMBOLIC_LINK      (1<<12)
#define FILE_FIFO               (1<<13)
#define FILE_NON_EMPTY          (1<<14)
#define FILE_EMPTY              (1<<15)
#define FILE_SOCKET             (1<<16)
#endif

/*---------------------------------------------------------------------------*/

static const char *help= "\n\
Usage: cp40import [OPTION]... SRC [...]\n\
\n\
  -f, --format=NUM   format version of raw data SRC\n\
                       NUM=1   512b header, sequences of 1Mb\n\
                       NUM=2   512b header, one file per observation\n\
                       NUM=3.0 16kb empty header, one file per observation\n\
                       NUM=3.1 16kb header, one file per observation\n\
  -o, --offset=COUNT skip COUNT bytes in every input file\n\
  -l, --list         only list contents of raw data file\n\
  -w, --wfs          parse wavefront sensor number\n\
  -?, --help         print short this help and exit\n\
  --version          print version and exit\n\n";

int main(int argc, const char *argv[])
{
  static gpl_opt_def_t opt[]= {
    {"format",  "f",   GPL_OPT_ARG_REQUIRED, 'f'},
    {"list",    "l",   GPL_OPT_ARG_NONE,     'l'},
    {"wfs",     "w",   GPL_OPT_ARG_NONE,     'w'},
    {"offset",  "o",   GPL_OPT_ARG_REQUIRED, 'o'},
    {"version", NULL,  GPL_OPT_ARG_NONE,     '~'},
    {"help",    "?",   GPL_OPT_ARG_NONE,     '?'},
    {NULL,      NULL,  0,                  0},
  };
  gpl_opt_data_t optData;
  ImportProc *Import= NULL;
  int flags= 0;
  int i;
  const char *str;
  int onlyList= 0;
  gpl_xdr_uint_t *buffer= NULL; /* Buffer for input raw data. */


  /* Parse options. */
  if (argc<=1) {
    fputs(help, stderr);
    exit(1);
  }
  gpl_opt_init(&optData, argc, argv, opt, GPL_OPT_PERMUTE_ARGS);    
  while ((i= gpl_opt_next(&optData))!=GPL_OPT_END_OF_OPTIONS) {
    if (i<0) {
      fputs(gpl_error_info(1), stderr);
      fputc('\n', stderr);
      exit(1);		
    }
    switch (opt[i].value) {
    case 'f':
      if (Import!=NULL) {
	fprintf(stderr, "%s: too many raw data formats\n", argv[0]);
	exit(1);
      }
      str= optData.opt_arg;
      if (strcmp(str, "1")==0 || strcmp(str, "1.0")==0) {
	Import= ImportVersion1p0;
      } else if (strcmp(str, "2")==0 || strcmp(str, "2.0")==0) {
	Import= ImportVersion2p0;
      } else if (strcmp(str, "2.1")==0) {
	Import= ImportVersion2p1;
      } else if (strcmp(str, "3.0")==0) {
	Import= ImportVersion3p0;
      } else if (strcmp(str, "3")==0 || strcmp(str, "3.1")==0) {
	Import= ImportVersion3p1;
      }
      if (Import==NULL) {
	fprintf(stderr, "%s: bad raw data format\n", argv[0]);
	exit(1);
      }
      break;
    case 'l':
      onlyList= 1;
      break;
    case 'o':
      if (gpl_str_get_integer(optData.opt_arg, &inputOffset)!=GPL_OK
	  || inputOffset<0) {
	fprintf(stderr, "%s: bad offset value\n", argv[0]);
	exit(1);
      }
      break;
    case 'w':
      flags|= CP40_IMPORT_WFS;
      break;
    case '~':
      fprintf(stderr, "%s: %s\n\tbased on libraries: ",
	      argv[0], version);
      fprintf(stderr, "CP40 (version %s)\n", CP40_VERSION_STRING);
      exit(0);
      break;
    case '?':
    default:
      fputs(help, stderr);
      exit(0);
      break;
    }
  }

  if (Import==NULL) {
    fprintf(stderr, "%s: use option --format to specify raw data format\n",
	    argv[0]);
    exit(1);
  }
  if (onlyList) {
    if (Import==ImportVersion3p0 || Import==ImportVersion3p1) {
      fprintf(stderr,
	      "%s: option `--list' does not work with `--format=3'\n",
	      argv[0]);
      exit(1);
    }
  } else if ((buffer= malloc(MEGA_BYTE))==NULL) {
    fprintf(stderr, "%s: unable to allocate working buffer\n", argv[0]);
    exit(1);
  }

  /* Import files. */
  for (i=optData.arg_index ; i<argc ; i++) {
    Import(argv[i], buffer, flags);
  }
  free(buffer);
  return 0;
}

/*---------------------------------------------------------------------------*/

static FILE *OpenInput(const char *name)
{
  FILE *stream;
    
  if ((stream= fopen(name, "r"))==NULL) {
    fprintf(stderr, "unable to open `%s'\n", name);
    exit(1);
  }
  if (inputOffset>0 && fseek(stream, inputOffset, SEEK_SET)!=0) {
    fclose(stream);
    fprintf(stderr, "unable to skip %ld bytes in `%s'\n", inputOffset,
	    name);
    exit(1);
	
  }
  return stream;
}

/*---------------------------------------------------------------------------*/

#define BUFFER 1024
#define CHUNK  1024	/* Read data by multiple of this size (in bytes). */

#undef BUFFER_COUNT
#define BUFFER_COUNT (MEGA_BYTE/4)

/*---------------------------------------------------------------------------*/

static void ImportVersion1p0(const char *name, gpl_xdr_uint_t buffer[],
			     int flags);
/* PURPOSE:
     Import or list raw CP40 data from file STREAM.  If LISTONLY is non-zero,
     a simple listing of file contents is printed; otherwise interactively
     import data from NAME.

   DATA FORMAT:
     Data for the same object is constituted by several sequences of 2049
     blocks (one block is 512 bytes),  the first block contains the header
     and the 2048 other blocks (1Mb) contain the data.  The *LAST* photon
     of a frame is flagged (i.e. highest bit set to 0 in XDR representation).
     The raw header is organized as follow:

     .	 offset   data  number  contents                      field
     .	 ------  -----  ------  ----------------------------  -----------
     .	      0  UInt8       1  year - 1900                   year
     .	      1  UInt8       1  month (1-12)                  month
     .	      2  UInt8       1  day of the month (1-31)       day
     .	      3  UInt8       1  hour (0-23)                   hour
     .	      4  UInt8       1  minutes (0-59)                minutes
     .	      5  UInt8       1  seconds (0-59)                seconds
     .	      6  UInt8       4  unused (padding)
     .	     10  char      490  comments                      comments
     .	    500  Int32       1  size (4 * number of photons)  size
     .	    504  Int32       1  number of frames              nframes
     .	    508  Int32       1  number of saturations         nsaturations
     .	 ------  -----  ------  ----------------------------  -----------
     .	    512  Int32  262144  photons
     .	 ------  -----  ------  ----------------------------  -----------*/

#undef COMMENT_LENGTH
#define COMMENT_LENGTH 490

/* Input/output macros (assume `stream' is the name of the input file
   stream): */
#undef XDR_READ
#undef XDR_READ_OK
#define XDR_READ(ADDR, COUNT, TYPE)    (gpl_xdr_read_##TYPE(ADDR,COUNT,stream))
#define XDR_READ_OK(ADDR, COUNT, TYPE) (XDR_READ(ADDR, COUNT, TYPE)==(COUNT))
#undef XDR_WRITE
#undef XDR_WRITE_OK
#define XDR_WRITE(ADDR, COUNT, TYPE)   (gpl_xdr_write_##TYPE(ADDR,COUNT,stream))
#define XDR_WRITE_OK(ADDR, COUNT, TYPE) (XDR_WRITE(ADDR, COUNT, TYPE)==(COUNT))

static void ImportVersion1p0(const char *name,
			     gpl_xdr_uint_t buffer[], int flags)
{
  char comment[COMMENT_LENGTH+1];
  char previousComment[COMMENT_LENGTH+1];
  gpl_xdr_uchar_t date[10];
  gpl_xdr_uint_t misc[3];
  int nread, count, number;
  cp40_instance_t *instance= NULL;
  const char *path;
  const char *timeStamp;
  int i1, i2, append;
  FILE *stream;

  /* Initialization. */
  stream= OpenInput(name);
  clearerr(stream);
  previousComment[0]= 0;
  flags|= CP40_IMPORT_LAST;
    
  /* Loop over sequences. */
  for (number=1 ;  ; number++) {
    /* Read next header and print it. */
    if ((nread= XDR_READ(date, 10, char))!=10) {
      if (feof(stream)) {
	if (nread!=0) goto prematureEndOfFile;
	break;
      }
      goto readError;
    }
    if (       ! XDR_READ_OK(comment, COMMENT_LENGTH, char)
	       || ! XDR_READ_OK(misc, 3, long)) {
      if (feof(stream)) goto prematureEndOfFile;
      goto readError;
    }
    comment[COMMENT_LENGTH]= 0;
    StripString(comment, comment);
    timeStamp= time_stamp(date);
    printf("%2d: [%s] %s\n", number, timeStamp, comment);

    /* Just skip data part if list-only mode. */
    if (buffer==NULL) {
      if (fseek(stream, MEGA_BYTE, SEEK_CUR)) {
	if (ferror(stream)) {
	  FatalError("seek error");
	}
	goto prematureEndOfFile;
      }
      continue;
    }

    /* Read raw data (skip first null events and make sure that the raw
       data buffer is terminated by an end-of-frame, i.e. a 4-byte word
       with its highest bit set to zero).  Note: we must pre-compute the
       number of long's to read since _read() would fail if end-of-file is
       encountered before read completed. */
    count= XDR_READ(buffer, BUFFER_COUNT, long);
    if (count<1) {
      if (feof(stream)) goto prematureEndOfFile;
      goto readError;
    }
    for (i2=count-1 ; i2>=0 && (buffer[i2]&0x80000000)!=0 ; i2--)
      ;
    for (i1=0 ; i1<=i2 && (buffer[i1]&0x00ffffff)==0 ; i1++)
      ;
	
    /* Prepare output file. */
    if (instance==NULL) {
      append= 0;
    } else {
      append= YesOrNo("Append to current file?", "(no help)",
		      (strcmp(comment, previousComment)==0?1:0));
    }
    strcpy(previousComment, comment);
    if (! append) {
      /* New output file. */
      path= GetPath("Name of new output data file", NULL, FILE_NEW);
      OpenOutput(&instance, path);
      if (cp40_set(instance, "date", timeStamp)!=GPL_OK) {
	FatalErrorInElib();
      }
    } else {
      /* Flush photon buffer. */
      if (instance->raw_frame->count!=0
	  && cp40_write_frame(instance) != GPL_OK) {
	FatalErrorInElib();
      }
    }
    instance->raw_frame->date= 50*(date[5]+60*(date[4]+60*date[3]));
	
    /* Import data and increment number of saturations. */
    if (i1<=i2) {
      if (cp40_import(instance, buffer+i1, i2-i1+1, flags)!=GPL_OK) {
	FatalErrorInElib();
      }
      instance->nsaturations+= misc[2];
    }
  }
  fclose(stream);
  CloseOutput(&instance);
  return;
    
 prematureEndOfFile:
  fclose(stream);
  CloseOutput(&instance);
  Warning("premature end of input file");
  return;

 readError:
  fclose(stream);
  CloseOutput(&instance);
  FatalError("read error");
}

/*---------------------------------------------------------------------------*/

struct Header2p0 {
  gpl_xdr_uchar_t creation[8];  /* date of creation */
  gpl_xdr_uchar_t update[8];    /* date of update */
  gpl_xdr_int_t   address;      /* address of data buffer */
  gpl_xdr_int_t   size;         /* size of data buffer */
  gpl_xdr_int_t   data_type;    /* type of data */
  gpl_xdr_int_t   obs_type;     /* type of observation (0=SPECKLE, 1=SPECKLE_DIF,
				 2=DECONV, 3=SFM, 4=GI2T) */
  gpl_xdr_int_t   xsize;
  gpl_xdr_int_t   ysize;
  gpl_xdr_int_t   nframes;      /* number of frames */
  gpl_xdr_int_t   nsaturations; /* number of saturations */
  gpl_xdr_int_t   nphotons;     /* number of photons */
  gpl_xdr_int_t   flags;        /* pre-processing flags */

  gpl_xdr_char_t  comments[81]; /* object / comments */
  gpl_xdr_uchar_t date[8];      /* date (again!) */
  gpl_xdr_float_t lambda_1;
  gpl_xdr_float_t lambda_2;
  gpl_xdr_float_t delta_1;
  gpl_xdr_float_t delta_2;
  gpl_xdr_float_t density;
  gpl_xdr_float_t rotation;
  gpl_xdr_float_t focus;
  gpl_xdr_float_t temperature;
  gpl_xdr_float_t voltage_1;
  gpl_xdr_float_t voltage_2;
  gpl_xdr_float_t obturator;
};

static void ImportVersion2p0(const char *name,
			     gpl_xdr_uint_t buffer[], int flags)
{
  struct Header2p0 header;
  FILE *stream= OpenInput(name);
  long offset= ftell(stream);

  /* Read header. */
  read_b(stream, (offset+  0L),  header.creation,     8);
  read_b(stream, (offset+  8L),  header.update,       8);
  read_i(stream, (offset+ 16L), &header.address,      1);
  read_i(stream, (offset+ 20L), &header.size,         1);
  read_i(stream, (offset+ 24L), &header.data_type,    1);
  read_i(stream, (offset+ 28L), &header.obs_type,     1);
  read_i(stream, (offset+ 32L), &header.xsize,        1);
  read_i(stream, (offset+ 36L), &header.ysize,        1);
  read_i(stream, (offset+ 40L), &header.nframes,      1);
  read_i(stream, (offset+ 44L), &header.nsaturations, 1);
  read_i(stream, (offset+ 48L), &header.nphotons,     1);
  read_i(stream, (offset+ 52L), &header.flags,        1);
  read_c(stream, (offset+220L),  header.comments,    80);
  header.comments[80]= 0;
  read_b(stream, (offset+300L),  header.date,         8);
  read_f(stream, (offset+308L), &header.lambda_1,     1);
  read_f(stream, (offset+312L), &header.lambda_2,     1);
  read_f(stream, (offset+316L), &header.delta_1,      1);
  read_f(stream, (offset+320L), &header.delta_2,      1);
  read_f(stream, (offset+324L), &header.density,      1);
  read_f(stream, (offset+328L), &header.rotation,     1);
  read_f(stream, (offset+332L), &header.focus,        1);
  read_f(stream, (offset+336L), &header.temperature,  1);
  read_f(stream, (offset+340L), &header.voltage_1,    1);
  read_f(stream, (offset+344L), &header.voltage_2,    1);
  read_f(stream, (offset+348L), &header.obturator,    1);

  if (buffer==NULL) {
    /* Only list contents. */
    printf("*** %s (v2.0) ***\n", name);
    printf("  creation ......... %s\n", time_stamp(header.creation));
    printf("  last update ...... %s\n", time_stamp(header.update));
    printf("  address (?) ...... 0x%x\n", (unsigned)header.address);
    printf("  size ............. %d\n", header.size);
    printf("  data type ........ %d\n", header.data_type);
    printf("  observation type . %d\n", header.obs_type);
    printf("  xsize yxsize ..... %d x %d\n", header.xsize, header.ysize);
    printf("  nframes .......... %d\n", header.nframes);
    printf("  nsaturations ..... %d\n", header.nsaturations);
    printf("  nphotons ......... %d\n", header.nphotons);
    printf("  flags ............ %d\n", header.flags);
    printf("  comments ......... %s\n",  header.comments);
    printf("  date ............. %s\n", time_stamp(header.date));
    printf("  wavelength_1 ..... %s\n",
	   PrintBW(header.lambda_1, header.delta_1));
    printf("  wavelength_2 ..... %s\n",
	   PrintBW(header.lambda_2, header.delta_2));
    printf("  density .......... %g\n", header.density);
    printf("  rotation ......... %g\n", header.rotation);
    printf("  focus ............ %g\n", header.focus);
    printf("  temperature ...... %g\n", header.temperature);
    printf("  voltage_1 ........ %g\n", header.voltage_1);
    printf("  voltage_2 ........ %g\n", header.voltage_2);
    printf("  obturator ........ %g\n", header.obturator);
  } else {
    /* Import data file. */
    char defaultPath[PATH_MAX+1];
    cp40_instance_t *instance= NULL;
    const char *path;
    int count;

    if (fseek(stream, offset+512, SEEK_SET) != 0) {
      gpl_sys_error("unable to seek to data part of file");
      exit(1);
    }
    strcpy(defaultPath, name);
    strcat(defaultPath, ".cp40");
    path= GetPath("Name of new output data file", defaultPath, FILE_NEW);
    OpenOutput(&instance, path);

    SetAttrQ(instance, "date", time_stamp(header.creation));
    SetAttrQ(instance, "comments", header.comments);
    SetAttrQ(instance, "wavelength_1",
	     PrintBW(header.lambda_1, header.delta_1));
    SetAttrQ(instance, "wavelength_2",
	     PrintBW(header.lambda_2, header.delta_2));
    SetAttrD(instance, "voltage_1", header.voltage_1);
    SetAttrD(instance, "voltage_2", header.voltage_2);
    SetAttrD(instance, "temperature", header.temperature);
    SetAttrD(instance, "focus", header.temperature);
    SetAttrD(instance, "rotation", header.temperature);
    SetAttrD(instance, "density", header.temperature);
    SetAttrD(instance, "obturator", header.obturator);
    instance->nsaturations= header.nsaturations;
    while ((count= gpl_xdr_read_long(buffer, BUFFER_COUNT, stream))>0) {
      if (cp40_import(instance, buffer, count, flags) != GPL_OK) {
	FatalErrorInElib();
      }
    }
    CloseOutput(&instance);
  }
  fclose(stream);
}

/*
#------------------------------------------------------------------------------
#
# Description:
#	Read raw CP40 data (version 2.1, 512 bytes) header organized as follow:
#
#       offset   data  number  contents                      field
#	------  -----  ------  ----------------------------  -----------
#	     0  Int32       1  identifier (0xDEADC0DE)
#	     4  Int32       1  contents identifier (C_DATA=0)
#	     8  DATE        1  date of creation              date  
#	    16  DATE        1  date of update
#	    24  Int32       1  address of data buffer
#	    28  Int32       1  size of data buffer
#	    32  Int32       1  type of data
#	    36  Int32       1  type of observation (0=SPECKLE, 1=SPECKLE_DIF,
#	                       2=DECONV, 3=SFM, 4=GI2T)      experiment
#	    40  Int32       1  xsize
#	    44  Int32       1  ysize
#	    48  Int32       1  number of frames              nframes
#	    52  Int32       1  number of saturations         nsaturations
#	    54  Int32       1  number of photons             nphotons
#	    60  Int32       1  pre-processing flags
#	    64  byte      164  unused
#	   228  char       80  object / comments             object
#	   308  DATE        1  date (again!)
#	   316  real32      1  lambda_1       |	s            lambda_1
#	   320  real32      1  lambda_2	      | p	     lambda_2
#	   324  real32      1  delta_1	      |	e            delta_1
#	   328  real32      1  delta_2	      |	c            delta_2
#	   332  real32      1  density	      |	k            density           
#	   336  real32      1  rotation	      |	l            rotation
#	   340  real32      1  focal length   |	e            focus
#	   344  real32      1  temperature    |	             temperature
#	   348  real32      1  voltage_1 (kV) |	o            voltage_1
#	   352  real32      1  voltage_2      |	n            voltage_2
#	   356  real32      1  obturator      |	l            obturator
#	   360  byte      152  unused         | y
#	------  -----  ------  ----------------------------  -----------
#	   512	Int32  262144  photons
#	------  -----  ------  ----------------------------  -----------
#
# Notes:
#	(1)	The *FIRST* photon of a frame is flagged (i.e. positive).
#	(2)	In addition, the total data file size is stored in
#		header(size) and the number of photons header(photons) is
#		fixed.
#
# Return:
#	1 if read succeed, 0 if there are no more data.
#
# Side effects:
#	Array headerName is created/updated.  The file access pointer is
#	moved the begining of raw data after the header of the recordNumber
#	data record.
#
#------------------------------------------------------------------------------
*/

struct Header2p1 {
  gpl_xdr_int_t identifier;    /* identifier (0xDEADC0DE) */
  gpl_xdr_int_t contents;      /* contents identifier (C_DATA=0) */
  gpl_xdr_uchar_t creation[8];   /* date of creation */
  gpl_xdr_uchar_t update[8];     /* date of update */
  gpl_xdr_int_t address;       /* address of data buffer */
  gpl_xdr_int_t size;          /* size of data buffer */
  gpl_xdr_int_t data_type;     /* type of data */
  gpl_xdr_int_t obs_type;      /* type of observation (0=SPECKLE, 1=SPECKLE_DIF,
				2=DECONV, 3=SFM, 4=GI2T) */
  gpl_xdr_int_t xsize;
  gpl_xdr_int_t ysize;
  gpl_xdr_int_t nframes;       /* number of frames */
  gpl_xdr_int_t nsaturations;  /* number of saturations */
  gpl_xdr_int_t nphotons;      /* number of photons */
  gpl_xdr_int_t flags;         /* pre-processing flags */

  gpl_xdr_char_t comments[81];  /* object / comments */
  gpl_xdr_uchar_t date[8];       /* date (again!) */
  gpl_xdr_float_t lambda_1;
  gpl_xdr_float_t lambda_2;
  gpl_xdr_float_t delta_1;
  gpl_xdr_float_t delta_2;
  gpl_xdr_float_t density;
  gpl_xdr_float_t rotation;
  gpl_xdr_float_t focus;
  gpl_xdr_float_t temperature;
  gpl_xdr_float_t voltage_1;
  gpl_xdr_float_t voltage_2;
  gpl_xdr_float_t obturator;
};

static void ImportVersion2p1(const char *name, gpl_xdr_uint_t buffer[],
			     int flags)
{
  struct Header2p1 header;
  FILE *stream= OpenInput(name);
  long offset= 0;

    
  /* Read header. */
  read_i(stream, (offset+  0L), &header.identifier,   1);
  read_i(stream, (offset+  4L), &header.contents,     1);
  read_b(stream, (offset+  8L),  header.creation,     8);
  read_b(stream, (offset+ 16L),  header.update,       8);
  read_i(stream, (offset+ 24L), &header.address,      1);
  read_i(stream, (offset+ 28L), &header.size,         1);
  read_i(stream, (offset+ 32L), &header.data_type,    1);
  read_i(stream, (offset+ 36L), &header.obs_type,     1);
  read_i(stream, (offset+ 40L), &header.xsize,        1);
  read_i(stream, (offset+ 44L), &header.ysize,        1);
  read_i(stream, (offset+ 48L), &header.nframes,      1);
  read_i(stream, (offset+ 52L), &header.nsaturations, 1);
  read_i(stream, (offset+ 54L), &header.nphotons,     1);
  read_i(stream, (offset+ 60L), &header.flags,        1);
  read_c(stream, (offset+228L),  header.comments,    80);
  header.comments[80]= 0;
  read_b(stream, (offset+308L),  header.date,         8);
  read_f(stream, (offset+316L), &header.lambda_1,     1);
  read_f(stream, (offset+320L), &header.lambda_2,     1);
  read_f(stream, (offset+324L), &header.delta_1,      1);
  read_f(stream, (offset+328L), &header.delta_2,      1);
  read_f(stream, (offset+332L), &header.density,      1);
  read_f(stream, (offset+336L), &header.rotation,     1);
  read_f(stream, (offset+340L), &header.focus,        1);
  read_f(stream, (offset+344L), &header.temperature,  1);
  read_f(stream, (offset+348L), &header.voltage_1,    1);
  read_f(stream, (offset+352L), &header.voltage_2,    1);
  read_f(stream, (offset+356L), &header.obturator,    1);

  if (buffer==NULL) {
    /* Only list contents. */
    printf("*** %s (v2.0) ***\n", name);
    printf("  indentifier ...... 0x%x\n", (unsigned)header.identifier);
    printf("  contents.. ....... %d\n", header.contents);
    printf("  creation ......... %s\n", time_stamp(header.creation));
    printf("  last update ...... %s\n", time_stamp(header.update));
    printf("  address (?) ...... 0x%x\n", (unsigned)header.address);
    printf("  size ............. %d\n", header.size);
    printf("  data type ........ %d\n", header.data_type);
    printf("  observation type . %d\n", header.obs_type);
    printf("  xsize yxsize ..... %d x %d\n", header.xsize, header.ysize);
    printf("  nframes .......... %d\n", header.nframes);
    printf("  nsaturations ..... %d\n", header.nsaturations);
    printf("  nphotons ......... %d\n", header.nphotons);
    printf("  flags ............ %d\n", header.flags);
    printf("  comments ......... %s\n", header.comments);
    printf("  date ............. %s\n", time_stamp(header.date));
    printf("  lambda_1 ......... %g\n", header.lambda_1);
    printf("  lambda_2 ......... %g\n", header.lambda_2);
    printf("  delta_1 .......... %g\n", header.delta_1);
    printf("  delta_2 .......... %g\n", header.delta_2);
    printf("  density .......... %g\n", header.density);
    printf("  rotation ......... %g\n", header.rotation);
    printf("  focus ............ %g\n", header.focus);
    printf("  temperature ...... %g\n", header.temperature);
    printf("  voltage_1 ........ %g\n", header.voltage_1);
    printf("  voltage_2 ........ %g\n", header.voltage_2);
    printf("  obturator ........ %g\n", header.obturator);
  } else {
    /* Import data file. */
    char defaultPath[PATH_MAX+1];
    cp40_instance_t *instance= NULL;
    const char *path;
    int count;

    if (fseek(stream, offset+512, SEEK_SET) != 0) {
      gpl_sys_error("unable to seek to data part of file");
      exit(1);
    }
#if 0
    if (header.identifier != 0xDEADC0DE || header.data_type != 0) {
      fprintf(stderr, "bad header (v2.0)\n");
      exit(1);
    }
#endif
    strcpy(defaultPath, name);
    strcat(defaultPath, ".cp40");
    path= GetPath("Name of new output data file", defaultPath, FILE_NEW);
    OpenOutput(&instance, path);
	
    while ((count= gpl_xdr_read_long(buffer, BUFFER_COUNT, stream))>0) {
      if (cp40_import(instance, buffer, count, flags) != GPL_OK) {
	FatalErrorInElib();
      }
    }
    CloseOutput(&instance);
  }
  fclose(stream);
}

/*---------------------------------------------------------------------------*/

static void ImportVersion3p0(const char *name, gpl_xdr_uint_t buffer[],
			     int flags)
{
  FILE *stream= OpenInput(name);
  char defaultPath[PATH_MAX+1];
  cp40_instance_t *instance= NULL;
  const char *path;
  int count;

  if (fseek(stream, 16*KILO_BYTE, SEEK_SET)) {
    FatalError("unable to skip 16 kb (empty) header");
  }
  strcpy(defaultPath, name);
  strcat(defaultPath, ".cp40");
  path= GetPath("Name of new output data file", defaultPath, FILE_NEW);
  OpenOutput(&instance, path);
  while ((count= gpl_xdr_read_long(buffer, BUFFER_COUNT, stream))>0) {
    if (cp40_import(instance, buffer, count, flags) != GPL_OK) {
      FatalErrorInElib();
    }
  }
  CloseOutput(&instance);
  return;
}

static void ImportVersion3p1(const char *name, gpl_xdr_uint_t buffer[],
			     int flags)
{
  FILE *stream= OpenInput(name);
  char defaultPath[PATH_MAX+1];
  cp40_instance_t *instance= NULL;
  const char *path;
  int count;

  /* Make output file header from input file header (just add identifier
     string, then overwrite output file header, re-read header and fix scale),
     this is messy but should work... */ 
  if (fread(buffer, KILO_BYTE, 16, stream) != 16) {
    FatalError("unable to read input header");
  }
  strcpy((char *)buffer, "CP40 DATA 1");
  strcpy(defaultPath, name);
  strcat(defaultPath, ".cp40");
  path= GetPath("Name of new output data file", defaultPath, FILE_NEW);
  OpenOutput(&instance, path);
  if (fseek(instance->stream.io.file.handle, 0, SEEK_SET) != 0 ||
      fwrite(buffer, KILO_BYTE, 16, instance->stream.io.file.handle) != 16) {
    FatalError("unable to overwrite output header");
  }
  if (cp40_read_header(instance) != GPL_OK ||
      (cp40_get_scale(instance) <= 0.0 &&
       cp40_set_scale(instance, 1.0) != GPL_OK)) {
    FatalErrorInElib();
  }

  /* Import data. */
  while ((count= gpl_xdr_read_long(buffer, BUFFER_COUNT, stream))>0) {
    if (cp40_import(instance, buffer, count, flags) != GPL_OK) {
      FatalErrorInElib();
    }
  }
  CloseOutput(&instance);
  return;
}

/*---------------------------------------------------------------------------*/

#if 0
static long ImportData(cp40_instance_t *instance, FILE *stream, long count,
		       int flags);
/*	Import raw photon counting data from file STREAM to instance DATAPTR.
	COUNT is the number of photons to import (<0 means up to the end of
	the file) and FLAGS is the same as in cp40_import().  The number
	of actually imported photons is returned (-1 in case of error). */
	

static void ImportData(cp40_instance_t *instance, FILE *stream,
		       long count, int flags)
{
  int nread, rest, nbuf= (1024*1024/4);
  gpl_xdr_uint_t *buffer;
    
  if (count==0) return 0;
  if (count>0 && count<nbuf) {
    nbuf= count;
  }
  if ((buffer= malloc(4*nbuf))==NULL) {
    gpl_sys_error("cannot allocate raw-photon buffer");
    return -1;
  }

  rest= count;
  count= 0;
  while (rest!=0) {
    if (rest>0 && nbuf>rest) nbuf= rest;
    nread= gpl_xdr_read_long(buffer, nbuf, stream);
    if (nread==0) break;
    if (nread<0 || cp40_import(instance, buffer, nread, flags)!=GPL_OK) {
      free(buffer);
      return -1;
    }
    count+= nread;
    if (rest>0) rest-= nread;
  }
  free(buffer);
  return count;
}
#endif

/*---------------------------------------------------------------------------*/

#undef HAVE_MODE
#undef HAVE_TYPE
#define HAVE_MODE(s, f)  (((s).st_mode&(f))==(f))
#define HAVE_TYPE(s, f)  (((s).st_mode&S_IFMT)==(f))

static const char *GetPath(const char *message, const char *defaultPath,
			   int flags)
{
  static char path[PATH_MAX+1];
  struct stat info;

 ask:
  fputs(message, stdout);
  if (defaultPath) {
    fputs(" [", stdout);
    fputs(defaultPath, stdout);
    putc(']', stdout);
  }
  fputs(": ", stdout);
  fflush(stdout);
  fgets(path, PATH_MAX, stdin);
  path[PATH_MAX]= 0;
  StripString(path, path); /* (this also remove the last newline) */
  if (path[0]==0) {
    if (defaultPath) return defaultPath;
    goto ask;
  }
  if (stat(path, &info)) {
    if (errno==EBADF) {
      Warning("invalid path name");
    } else if (errno==ENOENT) {
      if (flags&FILE_NEW) return path;
      Warning("no such file or directory");
    } else {
      Warning("unable to stat file");
    }
    goto ask;
  }
  if (flags&FILE_NEW) {
    Warning("file already exists");
    goto ask;
  }
  return path;
}

/*---------------------------------------------------------------------------*/
#define Y2K
static const char *time_stamp(unsigned char date[])
{
  static char stamp[30];
#ifdef Y2K
  int year= date[0];
  if (year>=89) {
    year+= 1900;
  } else {
    if (year>=20) {
      fprintf(stderr,
	      "WARNING (%s:%d): you should fix the coding of the year\n",
	      __FILE__, __LINE__);
    }
    year+= 2000;
  }
  sprintf(stamp, "%02d/%02d/%d %02d:%02d:%02d", date[2], date[1], year,
	  date[3], date[4], date[5]);
#else
  sprintf(stamp, "%02d/%02d/%02d %02d:%02d:%02d", date[2], date[1], date[0],
	  date[3], date[4], date[5]);
#endif
  return stamp;
}

/*---------------------------------------------------------------------------*/

static void CloseOutput(cp40_instance_t **instance)
{
  if (instance[0] != NULL && cp40_delete_instance(instance[0], 1) != GPL_OK) {
    FatalErrorInElib();
  }
  instance[0]= NULL;
}

static void OpenOutput(cp40_instance_t **instance, const char *path)
{
  CloseOutput(instance);
  if ((instance[0]= cp40_new_instance()) == NULL
      || cp40_assoc_file(instance[0], path, CP40_IO_MODE_WRITE)!=GPL_OK) {
    FatalErrorInElib();
  }
}

/*---------------------------------------------------------------------------*/

static void StripString(char *dst, register const char *src)
{
  register char c, *ptr= dst;
  register int flag= 0; /* Initialized to `false' to avoid copying leading
			   spaces. */
    
  for (;;) {
    switch ((c= *src++)) {
    case ' ':
    case '\t':
    case '\n':
    case '\v':
    case '\f':
    case '\r':
      if (flag) {
	*ptr++= ' ';
	flag= 0;
      }
      break;
    case '\0':
      /* Write terminating null and return. */
      ptr[ptr>dst && ptr[-1]==' ' ? -1 : 0]= '\0';
      return;
    default:
      *ptr++= c;
      flag= 1;
    }
  }
}

/*---------------------------------------------------------------------------*/

static void Warning(const char *message)
{
  fprintf(stderr, "*** WARNING: %s ***\n", message);
}

static void FatalError(const char *message)
{
  fprintf(stderr, "*** ERROR: %s ***\n", message);
  exit(1);
}

static void FatalErrorInElib(void)
{
  fprintf(stderr, "ELIB ERROR: %s\n", gpl_error_info(0));
  exit(1);
}


static int YesOrNo(const char *question, const char *help_str, int def)
{
  gpl_dstr_t dynStr;
  const char *ptr;
  int result= 0;
    
  gpl_dstr_init(&dynStr);
  for (;;) {
    fputs(question, stdout);
    fputs(" [", stdout);
    fputs((def==1 ? "Y/n" : (def==0 ? "y/N" : "y/n")), stdout);
    fputs((help_str==NULL ? "] " : "/?] "), stdout);
    fflush(stdout);
    if (gpl_dstr_read_line(&dynStr, stdin, NULL, NULL) != GPL_OK) {
      FatalErrorInElib();
    }
    gpl_dstr_trim(&dynStr, GPL_STR_TRIM_BOTH);
    gpl_dstr_lower(&dynStr);
    ptr= gpl_dstr_value(&dynStr);
    if (ptr[0]== 'y') {
      if (ptr[1]==0 || (ptr[1]=='e' && ptr[2]=='s' && ptr[3]==0)) {
	result= 1;
	break;
      }
    } else if (ptr[0]=='n') {
      if (ptr[1]==0 || (ptr[1]=='o' && ptr[2]==0)) {
	result= 0;
	break;
      }
    } else if (ptr[0]==0) {
      if (def==0 || def==1) {
	result= def;
	break;
      }
    } else if (ptr[0]=='?') {
      if (ptr[1]==0 && help_str!=NULL) {
	fputs(help_str, stdout);
	fflush(stdout);
	continue;
      }
    }
    fputs("*** ambiguous answer ***\n", stderr);
  }

  gpl_dstr_free(&dynStr);
  return result;
}
/*---------------------------------------------------------------------------*/
