/*
 * cp40cmd_integ.c --
 *
 *	Source code for command to integrate CP40 estimators.
 *
 * Copyright and warranty:
 *	Copyright (c) 1998 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Log$
 *_____________________________________________________________________________
 */
static const char *rcs_version= "$Id$";


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "cp40.h"
#include "gpl_fits.h"
#include "gpl_xdr.h"
#include "gpl_option.h"
#include "gpl_string.h"
/*---------------------------------------------------------------------------*/
static const char *prog_name= NULL;

static void fatal_error(const char *message, int call_abort)
						     __attribute__((noreturn));
static void fatal_error(const char *message, int call_abort)
{
  if (! message) message= gpl_error_info(1);
  if (prog_name) {
    fputs(prog_name, stderr);
    fputs(": ", stderr);
  }
  fputs(message, stderr);
  fputc('\n', stderr);
  if (call_abort) abort();
  exit(1);
}

static char *output_name(const char *prefix, cp40_task_t *task);
static char *output_name(const char *prefix, cp40_task_t *task)
{
  char *name= NULL;
  if (! (name= gpl_alloc(strlen(prefix)+strlen(task->token)+7))) {
    gpl_error("insufficient memory for output file name");
    return NULL;
  }
  sprintf(name, "%s-%s.fits", prefix, task->token);
  return name;
}

static int save_task(const char *prefix, cp40_task_t *task);
static int save_task(const char *prefix, cp40_task_t *task)
{
  int result= GPL_OK;
  char *name= NULL;
  FILE *file= NULL;

  if (! (name= output_name(prefix, task))) {
    result= GPL_ERROR;
    goto done;
  }
  sprintf(name, "%s-%s.fits", prefix, task->token);
  if (! (file= fopen(name, "w"))) {
    result= gpl_error("error opening output file");
    goto done;
  }
  result= tim_fits_write_image(file, task->image);
 done:
  if (name) gpl_free(name);
  if (file) fclose(file);
  return result;
}

extern char *gpl_str_basename(const char *string, const char *suffix);
char *gpl_str_basename(const char *string, const char *suffix)
{
  int result_len, suffix_len, string_len;
  char *ptr;

  if (! suffix) {
    result_len= (ptr= strrchr(string, '.')) ? ptr - string : strlen(string);
  } else {
    string_len= strlen(string);
    if ((suffix_len= strlen(suffix)) <= 0
	|| (result_len= string_len - suffix_len) < 0
	|| strcmp(suffix, string+result_len) != 0) {
      result_len= string_len;
    }
  }
  if (! (ptr= gpl_alloc(result_len+1))) {
    gpl_error("insufficient memory for basename string");
    return NULL;
  }
  memcpy(ptr, string, result_len);
  ptr[result_len]= 0;
  return ptr;
}

/*---------------------------------------------------------------------------*/

static const char *help= "\n\
Run CP40 task(s) and save result in FITS file(s).\n\
\n\
Usage: cp40integ [OPTIONS] TASK [...] SRC\n\
\n\
  TASK [...]              tasks to perform\n\
  SRC                     CP40 source file\n\
\n\
Options:\n\
  -a, --attribute=FILE    use pre-processing attributes from FILE\n\
  -p, --prefix=PREFIX     prefix for result images: \"PREFIX-TASK.fits\";\n\
                          default is SRC without \".cp40\" extension\n\
  -?, --help              print this short help and exit\n\
  --version               print version and exit\n\n";

int main(int argc, const char *argv[])
{
  static gpl_opt_def_t opt[]= {
    {"attribute", "a",   GPL_OPT_ARG_REQUIRED,   'a'},
    {"prefix",    "p",   GPL_OPT_ARG_REQUIRED,   'p'},
    {"version",   NULL,  GPL_OPT_ARG_NONE,       '~'},
    {"help",      "?",   GPL_OPT_ARG_NONE,       '?'},
    {NULL,        NULL,  0,                     0},
  };
  gpl_opt_data_t opt_data;
  int i, ntasks, nframes, update;
  const char *prefix= NULL, *attribute= NULL;
  const char *src= argv[argc-1], **token;
  cp40_instance_t *instance;
  
  /* Parse options. */
  prog_name= argv[0];
  if (argc<=1) {
    fputs(help, stderr);
    exit(1);
  }
  gpl_opt_init(&opt_data, argc, argv, opt, GPL_OPT_PERMUTE_ARGS);    
  while ((i= gpl_opt_next(&opt_data))!=GPL_OPT_END_OF_OPTIONS) {
    if (i<0) fatal_error(NULL, 0);
    switch (opt[i].value) {
    case 'a':
      if (attribute!=NULL) fatal_error("too many attribute files", 0);
      if (! (attribute= gpl_str_clone(opt_data.opt_arg))) {
	fatal_error("insufficient memory for attribute", 0);
      }
      break;
    case 'p':
      if (prefix!=NULL) fatal_error("too many prefixes", 0);
      if (! (prefix= gpl_str_clone(opt_data.opt_arg))) {
      alloc_prefix_error:
	fatal_error("insufficient memory for prefix", 0);
      }
      break;
    case '~':
      fprintf(stderr, "%s: %s\n", argv[0], rcs_version);
      exit(0);
      break;
    case '?':
    default:
      fputs(help, stderr);
      exit(0);
      break;
    }
  }

  ntasks= argc - opt_data.arg_index - 1;
  token= argv + opt_data.arg_index;
  if (ntasks <= 0) {
    fatal_error("no tasks to perform", 0);
  }
  if (! prefix && ! (prefix= gpl_str_basename(src, ".cp40"))) {
    goto alloc_prefix_error;
  }

  /* Prepare data instance. */
  if (! (instance= cp40_new_instance())
      || cp40_assoc_file(instance, src, CP40_IO_MODE_READ)!=GPL_OK) {
    fatal_error(NULL, 0);
  }
  if (attribute && cp40_restore_attr(instance, attribute,
				     CP40_ATTR_PREPROCESSING) != GPL_OK) {
    fatal_error(NULL, 0);
  }
  for (i=0 ; i<ntasks ; i++) {
    if (cp40_schedule_task(instance, token[i]) == NULL) {
      fatal_error(NULL, 0);
    }
  }

  /* Performs all tasks. */
  nframes= cp40_run_tasks(instance, -1);

  /* Save all tasks. */
  while (instance->task) {
    if (save_task(prefix, instance->task) != GPL_OK) {
      fatal_error(NULL, 0);
    }
    cp40_cancel_task(instance->task);
  }
  update= 0;
  cp40_delete_instance(instance, update);
  GPL_FREE_MAYBE((void *)prefix);
  return 0;
}

/*---------------------------------------------------------------------------*/
