/*
 * cp40correction.i --
 *
 *	Fit distortion correction from images of a grid.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id: cp40correction.i,v 1.1 2000/07/22 22:29:12 spid Exp spid $
 *	$Log: cp40correction.i,v $
 *	Revision 1.1  2000/07/22 22:29:12  spid
 *	Initial revision
 *
 */

require, "fits.i";
require, "Eric/poly.i";

local DC_DEBUG;
/* DOCUMENT local DC_DEBUG;
     Debug mode? (default false). */

local DC_TOLERANCE;
/* DOCUMENT local DC_TOLERANCE;
     Maximum distance in pixels between computed position and location
     of spots.  Default: 3 pixels. */

local DC_THRESHOLD;
/* DOCUMENT local DC_THRESHOLD;
     Total flux (minus local background) must be greater than DC_THRESHOLD
     times the local background times the area for the spot to be valid.
     Default: 0.5 */

local DC_BOX_SIZE;
/* DOCUMENT local DC_BOX_SIZE;
     Box size (in distorted pixels) for computing the spot centers.
     Default: 5 pixels. */

local DC_MAXITER;
/* DOCUMENT local DC_MAXITER;
     Maximum number of iterations to solve backward correction and to
     improve distortion coefficients.  Default: 10 iterations. */

local DC_PRECISION;
/* DOCUMENT local DC_PRECISION;
     Precision in grid step for convergence. Default: 10e-4. */

local DC_SCALE;
/* DOCUMENT local DC_SCALE;
     Approximate scale factor to convert distances in undistorted space
     into distorted space. Default: 24.0 distorted pixels per grid step. */

local DC_CHANNEL;
/* DOCUMENT local DC_CHANNEL;
     Current CP40 channel. */

local DC_IMAGE;
/* DOCUMENT local DC_IMAGE;
     Current CP40 distorted grid image. */

local DC_CUBE;
/* DOCUMENT local DC_CUBE;
     Cube of CP40 distorted grid images. */

local DC_W, DC_X, DC_Y, DC_XP, DC_YP;
/* DOCUMENT local DC_W, DC_X, DC_Y, DC_XP, DC_YP;
     Node weights (unused if <=0), distorted and undistorted coordinates in
     current grid image. */

local DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST;
/* DOCUMENT local DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST;
     Indices of next node in a given direction (-1 if irrelevant). */ 

local DC_FIT_WORSE;
/* DOCUMENT local DC_FIT_WORSE;
     Index of worst node with respect to fit. */

struct DC_DATA_S {
  pointer w, x, y, xp, yp;
  pointer coef;
  pointer north, south, west, east;
};

local DC_DATA;
/* DOCUMENT local DC_DATA;
     Array of structures to keep per-channel data. */

func dc_init(cube, box_size=, threshold=, precision=, scale=, tolerance=,
	     maxiter=, debug=)
/* DOCUMENT dc_init, cube;
     Start fitting of distortion correction. */
{
  extern DC_CUBE, DC_DATA, DC_CHANNEL, DC_TOLERANCE;
  extern DC_BOX_SIZE, DC_THRESHOLD, DC_MAXITER, DC_PRECISION, DC_SCALE;
  if ((dims= dimsof(cube))(1)!=3 || dims(2)!=384 || dims(3)!=288
      || dims(4)!=4) error, "expecting a (384,288,4) data cube";
  if (! is_void(debug)) {
    DC_DEBUG= !(!debug);
  } else if (is_void(DC_DEBUG)) {
    DC_DEBUG= 0n;
  }
  if (! is_void(box_size)) {
    DC_BOX_SIZE= box_size;
  } else if (is_void(DC_BOX_SIZE)) {
    DC_BOX_SIZE= 5;
  }
  if (! is_void(threshold)) {
    DC_THRESHOLD= threshold;
  } else if (is_void(DC_THRESHOLD)) {
    DC_THRESHOLD= 0.5;
  }
  if (! is_void(tolerance)) {
    DC_TOLERANCE= tolerance;
  } else if (is_void(DC_TOLERANCE)) {
    DC_TOLERANCE= 5.0;
  }
  if (! is_void(precision)) {
    DC_PRECISION= precision;
  } else if (is_void(DC_PRECISION)) {
    DC_PRECISION= 1e-4;
  }
  if (! is_void(maxiter)) {
    DC_MAXITER= maxiter;
  } else if (is_void(DC_MAXITER)) {
    DC_MAXITER= 10;
  }
  if (! is_void(scale)) {
    DC_SCALE= scale;
  } else if (is_void(DC_SCALE)) {
    DC_SCALE= 24.0;
  }
  
  /* Instanciate global per-channel ressources. */
  DC_DATA= array(DC_DATA_S, 4);
  DC_CUBE= cube;
  DC_CHANNEL= -1;
  dc_change, 1;
  dc_help;
  dc_wait_event;
}

func dc_help(void)
{
  write, "Use the following mouse/key:";
  write, " MOUSE  KEY  ACTION"
  write, "   1     A   select/toggle grid node";
  write, "   2     D   ";
  write, "   3     X   exit";
  write, "         I   improve coefficients";
  write, "         N   next channel";
  write, "         P   previous channel";
  write, "         Q   save coefficients and exit";
  write, "         R   reset settings for current channel";
  write, "         S   save coefficients in \"cp40correction.config\"";
  write, "         W   delete worst node";
  write, "        H ?  print this help";
  write, "        0-3  change to that channel number";
}

func dc_save_coefs(file)
{
  extern DC_DATA, DC_COEF, DC_SCALE;
  number= 10;
  s= "";
  for (channel=1 ; channel<=4 ; channel++) {
    coef= *DC_DATA(channel).coef;
    if (numberof(coef)!=4*number) {
      dc_error, swrite(format="bad coefficients for channel %d", channel);
      pause, 1000;
      return;
    }
    q= coef(,1:2);
    nq= numberof(q);
    /* Rotate coefficients appropriately. */
    if (channel==1 || channel==2) {
      q(,2)*= -1;
    } else {
      q(,1)*= -1;
    }
    for (i=1 ; i<=nq ; i++) s+= swrite(format=" % .15e", q(i));
  }
  if (is_void(file)) file= "cp40correction.config";
  file= open(file, "w");
  write, file, format="# %s\n", "CP40 distortion correction settings.";
  write, file, format="# -- created on: %s\n", timestamp();
  write, file, format="%20s = %s\n", "correction", "random";
  write, file, format="%20s = %.1f\n", "scale", DC_SCALE;
  write, file, format="%20s = %.1f\n", "angle", 0.0;
  write, file, format="%20s = %s\n", "raw_coefs", s;
}

func dc_wait_event(void)
{
  extern DC_CHANNEL, DC_FIT_WORST, DC_W;

  for (;;) {
    cursor= dc_get_cursor();
    key= cursor.key;
    if (key>='a' && key<='z') key-= 'a'-'A';
    if (key=='A') {
      dc_add_or_toggle, cursor.x, cursor.y;
    } else if (key=='N') {
      dc_change, (DC_CHANNEL>=4 ? 1 : DC_CHANNEL+1);
    } else if (key=='P') {
      dc_change, (DC_CHANNEL<=1 ? 4 : DC_CHANNEL-1);
    } else if (key=='Q') {
      dc_save_coefs;
      break;
    } else if (key=='R') {
      dc_reset;
    } else if (key=='S') {
      dc_save_coefs;
    } else if (key=='0') {
      dc_change, 1;
    } else if (key=='1') {
      dc_change, 2;
    } else if (key=='2') {
      dc_change, 3;
    } else if (key=='3') {
      dc_change, 4;
    } else if (key=='W') {
      /* Delete worst point. */
      if (! is_void(DC_FIT_WORST) && DC_FIT_WORST <= numberof(DC_W)) {
	DC_W(DC_FIT_WORST)= -abs(DC_W(DC_FIT_WORST));
	dc_update;
      }
    } else if (key=='H' || key=='?') {
      dc_help;
    } else if (key=='X') {
      break;
    }
  } 
}

func dc_change(channel)
/* DOCUMENT dc_change, channel;
     Save current per-channel ressources (see dc_save) and make CHANNEL the
     current one (see dc_restore). */
{
  extern DC_CHANNEL;
  dc_save;
  if (channel==DC_CHANNEL || channel<1 || channel>4) return;
  dc_restore, channel;
}

func dc_save(void)
/* DOCUMENT dc_save;
     Save current per-channel ressources into DC_DATA(DC_CHANNEL). */
{
  extern DC_COEF, DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  extern DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST, DC_CHANNEL;
  if (is_void(DC_CHANNEL)) DC_CHANNEL= -1;
  if (1<=DC_CHANNEL && DC_CHANNEL<=4) {
    /* Save last ressources for this channel. */
    DC_DATA(DC_CHANNEL)= DC_DATA_S(w=&DC_W, x=&DC_X, y=&DC_Y,
				   xp= &DC_XP, yp= &DC_YP, coef=&DC_COEF,
				   north=&DC_NORTH, south=&DC_SOUTH,
				   west= &DC_WEST, east=&DC_EAST);
  }
}

func dc_restore(channel)
/* DOCUMENT dc_restore, channel;
     Restore per-channel ressources from DC_DATA(CHANNEL) and redraw
     graphics.   Doe nothing if CHANNEL is invalid. */
{
  extern DC_CHANNEL, DC_IMAGE, DC_CUBE;
  extern DC_COEF, DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  extern DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST;
  
  if (channel<1 || channel>4) return;
  DC_W= *DC_DATA(channel).w;
  DC_X= *DC_DATA(channel).x;
  DC_Y= *DC_DATA(channel).y;
  DC_XP= *DC_DATA(channel).xp;
  DC_YP= *DC_DATA(channel).yp;
  DC_COEF= *DC_DATA(channel).coef;
  DC_NORTH= *DC_DATA(channel).north;
  DC_SOUTH= *DC_DATA(channel).south;
  DC_WEST= *DC_DATA(channel).west;
  DC_EAST= *DC_DATA(channel).east;
  DC_IMAGE= DC_CUBE(.., channel);
  DC_CHANNEL= channel;
  dc_redraw_all;
}

func dc_reset(void)
/* DOCUMENT dc_reset;
     Reset current per-channel ressources. */
{
  extern DC_COEF, DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  extern DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST, DC_CHANNEL;
  if (is_void(DC_CHANNEL) || DC_CHANNEL<1 || DC_CHANNEL>4) return;
  DC_W= DC_X= DC_Y= DC_XP= DC_YP= DC_COEF= [];
  DC_NORTH= DC_SOUTH= DC_WEST= DC_EAST= [];
  dc_redraw_all;
}

func dc_toggle(x, y, state, undistorted)
/* DOCUMENT dc_toggle, x, y;
       -or- dc_toggle, x, y, state;
       -or- dc_toggle, x, y, state, 1;
     Toggle the state of the node closest to the position (X,Y) in distorted
     image or in undistorted grid if optional 4th argument is true (i.e.
     non-nil and non-zero). Does nothing if (X,Y) is too far from any actual
     nodes.  If STATE is nil or zero, the node state is reversed.  If STATE
     is >0, the node will be accounted in the fit of distortion correction.
     If STATE is <0, the node will be omitted in the fit.

   SEE ALSO: dc_closest, dc_update. */
{
  extern DC_W;
  if ((i= dc_closest(x, y, undistorted)) > 0
      && (! state || state*DC_W(i) < 0.0)) {
    DC_W(i)= -DC_W(i);
    dc_update;
  }
}

func dc_add_or_toggle(x, y, undistorted)
/* DOCUMENT dc_add_or_toggle, x, y;
       -or- dc_add_or_toggle, x, y, 1;
 */
{
  extern DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  extern DC_BOX_X, DC_BOX_Y, DC_BOX_Z, DC_BOX_SIZE;
  extern DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST, DC_CHANNEL;
  
  /* Actually cannot add more than 2 nodes. */
  number= numberof(DC_W);
  if (number>=2) {
    dc_toggle, x, y, undistorted;
    return;
  }

  /* Localize new spot into grid image. */
  wxy= dc_localize(x, y);
  w= wxy(1);
  x= wxy(2);
  y= wxy(3);
  if (w<=0.0) return;
  if (dc_closest(x, y, undistorted) > 0) return;
  
  /* Figure out the new node undistorted position. */
  if (number==0) {
    xp= yp= 0;
  } else if (number==1) {
    dx= x - DC_X(1);
    dy= y - DC_Y(1);
    if (abs(dx) >= abs(dy)) {
      xp= sign(dx);
      yp= 0;
    } else {
      xp= 0;
      yp= sign(dy);
    }
  }

  for (;;) {
    /* Append new nodes. */
    grow, DC_W, w;
    grow, DC_X, x;
    grow, DC_Y, y;
    grow, DC_XP, xp;
    grow, DC_YP, yp;

    /* Because Yorick's arrays are dynamically typed, make sure that DC_XP
       and DC_YP are really integer arrays (have no fractionnal part would
       be sufficient in fact) by trucating them to the nearest integer. */
    DC_XP= long(floor(DC_XP+0.5));
    DC_YP= long(floor(DC_YP+0.5));

    /* Rebuild the connections (there is certainly a more clever and
       efficient way to just update the connection instead, but
       completely rebuilding the connections is much simpler...). */
    number= numberof(DC_W);
    DC_NORTH= DC_EAST= DC_SOUTH= DC_WEST= array(-1, number);
    if (number>=2) {
      for (i=1 ; i<=number ; i++) {
	if (is_array((j= where(DC_XP==DC_XP(i))))) {
	  if (is_array((k= where(DC_YP(j)==DC_YP(i)+1)))) {
	    k= j(k); DC_NORTH(i)= k; DC_SOUTH(k)= i;
	  }
	  if (is_array((k= where(DC_YP(j)==DC_YP(i)-1)))) {
	    k= j(k); DC_SOUTH(i)= k; DC_NORTH(k)= i;
	  }
	}
	if (is_array((j= where(DC_YP==DC_YP(i))))) {
	  if (is_array((k= where(DC_XP(j)==DC_XP(i)+1)))) {
	    k= j(k); DC_EAST(i)= k; DC_WEST(k)= i;
	  }
	  if (is_array((k= where(DC_XP(j)==DC_XP(i)-1)))) {
	    k= j(k); DC_WEST(i)= k; DC_EAST(k)= i;
	  }
	}
      }
    }

    /* Re-compute the coefficients and redraw the nodes. */
    if (! dc_update()) break;  /* stop because no distortion correction */

    /* Extrapolate backward correction to found more nodes which must
       not ne too close from any other existing nodes. */
    xp1= min(DC_XP)-1; xp2= max(DC_XP)+1; xp3= indgen(xp1:xp2);
    yp1= min(DC_YP)-1; yp2= max(DC_YP)+1; yp3= indgen(yp1:yp2)(2:-1);
    xn= numberof(xp3);
    yn= numberof(yp3);
    xp= yp= array(long, 2*(xn+yn));
    k2= 0;
    k1= k2+1; k2+= xn; xp(k1:k2)= xp3; yp(k1:k2)= yp1;
    k1= k2+1; k2+= yn; xp(k1:k2)= xp1; yp(k1:k2)= yp3;
    k1= k2+1; k2+= xn; xp(k1:k2)= xp3; yp(k1:k2)= yp2;
    k1= k2+1; k2+= yn; xp(k1:k2)= xp2; yp(k1:k2)= yp3;
    wxy= dc_localize(poly2(xp, yp, DC_COEF(,3)), poly2(xp, yp, DC_COEF(,4)));
    w= wxy(..,1);
    x= wxy(..,2);
    y= wxy(..,3);
    i= where(w>0.0);
    if (! is_array(i)) break; /* no more nodes found */
#if 0
    ip= where((abs(x(i)(-,)-DC_X)(min,) >= DC_BOX_SIZE)
	     +(abs(y(i)(-,)-DC_Y)(min,) >= DC_BOX_SIZE));
    if (! is_array(ip)) break; /* all new nodes are too close to
				  existing ones */
    i= i(ip);
#endif
    x= x(i);
    y= y(i);
    w= w(i);
    xp= xp(i);
    yp= yp(i);
  }
}

func dc_closest(x, y, undistorted)
/* DOCUMENT i= dc_closest(x, y);
       -or- i= dc_closest(x, y, 1);
     Get index of node closest to the position (X,Y) in distorted image
     or in undistorted grid if optional third argument is true (i.e. non-nil
     and non-zero). Returns -1 if (X,Y) is too far from any actual nodes. */
{
  extern DC_W, DC_X, DC_Y, DC_XP, DC_YP;

  if (numberof(DC_W)) {
    if (undistorted) {
      r= abs(x-DC_XP, y-DC_YP);
      rmax= 0.25;
    } else {
      r= abs(x-DC_X, y-DC_Y);
      rmax= DC_BOX_SIZE;
    }
    if (min(r) <= rmax) return r(mnx);
  }
  return -1;
}

func dc_update(degree)
{
  extern DC_SCALE, DC_COEF, DC_W, DC_X, DC_Y, DC_XP, DC_YP, DC_FIT_WORST;
  dc_redraw_nodes;
  degree= dc_fit_coefs(degree);
  if (degree>0) {
    /* For non-linear correction, refine the correction. */
    if (degree>1) dc_improve;

    /* Find the worst spot with respect to the fit. */
    if (is_array((i= where(DC_W>0.0)))) {
      x= DC_X(i); y= DC_Y(i);
      err= DC_SCALE * abs(poly2(x, y, DC_COEF(,1)) - DC_XP(i),
			  poly2(x, y, DC_COEF(,2)) - DC_YP(i));
      dc_draw_message, swrite(format="residuals: %.3f/%.3f pixels (avg/max)",
			      avg(err), max(err)), color="blue";
      DC_FIT_WORST= i(err(mxx));
      __pgsci, 7; /* yellow */
      dc_draw_box, DC_X(DC_FIT_WORST), DC_Y(DC_FIT_WORST), DC_BOX_SIZE;
    }
  }
  return degree;  
}


func dc_localize(x0, y0)
/* DOCUMENT wxy= dc_localize(x0, y0);
     Localize spot(s) in distorted grid image that is (are) close enough
     to (X0,Y0) and return array WXY with one more dimension than X0+Y0:
       WXY(..,1)= weights (zero where input position is not close enough
                           to a grid spot)
       WXY(..,2)= estimated abscissae
       WXY(..,3)= estimated ordinates

   SEE ALSO: dc_box. */
{
  extern DC_TOLERANCE, DC_THRESHOLD;
  extern DC_BOX_X, DC_BOX_Y, DC_BOX_Z, DC_BOX_SIZE;

  /* Make X0 and Y0 same shape. */
  dims= dimsof(x0, y0);
  if (is_void(dims)) error, "X0 and Y0 must be conformable";
  one= array(1.0, dims);
  x0*= one;
  y0*= one;
  number= numberof(x0);
  grow, dims, 3; dims(1)++;
  wxy= array(double, dims);

  /* Localize each spot. */
  for (i=1, j=0 ; i<=number ; i++, j+=3) { 
    /* Compute the barycentre around each starting position (compute
       barycentre twice to limit bias). */
    x= x0(i);
    y= y0(i);
    if (! dc_get_box(x, y, full=1, clip=1)) {
      w= 0.0;
    } else {
      tmp= dc_centroid(DC_BOX_X, DC_BOX_Y, DC_BOX_Z);
      if (tmp(1) > 0.0 && dc_get_box(tmp(2), tmp(3), full=1)) {
	tmp= dc_centroid(DC_BOX_X, DC_BOX_Y, DC_BOX_Z);
	w= tmp(1);
	x= tmp(2);
	y= tmp(3);
#if 1
	w= 1.0;
#endif
	/* Remove point if too far from its initial position. */
	if (abs(x-x0(i)) > DC_TOLERANCE || abs(y-y0(i)) > DC_TOLERANCE) {
	  w= 0.0;
	}
      } else {
	w= 0.0;
      }
    }
    wxy(i)= w;
    wxy(i+number)= x;
    wxy(i+2*number)= y;
  }
  return wxy;
}

func dc_centroid(x, y, z)
{
  extern DC_THRESHOLD;
  bg= array(int, dimsof(z));
  bg(1,)= 1n; bg(,1)= 1n;
  bg(0,)= 1n; bg(,0)= 1n;
  bg= avg(z(where(bg)));
  z-= bg;
  if ((w= sum(z)) <= DC_THRESHOLD * bg * numberof(z)) {
    return [0.0, avg(x), avg(y)];
  }
  return [w, sum(x*z)/w, sum(y*z)/w];
}

func dc_get_box(x, y, full=, clip=)
/* DOCUMENT test= dc_get_box(x, y);

     Return non-zero if a square region of half-widht DC_BOX_SIZE in
     DC_IMAGE around position (X,Y) could be found.  The pixel values
     and coordinates of the region get stored into global arrays DC_BOX_Z,
     DC_BOX_X and DC_BOX_Y.  If keyword FULL is true, all the coordinates
     are stored (instead of N-by-1 and 1-by-N vectors).  To be valid, a
     box must be completely within the image unless keyword CLIP is true
     in which case any non empty box is considered as valid. */
{
  extern DC_IMAGE, DC_BOX_X, DC_BOX_Y, DC_BOX_Z, DC_BOX_SIZE;

  dims= dimsof(DC_IMAGE);
  width= dims(2);
  height= dims(3);

  ximg= indgen(0:width-1);
  yimg= indgen(0:height-1);
  x= long(floor(x+0.5)); /* X rounded to nearest integer */
  y= long(floor(y+0.5)); /* Y rounded to nearest integer */
  if (clip) {
    /* Any non-empty box is valid. */
    if (! is_array((i= where(abs(x-ximg) <= DC_BOX_SIZE)))
	|| ! is_array((j= where(abs(y-yimg) <= DC_BOX_SIZE)))) return 0n;
  } else {
    /* Only complete box is valid. */
    len= 2*DC_BOX_SIZE+1;
    if (numberof((i= where(abs(x-ximg) <= DC_BOX_SIZE))) != len
	|| numberof((j= where(abs(y-yimg) <= DC_BOX_SIZE))) != len) return 0n;
  }
  if (full) {
    DC_BOX_X= double(ximg(i))(,-:1:numberof(j));
    DC_BOX_Y= double(yimg(j))(-:1:numberof(i),);
  } else {
    DC_BOX_X= double(ximg(i))(,-);
    DC_BOX_Y= double(yimg(j))(-,);
  }
  DC_BOX_Z= DC_IMAGE(i, j);
  return 1n;

#if 0
  /* Old version of the routine. */
  x+= 1.5; y+= 1.5; /* indices start at 1 in Yorick and 0.5 is
		       for rounding to nearest integer */
  if (clip) {
    if ((i1= long(floor(x-DC_BOX_SIZE))) < 1) i1= 1;
    if ((i2= long(floor(x+DC_BOX_SIZE))) > width) i2= width;
    if ((j1= long(floor(y-DC_BOX_SIZE))) < 1) j1= 1;
    if ((j2= long(floor(y+DC_BOX_SIZE))) > height) j2= height;
    if (i1>i2 || j1>j2) return 0n;
  } else {
    if ((i1= long(floor(x-DC_BOX_SIZE))) < 1
	|| (i2= long(floor(x+DC_BOX_SIZE))) > width
	|| (j1= long(floor(y-DC_BOX_SIZE))) < 1
	|| (j2= long(floor(y+DC_BOX_SIZE))) > height) return 0n;
  }
  if (full) {
    DC_BOX_X= (indgen(i1:i2)-1.0)(,-:1:j2-j1+1);
    DC_BOX_Y= (indgen(j1:j2)-1.0)(-:1:i2-i1+1,);
  } else {
    DC_BOX_X= (indgen(i1:i2)-1.0)(,-);
    DC_BOX_Y= (indgen(j1:j2)-1.0)(-,);
  }
  DC_BOX_Z= DC_IMAGE(i1:i2,j1:j2);
  return 1n;
#endif
}

func dc_fit_coefs(degree)
/* DOCUMENT dc_fit_coefs();
       -or- dc_fit_coefs(degree);
     Fit distortion correction coefficients and return the actual degree
     of polynomial correction (0 if no correction can be estimated).
     If the degree of polynomial correction is not supplied, it is
     automatically chosen.

   SEE ALSO: dc_fit_simplest. */
{
  extern DC_COEF;
  extern DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  if (! numberof(DC_W)) return 0;
  i= where(DC_W > 0.0);
  number= numberof(i);
  if (number<2) return 0;
  w= DC_W(i);
  x= DC_X(i);
  y= DC_Y(i);
  xp= DC_XP(i);
  yp= DC_YP(i);

  if (max(xp)==min(xp) || max(yp)==min(yp) || degree==0) {
    /* Not enough vertices to fit lowest (1st order) 2D polynomial
       correction.  Fit the simplest correstion (translation +
       magnification + rotation) in the form:
         xp= c1 + c2 * x - c3 * y;
	 yp= c4 + c3 * x + c2 * y; */
    cf= dc_fit_simplest(x, y, xp, yp, w); /* forward correction */
    cb= dc_fit_simplest(xp, yp, x, y, w); /* backward correction */
    DC_COEF= [[cf(1), cf(2), -cf(3)], [cf(4), cf(3), cf(2)],
	      [cb(1), cb(2), -cb(3)], [cb(4), cb(3), cb(2)]];
    return 1;
  }

  if (! degree) {
    if (number<12) degree= 1;
    else if (number<20) degree= 2;
    else degree= 3;
  }
  DC_COEF= array(double, ((degree+1)*(degree+2))/2, 4);
  if (catch(0x10)) {
    dc_warn, "fit error ("+catch_message+
      ") try with a lower polynomial degree...";
    return dc_fit_coefs(degree-1);
  }
  DC_COEF(,1)= poly2_fit(xp, x, y, degree, w);
  DC_COEF(,2)= poly2_fit(yp, x, y, degree, w);
  DC_COEF(,3)= poly2_fit(x, xp, yp, degree, w);
  DC_COEF(,4)= poly2_fit(y, xp, yp, degree, w);
  return degree;
}

func dc_fit_simplest(x, y, xp, yp, w)
/* DOCUMENT [c1,c2,c3,c4]= dc_fit_simplest(x, y, xp, yp);
       -or- [c1,c2,c3,c4]= dc_fit_simplest(x, y, xp, yp, weight);
     Return the coefficients C=[C1,C2,C3,C4] of the least square fit of
     a simple distortion correction (translation+magnification+rotation):
        XP= C1 + C2 * X - C3 * Y;
	YP= C4 + C3 * X + C2 * Y; */
{
  if (is_void(w)) w= 1.0;
  w*= array(1.0, dimsof(x, y));
  w1= sum(w);
  wx= sum(w*x);
  wy= sum(w*y);
  w2= sum(w*(x*x+y*y));
  a= [[ w1,  wx, -wy, 0.0],
      [ wx,  w2, 0.0,  wy],
      [-wy, 0.0,  w2,  wx],
      [0.0,  wy,  wx,  w1]];
  b= [sum(w*xp), sum(w*(x*xp+y*yp)), sum(w*(x*yp-y*xp)), sum(w*yp)];
  return LUsolve(a, b);
}

func dc_improve(void)
/* DOCUMENT dc_improve;
     Improve current set of distortion correction coefficient by accounting
     for the distortion into the computation of the barycentres.

   SEE ALSO: dc_fit_coefs. */
{
  extern DC_COEF, DC_PRECISION, DC_MAXITER;
  extern DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  extern DC_BOX_X, DC_BOX_Y, DC_BOX_Z;

  for (iter=1 ; iter<=DC_MAXITER ; iter++) {
    /* Compute barycenter of valid spots in undistorted space. */
    number= numberof(DC_W);
    xp= yp= array(double, number);
    ok= array(int, number);
    update= 0;
    for (i=1, ip=0 ; i<number ; ++i) {
      /* Get region of interest for node. */
      if (DC_W(i) <= 0.0) continue;
      if (! dc_get_box(DC_X(i), DC_Y(i), full=1, clip=0)) {
	DC_W(i)*= -1;
	update= 1;
	continue;
      }

      /* Compute undistorted barycenter of spot. */
      tmp= dc_centroid(poly2(DC_BOX_X, DC_BOX_Y, DC_COEF(,1)),
		       poly2(DC_BOX_X, DC_BOX_Y, DC_COEF(,2)), DC_BOX_Z);
      if (tmp(1) <= 0.0) {
	DC_W(i)*= -1;
	update= 1;
	continue;
      }
      ok(i)= 1n;
      ip++;
      xp(ip)= tmp(2);
      yp(ip)= tmp(3);
    }
    if (ip<number) {
      xp= xp(:ip);
      yp= yp(:ip);
      number= ip;
    }

    if (update) {
      dc_warn, "bad node(s) removed";
      pause, 1000;
      dc_redraw_nodes;
    }
    if (number < 10) {
      dc_error, "not enough nodes";
      pause, 1000;
      return;
    }
  
    /* Find corresponding positions in distorted space. */
    local x, y;
    dc_reverse, x, y, xp, yp;
    i= where(ok);
    w0= DC_W(i);
    x0= DC_X(i); DC_X(i)= x; xp= DC_XP(i);
    y0= DC_Y(i); DC_Y(i)= y; yp= DC_XP(i);
    dst= abs(x-x0, y-y0);
    dc_fit_coefs;
    if (DC_DEBUG) {
      write, format="dc_improve %2d: %.3g/%.3g pixels (avg/max)\n", iter,
	avg(dst), max(dst);
    }
    precision= max(dst);
    if (precision <= DC_PRECISION * DC_SCALE) return;
  }
  dc_error, "too many iterations in dc_improve.";
  pause, 1000;
  return;
}

func dc_reverse(&x, &y, xp, yp)
/* DOCUMENT local x, y;
            precision= dc_reverse(x, y, xp, yp);
     Reverse distortion correction (XP,YP) -> (X,Y) and return worse
     precision in grid step. */
{
  extern DC_COEF, DC_MAXITER;
  det_min= 1e-30;

  if (dimsof(DC_COEF)(2)==3) {
    /* Inverse linear coorection. */
    det= DC_COEF(2,1) * DC_COEF(3,2) - DC_COEF(3,1) * DC_COEF(2,2);
    if (abs(det) <= det_min) error, "the system is (near) singular";
    x= (DC_COEF(3,1)*DC_COEF(1,2) - DC_COEF(1,1)*DC_COEF(3,2) +
	DC_COEF(3,2)*xp - DC_COEF(3,1)*yp) / det;
    y= (DC_COEF(1,1)*DC_COEF(2,2) - DC_COEF(2,1)*DC_COEF(1,2) -
	DC_COEF(2,2)*xp + DC_COEF(2,1)*yp) / det;
    return 0.0;
  }

  /* Distorted position(s) predicted by backward model. */
  x= poly2(xp, yp, DC_COEF(,3));
  y= poly2(xp, yp, DC_COEF(,4));

  /* Use a simple shooting method to refine distorted positions. */
  for (iter=1 ; iter<=DC_MAXITER ; iter++) {
    dxp= xp - poly2(x, y, DC_COEF(,1));
    dyp= yp - poly2(x, y, DC_COEF(,2));
    precision= max(abs(dxp, dyp));
    if (DC_DEBUG) write, format="dc_reverse %2d: %g\n", iter, precision;
    if (precision <= DC_PRECISION) return precision;
    drv= poly2_deriv(x, y, DC_COEF(,1)); dxx= drv(..,1); dxy= drv(..,2);
    drv= poly2_deriv(x, y, DC_COEF(,2)); dyx= drv(..,1); dyy= drv(..,2);
    det= dxx*dyy-dxy*dyx;
    if (min(abs(det)) <= det_min) error, "the system is (near) singular";
    x+= (dyy * dxp - dxy * dyp) / det;
    y+= (dxx * dyp - dyx * dxp) / det;
  }
  dc_error, "too many iterations in inversion of distortion correction";
  pause, 1000;
}

/*---------------------------------------------------------------------------*/
/* DRAWING ROUTINES */

struct dc_cursor_struct { float x, y; char key; };

func dc_get_cursor(void)
{
  xref= yref= x= y= 0.0F;
  mode= 0;
  posn= 0;
  key= "1";
  if (! __pgband(mode, posn, xref, yref, x, y, key)) error, "PGBAND failed";
  return dc_cursor_struct(x=x, y=y, key=*pointer(key)(1));
}

func dc_draw_box(xa, ya, hw)
/* DOCUMENT dc_draw_box, xa, ya, hw;
     Draw box(es) at position(s) (XA,YA) with half-width HW (a scalar)
     in current line attributes. */ 
{
  number= numberof(xa);
  for (i=1 ; i<=number ; i++) {
    x= xa(i); y= ya(i);
    __pgline, 5, [x-hw, x+hw, x+hw, x-hw, x-hw],
                 [y-hw, y-hw, y+hw, y+hw, y-hw];
  }
}

DC_COLOR_NAMES= ["bg", "fg", "red", "green", "blue", "cyan", "magenta",
		"yellow", "orange"];
func dc_get_color(color)
{
  if (is_void(color)) return color= 1; /* default foreground */
  if (structof(color)==string) return where(DC_COLOR_NAMES==color)(1)-1;
  return color;
}

func dc_draw_message(msg, color=)
{
  color= dc_get_color(color);

  /* Save current viewport, window and attributes. */
  vp1= vp2= vp3= vp4= 0.0F;
  wn1= wn2= wn3= wn4= 0.0F;
  __pgqvp, 0, vp1, vp2, vp3, vp4;
  __pgqwin, wn1, wn2, wn3, wn4;
  __pgsave;

  /* Change to label box and redraw label. */
  __pgbbuf;
  __pgsvp, 0.05F, 0.95F, 0.85F, 0.95F;
  __pgswin, 0.0F, 1.0F, 0.0F, 1.0F;
  __pgsfs, 1; /* solid fill-style */
  __pgsci, 0; /* fill with background color */
  __pgrect, 0.0F, 1.0F, 0.0F, 1.0F;
  __pgsci, color; /* text foreground */
  __pgstbg, -1; /* transparent text background */
  __pgsch, 2.0F; /* character height */
  __pgptxt, 0.5F, 0.4F, 0.0F, 0.5F, msg;
  __pgebuf;


  /* Restore current viewport, window and attributes. */
  __pgunsa;
  __pgsvp, vp1, vp2, vp3, vp4;
  __pgswin, wn1, wn2, wn3, wn4;
}

func dc_draw_view(clear)
/* DOCUMENT dc_draw_view, clear;
     Select (create) drawing data window and, if CLEAR is true, erase it
     and set the viewport. */
{
  extern DC_PLOT;
  //if (! __pgqid()) DC_PLOT= __pgopen("/xwin");
  if (! is_void(DC_PLOT) && DC_PLOT>0) {
    __pgslct, DC_PLOT;
    if (clear) __pgeras;
  } else {
    DC_PLOT= __pgopen("/xwin");
    __pgask, 0;    
  }

  if (clear) __pgsvp, 0.05F, 0.95F, 0.05F, 0.80F;
}

func dc_warn(message) { dc_draw_message, "WARNING: "+message, color=8; }
func dc_error(message) { dc_draw_message, "ERROR: "+message, color=2; }

func dc_redraw_all(void)
{
  extern DC_IMAGE;
  
  dc_draw_view, 1;

  /* Plot the image should be a routine by itself). */
  if (0) {
    a= sqrt(max(0.0, DC_IMAGE));
  } else {
    a= log(max(1.0, DC_IMAGE));
  }
  if ((dims= dimsof(a))(1)!=2) error, "expecting 2D array";
  width= dims(2); height= dims(3);
  cmin= cmax= 0n; /* must be an int */
  __pgqcir, cmin, cmax;
  cmin= 16;
  amin= min(a);
  amax= max(a);
  x1= 0.5f;
  x2= (width-1)+x1;
  y1= 0.5f;
  y2= (height-1)+y1;
  if (amax<=amin) return;
  if (amin) a-= amin;
  scale= ((cmax-cmin+1.0f)/(amax-amin));
  bias= cmin + 0.5f - scale*amin; 
  //__pgsvp, 0.1, 0.9, 0.1, 0.9;
  __pgwnad, x1, x2, y1, y2;
  a= int(scale*a + bias);
  __pgpixl, a, width, height, 1, width, 1, height,
    x1, x2, y1, y2;
  xopt= "BCINMTS";
  yopt= "BCINMTS";
  xtick= ytick= 0.0F;
  nxsub= nysub= 0n;
  __pgsci, 1;
  __pgbox, xopt, xtick, nxsub, yopt, ytick, nysub;

  /* Draw nodes. */
  dc_redraw_nodes;
  dc_draw_message, swrite(format="channel: %d", DC_CHANNEL), color="green";
}

func dc_redraw_nodes(void)
{
  extern DC_BOX_SIZE;
  extern DC_W, DC_X, DC_Y, DC_XP, DC_YP;
  extern DC_NORTH, DC_SOUTH, DC_WEST, DC_EAST, DC_CHANNEL;
  
  dc_draw_view, 0;
 
  /* Draw grid. */
  if ((number= numberof(DC_W)) <= 0) return;
  __pgbbuf;
  __pgsci, 4; /* blue */
  for (i=1 ; i<=number ; i++) {
    if ((j= DC_NORTH(i)) > 0)
      __pgline, 2, [DC_X(i), DC_X(j)], [DC_Y(i), DC_Y(j)];
    if ((j= DC_EAST(i))  > 0)
      __pgline, 2, [DC_X(i), DC_X(j)], [DC_Y(i), DC_Y(j)];
  }

  /* Draw boxes. */
  if (is_array((i= where(DC_W>0.0)))) {
    __pgsci, 5; /* cyan */
    dc_draw_box, DC_X(i), DC_Y(i), DC_BOX_SIZE;
  }
  if (is_array((i= where(DC_W<=0.0)))) {
    __pgsci, 2; /* red */
    dc_draw_box, DC_X(i), DC_Y(i), DC_BOX_SIZE;
  }
  __pgebuf;
}

/*---------------------------------------------------------------------------*/
/* UNUSED/BOGGY ROUTINES */
#if 0
func dc_forward_x(coef, x, y) { return poly2(x, y, coef(,1)); }
func dc_forward_y(coef, x, y) { return poly2(x, y, coef(,2)); }

func dc_deriv(c,x,y)
{
  dims= dimsof(x,y);
  dims(1)++;
  grow, dims, 4;
  drv= array(double, dims);
  drv(..,1:2)= poly2_deriv(x, y, c(,1));
  drv(..,3:4)= poly2_deriv(x, y, c(,2));
  return drv;
}

func poly_pow(b,p)
{
  if (p==0) return array(1.0, dimsof(b));
  if (p==1) return b;
  if (p==2) return b*b;
  if (p==3) return b*b*b;
  q= (p&3);
  p>>= 2;
  if (p==1) c= b;
  else if (p==2) c= b*b;
  else if (p==3) c= b*b*b;
  else c= poly_pow(b, p);      /* recurse for large powers (p>15) */
  c*= c;
  if (q&2) c*= b;
  if (q&1) return c*c*b;
  else return c*c;
}

func poly2_fit(y, x1, x2, degree, w)
/* DOCUMENT: c= poly2_fit(y, x1, x2, degree);
        -or- c= poly2_fit(y, x1, x2, degree, w);
     Returns the ((DEGREE+1)*(DEGREE+2))/2 coefficients of the 2D polynomial
     of degree DEGREE:
       C(1) + C(2)*X1 + C(3)*X2 + C(4)*X1^2 + C(5)*X1*X2 + C(6)*X2^2 + ... 
     which best fits Y in the least squares sense:
       C = arg min sum(W*(poly2(X1,X2,C) - Y)^2)
     The weights W are optional (by default, W=1.0).

   SEE ALSO: poly2. */
{
  number= ((degree+1)*(degree+2))/2; /* number of coefficients */
  p1= p2= array(0, number);
  for (k=0, i=0 ; k<=degree ; k++) {
    for (k1=k, k2=0 ; k1>=0 ; k1--, k2++) {
      i++; p1(i)= k1; p2(i)= k2;
    }
  }

  if (is_void(w)) w= 1.0;

  if ((scale1= double(max(abs(x1)))) > 0.0) x1/= scale1;
  else scale1= 1.0;
  if ((scale2= double(max(abs(x2)))) > 0.0) x2/= scale2;
  else scale2= 1.0;

  b= array(0.0, number);
  a= array(0.0, number, number);
  //b(1)= sum(w*y);
  //a(1,1)= sum(w);
  for (i=1 ; i<=number ; i++) {
    q= w * poly_pow(x1,p1(i)) * poly_pow(x2,p2(i));
    b(i)= sum(q * y);
    for (j=i ; j>=1 ; j--) {
      r= sum(q * poly_pow(x1,p1(j)) * poly_pow(x2,p2(j)));
      a(i,j)= r;
      a(j,i)= r;
    }
  }

  //  for (i=1 ; i<=number ; i++) write, a(i,);

  c= LUsolve(a, b);

  if (scale1!=1.0 || scale2!=1.0) {
    for (i=1 ; i<=number ; i++) {
      c(i)*= poly_pow(scale1,p1(i)) * poly_pow(scale2,p2(i));
    }
  }
  return c;
}

func char2string(c)
{
  return string(&c);
}
#endif

/*---------------------------------------------------------------------------*/
