/*
 * cp40stream.c --
 *
 *	Data stream input/output routines for CP40 library.  These routines
 *	provide unified access to file/socket data stream and also deal
 *	with byte order and file header.
 *
 * Copyright (c) 1995-1999, Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id: cp40stream.c,v 1.7 1999/04/27 14:42:57 eric Exp $
 *	$Log: cp40stream.c,v $
 *	Revision 1.7  1999/04/27 14:42:57  eric
 *	 - gpl_error -> gpl_sys_error in file_write_frame().
 *
 *	Revision 1.6  1998/09/18 16:49:11  eric
 *	 - Fix a bug in cp40_import (instance->raw_frame->count was not
 *	   updated correctly).
 *
 *	Revision 1.5  1998/09/10 14:04:36  eric
 *	 - Major rewritting of driver interfaces.
 *	 - Support for sockets.
 *
 *	Revision 1.4  1998/06/08 17:20:59  eric
 *	 - CP40 Version ~ 4.2 saved before version 4.3.
 *
 *	Revision 1.3  1997/11/19 18:40:53  eric
 *	 - Fixed bad increment in CP40_StreamRead byte-swapping ;-<
 *
 *	Revision 1.2  1997/11/17 11:24:31  eric
 *	 - Modify byte-swapping in CP40_StreamRead (now similar to David
 *	   Munro's implementation in Yorick).
 *
 *	Revision 1.1  1997/10/31 10:40:20  eric
 *	Initial revision
 *-----------------------------------------------------------------------------
 */
#include "cp40.h"
#include "gpl_string.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#define HAVE_RECV 1 /* undef this macro if you don't have recv() function */
/*---------------------------------------------------------------------------*/
/* PRIVATE MACROS */

/* Loop over each CP40 channels. */
#undef EACH_CHANNEL
#define EACH_CHANNEL(var) (var=0 ; var<CP40_NCHANNELS ; var++)

/* Define the following macro if last photon in frame is positive
   (by default, first photon in frame is positive). */
#undef LAST_PHOTON_IS_POSITIVE

/* Figure out whether byte swapping is needed or not. */
#undef SWAP_BYTES
#if (GPL_XDR_BYTE_ORDER == GPL_XDR_BIG_ENDIAN)
# define SWAP_BYTES 0
#else
# if (GPL_XDR_BYTE_ORDER == GPL_XDR_LITTLE_ENDIAN)
#  define SWAP_BYTES 1
# else
#  error "unsupported byte order"
# endif
#endif


/* The following macros are used to read/write data on/from file.  The
   result is no-zero if an error occurs (no message is stored).   Beware
   that argument `count` is used twice. */

#define FileIO(iofunc, p, n, s)        (iofunc(p, n, file) != (n))

#define FileReadIntegerArray(s, p, n)  FileIO(gpl_xdr_read_int, p, n, s)
#define FileReadInteger(s, p)          FileReadIntegerArray(file, p, 1)
#define FileReadRealArray(s, p, n)     FileIO(gpl_xdr_read_double, p, n, s)
#define FileReadReal(s, p)             FileReadRealArray(file, p, 1)
#define FileReadString(s, p, n)        (fread(p, 1, n, s)!=(n))

#define FileWriteIntegerArray(s, p, n) FileIO(gpl_xdr_write_int, p, n, s)
#define FileWriteInteger(s, p)         FileWriteIntegerArray(s, p, 1)
#define FileWriteRealArray(s, p, n)    FileIO(gpl_xdr_write_double, p, n, s)
#define FileWriteReal(s, p)            FileWriteRealArray(s, p, 1)
#define FileWriteString(s, p, n)       (fwrite(p, 1, n, s)!=(n))


/* Macro for reading from a socket (EMPTY_INPUT_SOCKET_BUFFER is the errno
   value when reading from a socket fails because the socket is non-blocking
   and there is no data to read). */
#ifdef HAVE_RECV
#  define SocketRead(fd, buf, len)   recv(fd, buf, len, 0)
#  define EMPTY_INPUT_SOCKET_BUFFER  EWOULDBLOCK
#else   /* Use read() function. */
#  define SocketRead(fd, buf, len)   read(fd, buf, len)
#  define EMPTY_INPUT_SOCKET_BUFFER  EAGAIN
#endif

/*---------------------------------------------------------------------------*/

/*
 * Public routines defined in this file:
 */

extern int cp40_readable(const cp40_instance_t *instance);
extern int cp40_rewindable(const cp40_instance_t *instance);
extern int cp40_updatable(const cp40_instance_t *instance);
extern int cp40_writable(const cp40_instance_t *instance);
extern int cp40_assoc_file(cp40_instance_t *instance, const char *name,
			   int mode);
extern int cp40_assoc_socket(cp40_instance_t *instance, const char *host,
			     int port);
extern int cp40_close_stream(cp40_instance_t *instance, int update);
extern int cp40_get_file_version(const char *path);
extern int cp40_set_exposure(cp40_instance_t *instance, int exposure);


/*
 * Forward declarations for private procedures defined later in this file:
 */

static void shift_frame_stack(cp40_instance_t *data, int discard);

static cp40_read_frame_op_t    bogus_read_frame;
static cp40_write_frame_op_t   bogus_write_frame;
static cp40_read_header_op_t   bogus_read_header;
static cp40_write_header_op_t  bogus_write_header;
static cp40_close_stream_op_t  bogus_close;
static cp40_rewind_stream_op_t bogus_rewind;

static cp40_read_frame_op_t    file_read_frame;
static cp40_write_frame_op_t   file_write_frame;
static cp40_read_header_op_t   file_read_header;
static cp40_write_header_op_t  file_write_header;
static cp40_close_stream_op_t  file_close;
static cp40_rewind_stream_op_t file_rewind;

static cp40_read_frame_op_t    socket_read_frame;
static cp40_close_stream_op_t  socket_close;

static cp40_set_exposure_op_t set_exposure, set_exposure_value_only;

static void delete_input_frame_stack(cp40_instance_t *instance);
/*	Discard all frames for readable stream being closed. */

static const cp40_stream_ops_t none_ops= {
  CP40_IO_NONE, CP40_IO_MODE_NONE,
  bogus_close,
  bogus_rewind,
  set_exposure_value_only,
  bogus_read_frame,
  bogus_write_frame,
  bogus_read_header,
  bogus_write_header,
};

static const cp40_stream_ops_t socket_ops= {
  CP40_IO_SOCKET, CP40_IO_MODE_READ,
  socket_close,
  bogus_rewind,
  set_exposure,
  socket_read_frame,
  bogus_write_frame,
  bogus_read_header,
  bogus_write_header,
};

static const cp40_stream_ops_t file_read_ops= {
  CP40_IO_FILE, CP40_IO_MODE_READ,
  file_close,
  file_rewind,
  set_exposure,
  file_read_frame,
  bogus_write_frame,
  file_read_header,
  bogus_write_header,
};

static const cp40_stream_ops_t file_write_ops= {
  CP40_IO_FILE, CP40_IO_MODE_WRITE,
  file_close,
  bogus_rewind,
  set_exposure_value_only,
  bogus_read_frame,
  file_write_frame,
  file_read_header,
  file_write_header,
};

static const cp40_stream_ops_t file_update_ops= {
  CP40_IO_FILE, CP40_IO_MODE_UPDATE,
  file_close,
  file_rewind,
  set_exposure,
  file_read_frame,
  bogus_write_frame,
  file_read_header,
  file_write_header,
};

/*
 * Miscellaneous macros for header routines.
 */

/*
 * Private data: list of standard attributes that have to stored into the
 * textual part of the header.
 */

static struct {
  const char *name;
  int         one_per_channel;
} std_attr[] = {
  {"exposure",    0},
  {"parity",      0},
  {"pixel",       0},
  {"spurious",    0},
  {"correction",  0},
  {"window",      1},
  {"distance",    1},
  {"flux",        1},
  {"ratio",       1},
  {"spread",      1},
  {NULL,          0},
};

/*---------------------------------------------------------------------------*/
/* PRIVATE ROUTINES */

static int get_version(const char *ident);

static int read_element(char *dst, const char *src, char **end_ptr);
/*	Read list element (i.e. reverse the action of write_element).
	Attempt to mimic Tcl rule for splitting into elements:
	  1. List elements are separated by at least one space, spaces at
	     the begin or at the end of the string are ignored.  Space
	     separators are:
	       ' '  (normal space)
	       '\f' (form feed)
	       '\n' (newline)
	       '\r' (carriage return)
	       '\t' (tabulation)
	       '\v' (vertical tabulation).
	  2. If a list element begins with an open curly brace '{', then
	     spaces are no more considered as element separator until the
	     matching close brace '}' which must be the last character of
	     the element, i.e. followed by at least one space or at the end
	     of the string.
	  2. Special characters can be escaped by preceding them by a
	     backslash '\\'.

	Arguments:
	  DST     - Starting address for writing element.  Must be large
		    enough to store the maximum number of characters that
		    can be written: strlen(src)+1 (i.e. all input characters
		    plus a final '\0').
	  SRC	  - Source string (can be NULL and can point to the same area
		    as dst).
	  END_PTR - Address where to store address of first character in
	            src that does not belong to current parsed element:
		    this address is the adequate value of src to read
		    next element.  Can be NULL.

	Return:
	   1 if one element can be read, 0 if there is no more elements and
	  -1 in case of error.

	Side effects:
	  Address of last character read in src is stored into value pointed
	  by end_ptr (if not NULL). */

static int add_text_info(gpl_dbuf_t *buf_ptr, const char *name,
			 const char *value, int append);
/*	Write attribute name-value strings into a dynamic buffer. Return
	standard result: GPL_OK or GPL_ERROR. */

/*---------------------------------------------------------------------------*/

#define ReadInteger(this, long_ptr) ReadIntegerArray(this, long_ptr, 1L)
#define ReadReal(this, doublePtr)   ReadRealArray(this, doublePtr, 1L)

#ifndef COMPATIBLE_WITH_TCL_LIST
#   define COMPATIBLE_WITH_TCL_LIST 1	/* The default is to write elements in
					 * a way compatible with Tcl (this
					 * macro only affects writting of
					 * elements). */
#endif

#undef  IDENT
#define IDENT       "CP40 DATA 1"
#undef  IDENT_SIZE
#define IDENT_SIZE  16
#undef  EMPTY_SISE
#define EMPTY_SIZE  (CP40_TEXT_OFFSET-IDENT_SIZE-680)
#undef  TEXT_SISE
#define TEXT_SIZE   (CP40_HEADER_SIZE-CP40_TEXT_OFFSET)
#undef  NCOEF
#define NCOEF       80
#if CP40_COEF_COUNT != NCOEF
#   error "Bad number of coefficients"
#endif
#undef NCHANNELS
#define NCHANNELS 4L
#if CP40_NCHANNELS != NCHANNELS
#   error "Bad number of channels"
#endif


/*---------------------------------------------------------------------------*/
int cp40_read_frame(cp40_instance_t *instance)
{
  int status= instance->stream.ops->read_frame(instance);
  instance->stream.error= status<0;
  instance->stream.end= status<=0;
  return status;
}

/*_____________________________________________________________________________
 *
 * cp40_close_stream --
 *
 *	Close stream associated with a CP40 data instance and maybe update
 *	header of data stream.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).  An error occurs if `update` is
 *	requested whereas the stream is read-only; in that case, the
 *	`close` operation is nevertheless attempted.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_close_stream(cp40_instance_t *instance, int update)
{
  cp40_stream_t *stream= &instance->stream;
  
  /* Only update (rewrite header) if requested and if this is relevant. */
  if (update && stream->ops->write_header != bogus_write_header
      && stream->ops->write_header(instance) != GPL_OK) {
    stream->ops->close(instance);
    return GPL_ERROR;
  }
  return stream->ops->close(instance);
}

/*_____________________________________________________________________________
 *
 * cp40_import --
 *
 *      Import raw photon data into writable CP40 stream.
 *
 * Return:
 *      Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *      See description in cp40.h.
 *_____________________________________________________________________________
 */


/* These macro are to check if a position is marked as being a wave-front
   sensor number. */
#undef  WFS_LOW
#define WFS_LOW  0x0000ffff
#undef  WFS_HIGH
#define WFS_HIGH 0xffff0000
#undef  WFS_SEAL
#define WFS_SEAL 0x007f0000
#undef  IS_WFS_NUMBER
#define IS_WFS_NUMBER(p) (((p) & WFS_HIGH) == WFS_SEAL)

#undef LAST_OR_FIRST
#define LAST_OR_FIRST(p) (((p)&0x80000000)==0)

#undef CHANNEL_ONLY
#define CHANNEL_ONLY(p)  ((p)&0x00800800)

#undef INSIDE_CCD
#define INSIDE_CCD(p)    (((p)&0x7ff)<(384<<2) && ((p)&0x7ff000)<(288<<14))

/* Set current raw frame count, write the frame and reset count (must be
   enclosed by braces). */
#undef WRITE_FRAME
#define WRITE_FRAME \
	raw_frame->count= count; \
	if (stream->ops->write_frame(instance) != GPL_OK) \
	    return GPL_ERROR; \
	count= 0

int cp40_import(cp40_instance_t *instance, const cp40_raw_photon_t *buffer,
		long number, int flags)
{
  cp40_raw_photon_t photon, *output;
  cp40_stream_t *stream= &instance->stream;
  cp40_raw_frame_t *raw_frame= instance->raw_frame;
  int *nphotons, i, size, count, date;
  int last= flags & CP40_IMPORT_LAST;
  int wfs= flags & CP40_IMPORT_WFS;
    
  size= raw_frame->size;
  count= raw_frame->count;
  output= raw_frame->buffer;
  nphotons= instance->nphotons;
  for (i=0 ; i<number ; ++i) {
    photon= buffer[i];
    if (wfs && IS_WFS_NUMBER(photon)) {
      /* Expect 2 successive photons marked by the seal 0x007F0000. */
      if (++i >= number) {
	fputs("WARNING *** wavefront number split at end of buffer\n",
	      stderr);
	break;
      }
      date= photon & WFS_LOW;
      photon= buffer[i];
      if (IS_WFS_NUMBER(photon)) {
	/* Flush photon buffer and update frame number. */
	date |= (photon << 16) & WFS_HIGH;
	if (count) {
	  WRITE_FRAME;
	}
	raw_frame->date= date;
	continue;
      }
    }
    if (! last && LAST_OR_FIRST(photon)) {
      WRITE_FRAME;
    }
    if (INSIDE_CCD(photon)) {
      if (count >= size) {
	/* No space for one more photon: realloc buffer. */
	if (size < 256) {
	  size= 256;
	}
	while (count >= size) {
	  size *= 2;
	}
	raw_frame->count= count;  /* Otherwise, cp40_resize_raw_frame will
				     omit to copy some photons. */
	if (cp40_resize_raw_frame(raw_frame, size) != GPL_OK) {
	  return GPL_ERROR;
	}
	size= raw_frame->size;
	output= raw_frame->buffer;
      }
      output[count++]= 0x01000000 | (photon & 0x00ffffff);
      switch (photon & 0x800800) {
      case 0x000000: nphotons[0]++; break;
      case 0x000800: nphotons[1]++; break;
      case 0x800000: nphotons[2]++; break;
      default:       nphotons[3]++; break;
      }
    }
    if (last && LAST_OR_FIRST(photon)) {
      WRITE_FRAME;
    }
  }

  /* Update structure and return. */
  raw_frame->count= count;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/
/* BOGUS STREAM OPERATIONS */

static int bogus_read_frame(cp40_instance_t *instance)
{ gpl_error("cannot read data"); return -1; }

static int bogus_write_frame(cp40_instance_t *instance)
{ return gpl_error("cannot write data"); }

static int bogus_read_header(cp40_instance_t *instance)
{ return gpl_error("cannot read header"); }

static int bogus_write_header(cp40_instance_t *instance)
{ return gpl_error("cannot write header"); }

static int bogus_close(cp40_instance_t *instance)
{ return gpl_error("cannot close data stream"); }

static int bogus_rewind(cp40_instance_t *instance)
{ return gpl_error("cannot rewind data stream"); }

/*---------------------------------------------------------------------------*/
/* CHECKING STREAM CAPABILITIES */

#undef FUNCTION
#define FUNCTION(NAME, TEST)						      \
int NAME(const cp40_instance_t *instance)				      \
{ return instance!=NULL && instance->stream.ops->TEST; }
FUNCTION(cp40_readable,   read_frame   != bogus_read_frame)
     FUNCTION(cp40_rewindable, rewind       != bogus_rewind)
     FUNCTION(cp40_updatable,  write_header != bogus_write_header)
     FUNCTION(cp40_writable,   write_frame  != bogus_write_frame)

/*---------------------------------------------------------------------------*/
const char *cp40_get_io_filename(cp40_instance_t *instance)
{
  if (! instance || instance->stream.ops->type != CP40_IO_FILE) return NULL;
  return instance->stream.io.file.name;
}

const char *cp40_get_io_hostname(cp40_instance_t *instance)
{
  if (! instance || instance->stream.ops->type != CP40_IO_SOCKET) return NULL;
  return instance->stream.io.socket.host;
}

int cp40_get_io_port(cp40_instance_t *instance)
{
  if (! instance || instance->stream.ops->type != CP40_IO_SOCKET) return -1;
  return instance->stream.io.socket.port;
}

int cp40_get_io_mode(cp40_instance_t *instance)
{ return instance ? instance->stream.ops->mode : CP40_IO_MODE_NONE; }

int cp40_get_io_type(cp40_instance_t *instance)
{ return instance ? instance->stream.ops->type : CP40_IO_NONE; }

/*---------------------------------------------------------------------------*/

/* Slow but sure way (against interrupts) to clear stream descriptor
   while keaping a sensible operation table. */
void cp40_init_stream(cp40_stream_t *stream)
{
  char *p1= (char *)stream;
  char *p2= (char *)&stream->ops;
  char *p3= p2 + sizeof stream->ops;
  char *p4= p1 + sizeof *stream;
  stream->ops= &none_ops;
  while (p1 < p2) *p1++= 0;
  while (p3 < p4) *p3++= 0;
}
/*_____________________________________________________________________________
 *
 * cp40_assoc_file --
 *
 *	Associate a CP40 data instance with a file stream.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).
 *
 * Side effects:
 *	See description.
 *-----------------------------------------------------------------------------
 */

int cp40_assoc_file(cp40_instance_t *instance, const char *name, int mode)
{
  cp40_io_file_t *io= &instance->stream.io.file;

  /* Check argument. */
  if (instance->stream.ops != &none_ops) {
    return gpl_error("CP40 data instance already has a data stream");
  }
  io->handle= NULL;
  io->name= NULL;

  /* Store file name for further reference. */
  if (! (io->name= gpl_str_clone(name))) {
    return gpl_error("insufficient memory to store file name");
  }

  /* Open stream and set its operation table. */
  switch (mode) {
  case CP40_IO_MODE_READ:
    if (! (io->handle= fopen(name, "r"))) goto open_error;
    instance->stream.ops= &file_read_ops;
    break;
  case CP40_IO_MODE_WRITE:
    if (access(name, F_OK) == 0) {
      gpl_error("cannot open file to write: file already exists");
      goto error;
    }
    if (! (io->handle= fopen(name, "w+"))) goto open_error;
    instance->stream.ops= &file_write_ops;
    break;
  case CP40_IO_MODE_UPDATE:
    if (access(name, F_OK) != 0) {
      return gpl_error("cannot open file to update: file does not exists");
      goto error;
    }
    if (! (io->handle= fopen(name, "r+"))) goto open_error;
    instance->stream.ops= &file_update_ops;
    break;
  default:
    gpl_error("cannot open file stream: bad mode");
    goto error;
  }

  /* Read or write (skip) header. */
  if ((mode==CP40_IO_MODE_WRITE ? file_write_header
       : file_read_header)(instance) == GPL_OK
      && cp40_set_exposure(instance, GPL_MAX(instance->exposure, 1)) == GPL_OK) {
    return GPL_OK;
  }
 error:
  instance->stream.ops= &none_ops;
  if (io->handle) {
    fclose(io->handle);
    io->handle= NULL;
  }
  if (io->name) {
    gpl_free(io->name);
    io->name= NULL;
  }
  return GPL_ERROR;
 open_error:
  gpl_sys_error("open file error");
  goto error;
}

/*_____________________________________________________________________________
 *
 * file_close --
 *
 *	Close a CP40 data file stream (does not update header).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	Discard all frames for readable file stream and flush raw photon
 *	buffer for writable file stream.  Data stream cannot be accessed
 *	anymore.
 *_____________________________________________________________________________
 */

static int file_close(cp40_instance_t *instance)
{
  cp40_stream_t *stream= &instance->stream;
  int flush= (stream->ops->write_frame == file_write_frame);
  FILE *file= stream->io.file.handle;
  
  stream->ops= &none_ops;
  GPL_FREE_MAYBE(stream->io.file.name);
  if (flush) {
    /* Flush frame buffer. */
    if (instance->raw_frame->count > 0
	&& file_write_frame(instance) != GPL_OK) {
      fclose(file);
      return GPL_ERROR;
    }
  } else {
    /* Discard all frames for readable file stream. */
    delete_input_frame_stack(instance);
  }
  if (fclose(file)) {
    return gpl_sys_error("close file error");
  }
  return GPL_OK;
}


/*_____________________________________________________________________________
 *
 * file_rewind --
 *
 *	Go back to the beginning of data part of a CP40 data stream.
 *	This routine must only be part of stream operations for data
 *	file stream in read/update mode.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	Stream access is moved to the end of the header part and whole frame
 *	stack is shifted.
 *_____________________________________________________________________________
 */

static int file_rewind(cp40_instance_t *instance)
{
  if (fseek(instance->stream.io.file.handle, CP40_HEADER_SIZE, SEEK_SET)) {
    return gpl_sys_error("seek file error");
  }
  shift_frame_stack(instance, instance->exposure);
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * file_read_frame --
 *
 *	Read next frame of raw photons in file stream data.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int file_read_frame(cp40_instance_t *instance)
{
  FILE *file= instance->stream.io.file.handle;
  cp40_raw_frame_t *raw_frame= instance->raw_frame;
  gpl_xdr_int_t raw_frame_header[2];
  int count, new_count, nread;

  /* Read frame header: frame number and number of photons. */
  if ((nread= gpl_xdr_read_int(raw_frame_header, 2, file)) != 2) {
    if (nread == 0) {
      return 0;	/* Normal end-of-file. */
    }
    if (nread > 0) {
      gpl_error("read incomplete frame header");
    }
    return -1;	/* Error. */
  }
  
  /* Get rid of the oldest frame(s) and set new frame number. */
  count= raw_frame_header[0] - raw_frame->date; /* # of frames to discard */
  if (count <= 0) {
    if (ftell(file) != (CP40_HEADER_SIZE + 8)) {
      gpl_warning("frames numbered in reverse order!");
    }
    count= instance->exposure;	/* Get rid of all previous frames. */
  }
  shift_frame_stack(instance, count);
  raw_frame->date= raw_frame_header[0];
  
  /* Read photons into frame. */
  if ((count= raw_frame_header[1]) > 0) {
    if ((new_count= raw_frame->count + count) > raw_frame->size) {
      if (cp40_resize_raw_frame(raw_frame, new_count) != GPL_OK) {
	/* Skip until end-of-frame otherwise
	   this error cannot be recovered. */
	fseek(file, count*sizeof(cp40_raw_photon_t), SEEK_CUR);
	return -1;
      }
    }
    nread= gpl_xdr_read_int(&raw_frame->buffer[raw_frame->count], count,
			  file);
    if (nread != count) {
      gpl_error("incomplete frame data");
      return -1;	/* Error. */
    }
    raw_frame->count= new_count;
    instance->count_list[instance->exposure - 1]= count;
  } else if (count < 0) {
    gpl_error("bad frame header");
    return -1;	/* Error. */
  }

  return 1;
}

/*_____________________________________________________________________________
 *
 * file_write_frame --
 *
 *	Write current frame of raw photons in file stream data, reset
 *	raw frame count instance->raw_frame->count and increment frame
 *	number and count (instance->raw_frame->date and instance->nframes).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int file_write_frame(cp40_instance_t *instance)
{
  FILE *file= instance->stream.io.file.handle;
  cp40_raw_frame_t *raw_frame= instance->raw_frame;
  gpl_xdr_int_t raw_frame_header[2];
  
  if (raw_frame->count > 0) {
    if (fseek(file, 0L, SEEK_END)) {
      return gpl_sys_error("cannot seek to end of stream file");
    }
    raw_frame_header[0]= raw_frame->date;
    raw_frame_header[1]= raw_frame->count;
    if (gpl_xdr_write_int(raw_frame_header, 2, file) != 2L) {
      return gpl_sys_error("error writing frame header");
    }
    if (gpl_xdr_write_int(raw_frame->buffer, raw_frame->count, file)
	!= raw_frame->count) {
      return gpl_sys_error("error writing frame data");
    }
  }
  raw_frame->count= 0;
  raw_frame->date+= 1;
  instance->nframes+= 1;
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * cp40_assoc_socket --
 *
 *	Associate a CP40 data instance with a socket stream connected to a
 *	raw photon server.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).
 *
 * Side effects:
 *	Initialize CP40_Stream structure.
 *_____________________________________________________________________________
 */

int cp40_assoc_socket(cp40_instance_t *instance, const char *host, int port)
{
  cp40_io_socket_t *io= &instance->stream.io.socket;
  
  /* Check argument. */
  if (instance->stream.ops != &none_ops) {
    return gpl_error("CP40 data instance already has a data stream");
  }
  io->handle= -1;
  io->host= NULL;
  io->port= port;
  io->buffer= NULL;
  io->size= 0;
  io->pending= 0;

  /* Store host name for further reference. */
  if (! (io->host= gpl_str_clone(host))) {
    gpl_error("insufficient memory to store server host name");
    goto error;
  }

  /* Open socket stream in non-blocking mode.  So that reading data will
     return 0 (no frame) until (1) the server accept the connection and
     (2) enough photons are available. */
#if 0
  if ((io->handle= gpl_create_socket(port, host,
				     NULL, 0, GPL_SOCK_ASYNC)) == -1) {
    gpl_sys_error("cannot open client socket");
    goto error;
  }
#else
  if ((io->handle= gpl_create_socket(port, host, NULL, 0, 0)) == -1) {
    gpl_sys_error("cannot open client socket");
    goto error;
  }
  if (fcntl(io->handle, F_SETFL,
	    fcntl(io->handle, F_GETFL) | O_NONBLOCK) == -1) {
    gpl_sys_error("cannot set non-blocking mode");
    goto error;
  }
#endif

  /* Set socket stream and stream access functions table pointers. */
  instance->stream.ops= &socket_ops;
  if (cp40_set_exposure(instance, GPL_MAX(instance->exposure, 1)) != GPL_OK) {
    return GPL_ERROR;
  }
  return GPL_OK;

 error:
  instance->stream.ops= &none_ops;
  GPL_FREE_MAYBE(io->host);
  if (io->handle>=0) close(io->handle);
  return GPL_ERROR;
}

/*_____________________________________________________________________________
 *
 * socket_close --
 *
 *	Close a CP40 data socket stream (does not update header).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	Data stream cannot be accessed anymore; also see description.
 *_____________________________________________________________________________
 */

static int socket_close(cp40_instance_t *instance)
{
  cp40_stream_t *stream= &instance->stream;
  int fd= stream->io.socket.handle;

  /* To avoid problems when one of the open/close stream routine gets
     interrupted by some signal (e.g. user hits Ctrl-C), the trick is to
     set the stream operation table after and before every other settings
     in `open` and `close` operations respectively. */
  instance->stream.ops= &none_ops;
  
  /* Discard all frames for readable file stream. */
  delete_input_frame_stack(instance);

  /* Free socket hostname and buffer and close socket descriptor. */
  GPL_FREE_MAYBE(stream->io.socket.host);
  GPL_FREE_MAYBE(stream->io.socket.buffer);
  if (fd > 0 && close(fd)) {
    return gpl_sys_error("close file error");
  }
  return GPL_OK;
}

#if SWAP_BYTES
#  define MASK (1<<7)
#else
#  define MASK (1<<31)
#endif
#define BLOCK 1024

static int socket_read_frame(cp40_instance_t *instance)
{
#if SWAP_BYTES
  char *dst, *src;
#else
  int *dst, *src;
#endif
  int count, i, nbytes, new_count;
  int *ptr, *end;
  cp40_io_socket_t *io= &instance->stream.io.socket;
  cp40_raw_frame_t *raw_frame= instance->raw_frame;

  /* First allocation of pending data buffer. */
  if (io->buffer == NULL) {
    io->pending= 0;
    if ((io->buffer= gpl_alloc(2*BLOCK)) == NULL) {
      (void)gpl_sys_error("insufficient memory for socket buffer");
      return -1;
    }
    io->size= 2*BLOCK;
  }
  ptr= (int *)io->buffer;
  
  /* Search for end/begin of frame (i.e. a positive photon)
     from the start of the buffer. */
  if (io->pending >= 4) {
    end= (int *)io->buffer + (io->pending >> 2);
#ifdef LAST_PHOTON_IS_POSITIVE
    while (ptr < end) {
      if (! (*ptr++ & MASK)) goto found_frame;
    }
#else
    while (++ptr < end) {
      if (! (*ptr & MASK)) goto found_frame;
    }
#endif
  }

  /* No frame found in pending data buffer.  Possibly augment size of buffer
     before reading more bytes. */
  if ((io->size - io->pending) < BLOCK) {
    char *new, *old= io->buffer;
    if ((new= gpl_alloc(io->size + BLOCK)) == NULL) {
      (void)gpl_sys_error("insufficient memory for socket buffer");
      return -1;
    }
    if (io->pending) memcpy(new, old, io->pending);
    io->buffer= new;
    io->size+= BLOCK;
    if (old) gpl_free(old);
    ptr= (int *)new + (ptr - (int *)old);
  }
  
  /* Try to read some more data and search for end/begin of frame
     (i.e. a positive photon) from where we previously stop. */
  if ((nbytes= SocketRead(io->handle, io->buffer + io->pending,
			  io->size - io->pending)) <= 0) {
#if 0
    if (errno == ECONNRESET)
      fprintf(stderr, "SocketRead: errno= ECONNRESET\n");
#endif
    if (nbytes == 0 || errno == EMPTY_INPUT_SOCKET_BUFFER) {
      errno= 0;	/* Clear error code. */
      return 0;
    }
    gpl_sys_error("read error");
    return -1;
  }
  io->pending+= nbytes;
  end= (int *)io->buffer + (io->pending >> 2);
#ifdef LAST_PHOTON_IS_POSITIVE
  while (ptr < end) {
    if (! (*ptr++ & MASK))
      goto found_frame;
  }
#else
  if (ptr < end) {
    do {
      if (! (*ptr & MASK))
	goto found_frame;
    } while (++ptr < end);
  }
#endif
  
  /* No complete frame read. */
  return 0;
  
 found_frame:
  /* Found a complete frame: store it in a the instance photon buffer.
     Note: PTR is the address of first photon in next frame. */
  nbytes= (count= ptr - (int *)io->buffer) << 2;
  shift_frame_stack(instance, 1);	/* Get rid of the previous frame. */
  if ((new_count= raw_frame->count + count) > raw_frame->size) {
    if (cp40_resize_raw_frame(raw_frame, new_count) != GPL_OK) {
      /* Unable to store this frame this time, anyway we'll
	 get it next time (if there is enough memory!). */
      return -1;
    }
  }
#if SWAP_BYTES
  src= (char *)io->buffer;
  dst= (char *)&raw_frame->buffer[raw_frame->count];
  for (i=0; i<nbytes; i+=4) {
    dst[i  ]= src[i+3];
    dst[i+1]= src[i+2];
    dst[i+2]= src[i+1];
    dst[i+3]= 1 /*src[i  ]*/;
  }
#else
  src= (int *)io->buffer;
  dst= (int *)&raw_frame->buffer[raw_frame->count];
  for (i=0; i<count; ++i) {
    dst[i]= src[i] & 0x01ffffff;
  }
#endif
  if ((io->pending-= nbytes) > 0)
    memcpy(io->buffer, io->buffer+nbytes, io->pending);
  raw_frame->date += 1;
  raw_frame->count= new_count;
  instance->count_list[instance->exposure - 1]= count;
  return 1;
}

/*_____________________________________________________________________________
 *
 * cp40_set_exposure, set_exposure, set_exposure_value_only --
 *
 *	Set number of elementary frames in a raw frame.  Additional frames,
 *	if any, are considered to have zero count.  Only effective if
 *	data is "readable".
 *
 * Arguments:
 *	cp40_instance_t *instance	Photon counting data descriptor.
 *	long       exposure	Number of short-exposure frames per
 *				super-frame.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_set_exposure(cp40_instance_t *instance, int exposure)
{
  if (instance == NULL) return GPL_ERROR;
  return instance->stream.ops->set_exposure(instance, exposure);
}

static int set_exposure_value_only(cp40_instance_t *instance, int exposure)
{
  if (exposure <= 0) return gpl_error("bad value for \"exposure\"");
  instance->exposure= exposure;
  return GPL_OK;
}

static int set_exposure(cp40_instance_t *instance, int exposure)
{
  int *new_ptr, *old_ptr;
  int  i, n;


  /* Check arguments and return earlier if nothing have to be done. */
  if (exposure <= 0) {
    return gpl_error("bad value for \"exposure\"");
  }
  if (instance->exposure == exposure && instance->count_list != NULL) {
    return GPL_OK;
  }

  /* (Re)allocate array of photon counts. */
  if (! GPL_ALLOC_AT(new_ptr, exposure)) {
    return gpl_sys_error("cannot (re)allocate array of photon counts");
  }
  n= exposure - instance->exposure;	/* Number of additional frames. */
  if (n > 0) {
    /* Number of frames per super-frame has been augmented: adjust
       the new frame stack so that oldest frames (i.e., very first ones)
       can be considered as empty. */
    for (i=0 ; i<n ; ++i) {
      new_ptr[i]= 0;
    }
    for (i=n ; i<exposure ; ++i) {
      new_ptr[i]= instance->count_list[i-n];
    }
  } else if (n < 0) {
    /* Number of frames per super-frame has been reduced: shift photons
     * and counts out of the stack. */
    shift_frame_stack(instance, -n);
    for (i=0 ; i<exposure ; ++i) {
#warning "BUG: was count_list[i-n]"
      new_ptr[i]= instance->count_list[i];
    }
  }

  /* Replace array of photon counts. */
  old_ptr= instance->count_list;
  instance->count_list= new_ptr;
  instance->exposure= exposure;
  GPL_FREE_MAYBE(old_ptr);
  return GPL_OK;
}

static void delete_input_frame_stack(cp40_instance_t *instance)
{
  if (instance && instance->count_list != NULL) {
    int *count_list= instance->count_list;
    shift_frame_stack(instance, instance->exposure);
    instance->count_list= NULL;
    gpl_free(count_list);
  }
}


/*_____________________________________________________________________________
 *
 * shift_frame_stack --
 *
 *	Shift frames out of the frame stack.  The number of frames to discard
 *	must be positive and may be greater than instance->exposure.
 *	Beware that instance *MUST* be in read mode.
 *
 * Arguments:
 *	cp40_instance_t *this	Photon counting data descriptor.
 *	long       discard	Number of frames to discard (must be >= 0).
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static void shift_frame_stack(cp40_instance_t *this, int discard)
{
  
  int i, count, *src, *dst, keep;
  cp40_raw_frame_t *raw_frame;

  if ((dst= this->count_list) == NULL) {
    return;
  }
  if (discard >= this->exposure) {
    /* Get rid of the whole frame stack (this is faster than the more
       general code that follow). */
    discard= this->exposure;
    for (i=0 ; i<discard ; ++i)
      dst[i]= 0;
    this->raw_frame->count= 0;
    return;
  }
  
  /* Shift counts (also integrate number of photons to keep) and set
     count of frames shifted in to 0. */
  src= dst + discard;		   /* Address of count in 1st frame to keep. */
  keep= this->exposure - discard;  /* Number of frames to keep. */
  count= 0;                        /* Reset number of photons to keep. */ 
  for (i=0 ; i<keep ; ++i) {
    int number= src[i];
    src[i]= 0;
    dst[i]= number;
    count+= number;
  }
 
  /* Shift photons. */
  raw_frame= this->raw_frame;
  memmove(raw_frame->buffer, &raw_frame->buffer[raw_frame->count - count],
	  count * sizeof(*raw_frame->buffer));
  raw_frame->count= count; /* new photon count */
}

/*_____________________________________________________________________________
 *
 * cp40_read_header --
 *
 *	Reread CP40 data instance attributes from stream header.
 *
 * Arguments:
 *	cp40_instance_t *instance	CP40 data instance (must have a
 *					readable and seekable associated
 *					stream).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_read_header(cp40_instance_t *instance)
{
  if (instance == NULL) return GPL_ERROR;
  return instance->stream.ops->read_header(instance);
}

/*_____________________________________________________________________________
 *
 * cp40_write_header --
 *
 *	Rewrite CP40 data instance attributes into stream header.
 *
 * Arguments:
 *	cp40_instance_t *instance	CP40 data instance (must have a
 *					writable and seekable associated
 *					stream).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_write_header(cp40_instance_t *instance)
{
  if (instance == NULL) return GPL_ERROR;
  return instance->stream.ops->write_header(instance);
}

/*_____________________________________________________________________________
 *
 * cp40_get_file_version --
 *
 *	Get CP40 data file version.
 *
 * Arguments:
 *	const char *path	File name.
 *
 * Return:
 *	An integer > 0, or -1 in case of error.
 *
 * Side effects:
 *	In case of error, an error message is left.
 *_____________________________________________________________________________
 */

int cp40_get_file_version(const char *path)
{
  FILE *filePtr;
  char ident[IDENT_SIZE];
  int version;

  filePtr= fopen(path, "r");
  if (filePtr == NULL) {
    gpl_sys_error("cannot open file for reading");
    return -1;
  }
  if (fread(ident, 1, IDENT_SIZE, filePtr) != IDENT_SIZE
      || (version= get_version(ident)) < 0) {
    gpl_error("bad file identifier");
    fclose(filePtr);
    return -1;
  }
  if (version != 1) {
    gpl_error("bad file version");
    fclose(filePtr);
    return -1;
  }
  if (fclose(filePtr)) {
    gpl_sys_error("cannot close file");
    return -1;
  }
  return version;
}

/*_____________________________________________________________________________
 *
 * file_read_header --
 *
 *	Read or read again CP40 data instance attributes from file stream
 *	header.  Can be called at any time since the routine takes care of
 *	seeking the file stream.
 *
 * Arguments:
 *	cp40_instance_t *instance	CP40 data instance (must have a
 *					readable associated stream).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int file_read_header(cp40_instance_t *instance)
{
  char ident[IDENT_SIZE];
  char empty[EMPTY_SIZE];
  char text[TEXT_SIZE+1];
  char name[TEXT_SIZE+1];	/* Split elements cannot be larger than actual
				 * size of textual header. */
  char *next, *end, *value;
  int nframes, nsaturations, nphotons[4];
  double scale, angle, raw_coefs[NCOEF];
  int status;
  int result= GPL_OK;
  FILE *file;
  long offset;

  if (instance->task) {
    return gpl_error("cannot read header: scheduled tasks exist");
  }

  /* Record current file stream position to be able to return where
     we previously were (at least after the header) and rewind file. */
  file= instance->stream.io.file.handle;
  if ((offset= ftell(file)) < 0 || fseek(file, 0L, SEEK_SET)) {
    return gpl_sys_error("cannot seek to beginning of file header");
  }
  if (offset < CP40_HEADER_SIZE) offset= CP40_HEADER_SIZE;
  
  
  /* Read and check identifier string. */
  if (FileReadString(file, ident, IDENT_SIZE)) {
    result= gpl_sys_error("cannot read identifier");
    goto seek_back;
  }
  if (get_version(ident) != 1) {
    result= gpl_error("bad file identifier");
    goto seek_back;
  }
  
  /* Read rest of header (this should left stream access pointer at the
   * beginning of the photon counting data). */
  if (/**/ FileReadInteger(file, &nframes)
      ||   FileReadInteger(file, &nsaturations)
      ||   FileReadIntegerArray(file, nphotons, NCHANNELS)
      ||   FileReadReal(file, &scale)
      ||   FileReadReal(file, &angle)
      ||   FileReadRealArray(file, raw_coefs, NCOEF)
      ||   FileReadString(file, empty, EMPTY_SIZE)
      ||   FileReadString(file, text, TEXT_SIZE)) {
    result= gpl_sys_error("cannot read header");
    goto seek_back;
  }


  /* Left file stream position where it previously was. */
 seek_back:
  if (fseek(file, offset, SEEK_SET) && result == GPL_OK) {
    return gpl_sys_error("cannot seek back after reading header");
  }
  if (result != GPL_OK) return GPL_ERROR;


  /* Instanciate CP40 data from binary header information so that
     subsequent errors may return immediatly. */
  instance->nframes= nframes;
  instance->nsaturations= nsaturations;
  memcpy(instance->nphotons, nphotons, sizeof(nphotons));
  instance->scale= scale;
  instance->angle= angle;
  memcpy(instance->raw_coefs, raw_coefs, sizeof(raw_coefs));
  _cp40_invalidate_cache(instance);

  /* Parse textual information (any error in this part result in a
     warning). */
  text[TEXT_SIZE]= 0;	/* Make sure end-of-string is written. */
  next= text;
  for (;;) {
    /* Get next attribute name-value pair. */
    status= read_element(name, next, &next);
    if (status != 1) {
      if (status != 0) {
	gpl_warning("cannot get attribute name-value (%s)",
		  gpl_error_info(1));
      }
      break;	/* Immediate return. */
    }

    /* Split name-value pair (this can be done in-place since read
     * elements will always be shorter than their written counterparts). */
    status= read_element(name, name, &value);
    if (status != 1) {
      gpl_warning("cannot get attribute name (%s)", status == -1 ?
		gpl_error_info(1) : "unexpected end of string");
      break;	/* Immediate return. */
    }
    status= read_element(value, value, &end);
    if (status != 1) {
      gpl_warning("cannot get attribute value (%s)",
		status == -1 ? gpl_error_info(1) : "no more elements");
      break;	/* Immediate return. */
    }
    while (GPL_IS_SPACE(*end)) {
      end++;
    }
    if (*end != 0) {
      gpl_warning("too many elements in attribute name-value");
    }

    /* Set attribute (just print out warnings in case of error). */
    if (cp40_set(instance, name, value) != GPL_OK) {
      gpl_warning(gpl_error_info(1));
    }
  }

  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * file_write_header --
 *
 *	Write or rewrite CP40 data instance attributes into file stream
 *	header.  Can be called at any time since the routine takes care of
 *	seeking the file stream.  Extra attributes are written in the order
 *	they were defined.  Extra attributes that would overflow the header
 *	size are discarded (a warning message is print to stderr).
 *
 * Arguments:
 *	cp40_instance_t *INSTANCE   CP40 data instance (must have a writable
 *				    associated stream).
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR (error message can be
 *	recovered by gpl_error_info).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int file_write_header(cp40_instance_t *instance)
{
  gpl_dbuf_t dbuf;
  gpl_dstr_t dstr;
  char ident[IDENT_SIZE];
  char empty[EMPTY_SIZE];
  char text[TEXT_SIZE];
  int index, status, chn, one_per_channel;
  unsigned length;
  const char *name, *value;
  char fullname[32];	/* Must be large enough to store name+channel. */
  FILE *file;
  long offset;

  /* Setup buffers. */
  gpl_dbuf_init(&dbuf);
  gpl_dstr_init(&dstr);
  memset(ident, 0, IDENT_SIZE);
  memset(empty, 0, EMPTY_SIZE);
  memset(text, 0, TEXT_SIZE);
  strcpy(ident, IDENT);

  /* Prepare textual header. */
  length= 0;
  for (index= 0; std_attr[index].name != NULL; index++) {
    one_per_channel= std_attr[index].one_per_channel;
    for EACH_CHANNEL(chn) {
      if (one_per_channel) {
	sprintf(fullname, "%s,%d", std_attr[index].name, chn);
	name= fullname;
      } else {
	name= std_attr[index].name;
      }
      if (cp40_get(instance, name, &dstr) != GPL_OK) {
	gpl_warning("cannot get attribute \"%s\": %s", name,
		    gpl_error_info(1));
	continue;
      }
      if (add_text_info(&dbuf, name,
			gpl_dstr_value(&dstr), length!=0) != GPL_OK) {
	gpl_warning(gpl_error_info(1));
	break;
      }
      if (GPL_DBUF_SIZE(&dbuf) > TEXT_SIZE) {
	gpl_warning("cannot store attribute \"%s\"", name);
      } else {
	length= GPL_DBUF_SIZE(&dbuf);
      }
      if (! one_per_channel) break;
    }
  }
  index= -1;
  for (;;) {
    status= cp40_get_next_extra(instance, &index, &name, &value);
    if (status == 0) {
      break;	/* Last attribute. */
    }
    if (status == -1) {
      gpl_warning(gpl_error_info(1));
      break;
    }
    if (add_text_info(&dbuf, name, value, length!=0) != GPL_OK) {
      gpl_warning(gpl_error_info(1));
      break;
    }
    if (GPL_DBUF_SIZE(&dbuf) > TEXT_SIZE) {
      gpl_warning("cannot store attribute \"%s\"", name);
    } else {
      length= GPL_DBUF_SIZE(&dbuf);
    }
  }
  if (length) {
    strncpy(text, GPL_DBUF_BASE(&dbuf), length);
  }
  gpl_dbuf_free(&dbuf);
  gpl_dstr_free(&dstr);

  
  /* Record current file stream position to be able to return where
     we previously were (at least after the header) and rewind file
     before writting header. */
  file= instance->stream.io.file.handle;
  if ((offset= ftell(file)) < 0 || fseek(file, 0L, SEEK_SET)) {
    return gpl_sys_error("cannot seek to beginning of file header");
  }
  if (offset < CP40_HEADER_SIZE) {
    offset= CP40_HEADER_SIZE;
  }
  if (FileWriteString(file, ident, IDENT_SIZE)
      || FileWriteInteger(file, &instance->nframes)
      || FileWriteInteger(file, &instance->nsaturations)
      || FileWriteIntegerArray(file, instance->nphotons, NCHANNELS)
      || FileWriteReal(file, &instance->scale)
      || FileWriteReal(file, &instance->angle)
      || FileWriteRealArray(file, instance->raw_coefs, NCOEF)
      || FileWriteString(file, empty, EMPTY_SIZE)
      || FileWriteString(file, text, TEXT_SIZE)
      || fseek(file, offset, SEEK_SET)) {
    return gpl_sys_error("cannot write/seek header");
  }
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * get_version --
 *
 *	Scan identifier to get CP40 data stream version..
 *
 * Arguments:
 *	const char *ident	Identifier string, must have IDENT_SIZE bytes.
 *
 * Return:
 *	Version number, -1 in case of error (no error message left).
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

static int get_version(const char *ident)
{
  int i, c, version;
  char buffer[IDENT_SIZE+1], dummy[2];

  /* Copy identifier in buffer to make sure that it is zero terminated. */
  for (i=0; i<IDENT_SIZE && (c= ident[i]) != '\0'; ++i) {
    buffer[i]= c;
  }
  buffer[i]= '\0';

  if (sscanf(buffer, "CP40 DATA %d%1s", &version, dummy) != 1) {
    return -1;
  }
  return version;
}

/*_____________________________________________________________________________
 *
 * write_element --
 *
 *	Write list element enclosed in braces as needed and with special
 *	characters substituted by backslash sequences.
 *
 * Arguments:
 *	char     *dst	Starting address for writing element.  Must be large
 *			enough to store the maximum number of characters
 *			that can be written: 2*strlen(src)+3 (i.e. all input
 *			characters get escaped, input string is enclosed into
 *			braces plus a final '\0').
 *	const char *src	Source string (can be NULL).
 *
 * Return:
 *	Number of characters written (excluding final '\0').
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static long write_element(char *dst, const char *src)
{
  char *ptr;
  int c, needBraces;

  /* Figure out whether enclosing curly braces are
   * needed or not and compute string length as well. */
  if (src == NULL || src[0] == 0) {
    dst[0] = '{';
    dst[1] = '}';
    dst[2] = 0;
    return 2L;
  }
  needBraces = 0;
  ptr = (char *)src;
  while ((c = *ptr++) != 0) {
    if (c == ' ') {
      needBraces = 1;
      break;
    }
  }

  /* Copy string with enclosing curly braces, if needed.  Adopt a
   * conservative approach to quote all characters that have a
   * special meaning: curly braces and backslash. */
  ptr = dst;
  if (needBraces) {
    *ptr++ = '{';
  }
  while ((c = *src++) != 0) {
    switch (c) {
    case '\a':
      *ptr++ = '\\';
      *ptr++ = 'a';
      break;
    case '\b':
      *ptr++ = '\\';
      *ptr++ = 'b';
      break;
    case '\f':
      *ptr++ = '\\';
      *ptr++ = 'f';
      break;
    case '\n':
      *ptr++ = '\\';
      *ptr++ = 'n';
      break;
    case '\r':
      *ptr++ = '\\';
      *ptr++ = 'r';
      break;
    case '\t':
      *ptr++ = '\\';
      *ptr++ = 't';
      break;
    case '\v':
      *ptr++ = '\\';
      *ptr++ = 'v';
      break;
#if COMPATIBLE_WITH_TCL_LIST
    case ']':
    case '[':
    case '$':
    case ';':
    case '"':
#endif
    case '\\':
    case '{':
    case '}':
      *ptr++ = '\\';
    default:
      *ptr++ = c;
    }
  }
  if (needBraces) {
    *ptr++ = '}';
  }
  *ptr = 0;
  return ptr-dst;
}

/*---------------------------------------------------------------------------*/

static int read_element(char *dst, const char *src, char **end_ptr)
{
  int c;

  if (src != NULL) {
    /* Skip leading spaces. */
    while (GPL_IS_SPACE(*src)) {
      src++;
    }
  }
  if (src == NULL || (c= *src) == '\0') {
    *dst= 0;
    if (end_ptr != NULL) {
      *end_ptr= (char *)src;
    }
    return 0;
  } else if (c == '{') {
    int depth= 1;
    for (;;) {
      c= *++src;
      if (c == '{') {
	++depth;
      } else if (c == '}') {
	if (--depth == 0) {
	  /* End of element in braces, check the presence of a separator. */
	  if (*++src != 0 && ! GPL_IS_SPACE(*src)) {
	    gpl_error("list element in braces followed by non-space character");
	    return -1;
	  }
	  goto done;	/* Normal termination. */
	}
      } else if (c == '\\') {
	/* Escaped character, get and translate next character. */
	switch ((c= *++src)) {
	case 'a': c= '\a'; break;
	case 'b': c= '\b'; break;
	case 'f': c= '\f'; break;
	case 'n': c= '\n'; break;
	case 'r': c= '\r'; break;
	case 't': c= '\t'; break;
	case 'v': c= '\v'; break;
	case '\0':
	  gpl_error("bad null escaped character");
	  return -1;
	}
      } else if (c == '\0') {
	gpl_error("unmatched open brace");
	return -1;
      }
      *dst++= c;
    }
  } else {
    for (;;) {
      switch (c) {
      case ' ':
      case '\0':
      case '\f':
      case '\n':
      case '\r':
      case '\t':
      case '\v':
	goto done;	/* Normal termination. */
      case '\\':
	switch ((c= *++src)) {
	case 'a': c= '\a'; break;
	case 'b': c= '\b'; break;
	case 'f': c= '\f'; break;
	case 'n': c= '\n'; break;
	case 'r': c= '\r'; break;
	case 't': c= '\t'; break;
	case 'v': c= '\v'; break;
	case '\0':
	  gpl_error("bad escaped character");
	  return -1;
	}
      }
      *dst++= c;
      c= *++src;
    }
    
  }

 done:
  if (end_ptr != NULL) {
    /* Must advance to next non-space char *BEFORE* writting string
     * terminator in DST (in case SRC and DST both point to the same
     * area). */
    while (GPL_IS_SPACE(*src)) {
      src++;
    }
    *end_ptr= (char *)src;
  }
  *dst= 0;
  return 1;
}

/*---------------------------------------------------------------------------*/

static int add_text_info(gpl_dbuf_t *dbuf, const char *name,
			 const char *value, int append)
{
  unsigned size;
  char *ptr;

  /* Make sure destination buffer is large enough; maximal size of result
   * (if not append, omit the actual contents and 1st space separator):
   *     actual contents + 1 space separator + 1 opening brace
   *   + 2*strlen(name)+2 + 1 space separator + 2*strlen(value)+2
   *   + 1 closing brace + 1 null character
   * which reduce to:
   *     9 + actual contents + 2*strlen(name) + 2*strlen(value)
   * or (not append):
   *     8 + 2*strlen(name) + 2*strlen(value) */
  if (append) {
    size= 9 + GPL_DBUF_SIZE(dbuf);
  } else {
    size= 8;
    GPL_DBUF_RESET(dbuf);
  }
  if (name) {
    size+= 2*strlen(name);
  }
  if (value) {
    size+= 2*strlen(value);
  }
  if (size > GPL_DBUF_MAX_SIZE(dbuf) &&
      gpl_dbuf_resize(dbuf, size) != GPL_OK) {
    return GPL_ERROR;
  }
  ptr= (char *)GPL_DBUF_BASE(dbuf) + GPL_DBUF_SIZE(dbuf);
    
  /* Write space separator if append mode and opening brace. */
  if (append) *ptr++= ' ';
  *ptr++= '{';

  /* Write name-value list. */
  ptr+= write_element(ptr, name);
  *ptr++= ' ';
  ptr+= write_element(ptr, value);

  /* Write closing brace and end of string null character. */
  *ptr++= '}';
  *ptr= 0;
  GPL_DBUF_SIZE(dbuf)= ptr - (char *)GPL_DBUF_BASE(dbuf);
  return GPL_OK;
}
/*---------------------------------------------------------------------------*/
