/*---------------------------------------------------------------------------*/
/* APPLY WEIGHTED DISTORTION CORRECTION */

typedef struct _wght_table wght_table_t;

#ifndef WGHT_NDIV
#  define WGHT_NDIV 15 /* Number of raw pixel subdivisions along each axis. */
#endif

static wght_table_t *wght_new(cp40_instance_t *this, int channel);
static void          wght_delete(void *wght);

struct _wght_table {
  /* This structure is used  to store data for weighted distortion correction.
     In this case, each raw-event is replaced by, possibly several, sub-events
     with   different  undistorted   coordinates  and   weights.   Let  I0  be
     INDEX[raw_x + WIDTH*raw_y], then:
       DATA[3*(i0+i)+0]= W      weight   of i-th sub-event
       DATA[3*(i0+i)+1]= X      abscissa of i-th sub-event
       DATA[3*(i0+i)+2]= Y      ordinate of i-th sub-event
     for i=0...NSUB-1, with NSUB the number of sub-events given by:
       NSUB= INDEX[raw_x+WIDTH*raw_y + 1] - INDEX[raw_x+WIDTH*raw_y] */
  int  width, height; /* Raw CCD dimension must match current raw "pixel"
			 settings. */
  int  weight;        /* Sum of weights for one event. */
  short *data;        /* Correction data: position and weights. */
  int  index[1];      /* A WIDTH*HEIGHT+1 table of offsets in member array
			 DATA (*MUST* be last member, since real size is not 1
			 but WIDTH*HEIGHT+1). */
};

static int correction_weighted(cp40_frame_t *frame)
{
  static cp40_photon_t *buffer= NULL;
  static int size= 0;
  int nsub; /* number of weighted sub-pixels for the current position */
  int j, w, pending, count, limit;
  cp40_photon_t *src, *dst;
  const short *data, *wght_data;
  const int *wght_index;
  int wght_width;
  const wght_table_t *wght;
  cp40_instance_t *owner= frame->owner;

  /* Cope with interrupts. */
  if (! buffer || size<=0) {
    buffer= NULL;
    size= 0;
  }

  /* Get table of correction. */
  if ((wght= (wght_table_t *)owner->wght[frame->channel]) == NULL &&
      (wght= wght_new(owner, frame->channel)) == NULL) {
    return GPL_ERROR;
  }
  wght_data= wght->data;
  wght_index= wght->index;
  wght_width= wght->width;
  
  /* Apply correction. */
  dst= buffer;
  count= 0;
  for (src= frame->buffer, pending= frame->count ;
       pending > 0 ; ++src, --pending) {
    if ((w= src->w) <= 0) continue;
    j= src->x + wght_width * src->y;
    nsub= wght_index[j+1] - wght_index[j];
    if (nsub<=0) continue;
    data= wght_data + 3*wght_index[j];
    if ((limit= count + nsub) > size) {
      /* Resize buffer. */
      unsigned new_size= count + nsub * pending;
      cp40_photon_t *old_buf= buffer;
      if (! (buffer= gpl_alloc(new_size * sizeof(*buffer)))) {
	if (old_buf) gpl_free(old_buf);
	size= 0;
	return gpl_error("not enough memory to apply weighted correction");
      }
      if (count>0) memcpy(buffer, old_buf, count * sizeof(*buffer));
      if (old_buf) gpl_free(old_buf);
      size= new_size;
      dst= buffer;
    }
    for ( ; count<limit ; ++count, data+=3) {
      dst[count].w= w * data[0];
      dst[count].x= data[1];
      dst[count].y= data[2];
    }
  }

  /* Swap buffers -- try to be safe against interrupts, although optimizer
     will probably break this attempt ;-( */
  {
    void *old_buf=  buffer;
    int   old_size= size;
    void *new_buf=  frame->buffer;
    int   new_size= frame->size;

    size= 0;
    buffer=  NULL;
    frame->count= 0;
    frame->size= 0;
    frame->buffer= NULL;

    buffer= new_buf;
    size= new_size;
    
    frame->buffer= old_buf;
    frame->size= old_size;
    frame->count= count;
  }
  return GPL_OK;
}

static wght_table_t *wght_new(cp40_instance_t *this, int channel)
{
#define BOX_WIDTH  7  /* 2 + maximum allowed raw pixel diagonal in units of
			 undistorted pixels */
#define BOX_COUNT (BOX_WIDTH*BOX_WIDTH)

  int *index;
  double x, y; /* raw pixel coordinates and boundaries */
  double start, step;
  double cx0, cx1, cx2, cx3;
  double cy0, cy1, cy2, cy3;
  int box[BOX_COUNT]; /* used temporarily to integrate the undistorted
			 pixel weights */
  int ix0, iy0;      /* undistorted coordinates of the lower left box corner */
  int raw_x, raw_y;  /* coordinates of raw pixel (in raw pixel units) */
  wght_table_t *wght= NULL;
  int data_count= 0, data_index= 0;
  short *data;
  const int data_elsize= (3*sizeof(*data));
  int i;
  cp40_frame_t *frame= &this->frame[channel];
  int xmin= frame->xmin;
  int xmax= frame->xmax;
  int ymin= frame->ymin;
  int ymax= frame->ymax;
  int raw_width= CP40_WIDTH(this->pixel);
  int raw_height= CP40_HEIGHT(this->pixel);
  const double *coef= this->coefs + channel*(2*CP40_NCOEFS);

  if (CP40_NCOEFS != 10) {
    gpl_error("CP40_NCOEFS=%d not yet implemented", CP40_NCOEFS);
    return NULL;
  }

  /* Initialize and get number of subdivisions. */
  for (i=0 ; i<BOX_COUNT ; ++i) box[i]= 0;
  step= 1.0/(double)WGHT_NDIV; /* sub-raw-pixel step */
  start= step/2.0;     /* starting index with respect to (X0,Y0) corner */
  
  /* Allocate and instanciate wght_table_t structure. */
  if ((wght= this->wght[frame->channel]) != NULL) {
    this->wght[frame->channel]= NULL;
    wght_delete(wght);
  }
  wght= gpl_alloc(GPL_OFFSET_OF(wght_table_t, index)
		+ (raw_width*raw_height + 1)*sizeof(int));
  if (wght == NULL) goto not_enough_memory;
  index= wght->index;
  *index++= data_index;
  wght->width= raw_width;
  wght->height= raw_height;
  wght->weight= WGHT_NDIV*WGHT_NDIV;
  wght->data= data= NULL;

  /* Loop over raw pixels. */
  for (raw_y=0 ; raw_y<raw_height ; ++raw_y) {
    double y0= raw_y-0.5, y1= raw_y+0.5; /* raw pixel Y-boundaries */
    for (raw_x=0 ; raw_x<raw_width ; ++raw_x) {
      /* Get bounding box (CX0:CX1,CY0:CY1) of undistorted image of raw pixel
	 by computing undistorted coordinates of raw pixel corners (I'm too
	 lazzy to figure out how to avoid redundancy in this computation, i.e,
	 next pixel will share 2 corners with previous one unless it is at the
	 begin of a scanline; nevertheless the CPU gain would be very low). */
      double x0= raw_x-0.5, x1= raw_x+0.5; /* raw pixel X-boundaries */
      cx1= cx0= CP40_XCOR(coef,x0,y0);
      if ((x= CP40_XCOR(coef,x0,y1)) < cx0) cx0= x; else if (x > cx1) cx1= x;
      if ((x= CP40_XCOR(coef,x1,y0)) < cx0) cx0= x; else if (x > cx1) cx1= x;
      if ((x= CP40_XCOR(coef,x1,y1)) < cx0) cx0= x; else if (x > cx1) cx1= x;
      cy1= cy0= CP40_YCOR(coef,x0,y0);
      if ((y= CP40_YCOR(coef,x0,y1)) < cy0) cy0= y; else if (y > cy1) cy1= y;
      if ((y= CP40_YCOR(coef,x1,y0)) < cy0) cy0= y; else if (y > cy1) cy1= y;
      if ((y= CP40_YCOR(coef,x1,y1)) < cy0) cy0= y; else if (y > cy1) cy1= y;
      if (cx0<0.0 || cy0<0.0) {
	gpl_error("(BUG) offsets incorrectly set by cp40_update_cache()");
	goto error;
      }

      /* Check that the undistorted pixel (plus one pixel around) is enclosed
	 in the box (the check is not 100% safe but should only fail for
	 non-realistic coefficients). */
      if (/**/(int)cx1 - (int)cx0 >= BOX_WIDTH - 2
	  ||  (int)cy1 - (int)cy0 >= BOX_WIDTH - 2) {
	gpl_error("BOX_WIDTH=%d too small (try with a smaller SCALE)",
		BOX_WIDTH);
	goto error;
      }
      ix0= (int)cx0 - 1; /* x-offset plus one pixel */
      iy0= (int)cy0 - 1; /* y-offset plus one pixel */

      /* Loop over sub-raw-pixel. */
      for (y=y0+start ; y<y1 ; y+=step) {
	/* Compute factorized coefficients with offset (IX0,IY0). */
	cx0= ((coef[9]*y + coef[5])*y + coef[2])*y + coef[0] - ix0;
	cx1= (coef[8]*y + coef[4])*y + coef[1];
	cx2= coef[7]*y + coef[3];
	cx3= coef[6];
	cy0= ((coef[19]*y + coef[15])*y + coef[12])*y + coef[10] - iy0;
	cy1= (coef[18]*y + coef[14])*y + coef[11];
	cy2= coef[17]*y + coef[13];
	cy3= coef[16];

	for (x=x0+start ; x<x1 ; x+=step) {
	  box[(int)(((cx3*x + cx2)*x + cx1)*x + cx0)
	     +(int)(((cy3*x + cy2)*x + cy1)*x + cy0)*BOX_WIDTH]++;
	}
      }

      /* Re-allocate buffer by chunks if it becomes too small. */
      if (data_index+BOX_COUNT > data_count) {
	data_count+= raw_width*raw_height;
	data= gpl_alloc(data_count*data_elsize);
	if (data == NULL) goto not_enough_memory;
	if (data_index>0) memcpy(data, wght->data, data_index*data_elsize);
	if (wght->data) gpl_free(wght->data);
	wght->data= data;
      }

      /* Instanciate table and index. */
      for (i=0 ; i<BOX_COUNT ; ++i) {
	if (box[i]) {
	  int ix= i%BOX_WIDTH + ix0;
	  int iy= i/BOX_WIDTH + iy0;
	  if (ix>=xmin && ix<=xmax && iy>=ymin && iy<=ymax) {
	    data[3*data_index]= box[i];
	    data[3*data_index+1]= ix;
	    data[3*data_index+2]= iy;
	    ++data_index;
	  }
	  box[i]= 0;
	}
      }
      *index++= data_index;

    }
  }

  /* Resize table if it is too large by more than 10%
     (ignore malloc error). */ 
  if (data_count > 1.1*data_index) {
    if (data_index > 0) {
      if ((data= gpl_alloc(data_index*data_elsize)) != NULL) {
	memcpy(data, wght->data, data_index*data_elsize);
	if (wght->data) gpl_free(wght->data);
	wght->data= data;
      }
    } else {
      if (wght->data) gpl_free(wght->data);
      wght->data= NULL;
    }
    data_count= data_index;
  }
  this->wght[frame->channel]= wght;
  return wght;

 not_enough_memory:
    gpl_error("insuficient memory to store weighted correction table");
 error:
    wght_delete(wght);
    return NULL;
}

static void wght_delete(void *ptr)
{
  if (ptr) {
    GPL_FREE_MAYBE(((wght_table_t *)ptr)->data);
    gpl_free(ptr);
  }
}

/*---------------------------------------------------------------------------*/
