/*
 * gpl_file.c --
 *
 *	File utilities for GPLib library: provides management of version
 *	string, static file, socket connection, ...
 *
 * Copyright and warranty:
 *	Copyright (c) 1997-1999 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the delete Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   delete  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *_____________________________________________________________________________
 */

#include <stdio.h>
#include "gpl_base.h"

/* Private routines used in this file.  */
static long ndigits(long number);


/*_____________________________________________________________________________
 *
 * gpl_close_static_file --
 *
 *	Close file stream that is statically memorized.  
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	These routines are designed to allow interruption to occurs while the
 *	file stream is open: it will be automatically closed next time
 *	gpl_open_static_file() is called.
 *_____________________________________________________________________________
 */

void gpl_close_static_file(FILE **file_ptr)
{
  if (*file_ptr != NULL) {
    FILE *tmp= *file_ptr;
    *file_ptr= NULL;
    (void)fclose(tmp);
  }
}

/*_____________________________________________________________________________
 *
 * gpl_open_static_file --
 *
 *	Open file stream that is statically memorized.  
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	These routines are designed to allow interruption to occurs while the
 *	file stream is open: it will be automatically closed next time
 *	gpl_open_static_file() is called.
 *_____________________________________________________________________________
 */

int gpl_open_static_file(FILE **file_ptr, const char *filename,
	const char *mode)
{
  gpl_close_static_file(file_ptr);
  if ((*file_ptr= fopen(filename, mode)) == NULL) {
    return gpl_sys_error("cannot open file");
  }
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * ndigits --
 *
 *	Return number of base 10 digits needed to encode an integer (plus
 *	its minus sign if negative).
 *
 * Return:
 *	Number of digits.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

static long ndigits(long number)
{
  long n= number >= 0 ? 1 : 2;
  
  while ((number /= 10) != 0) {
    n++;
  }
  return n;
}

/*_____________________________________________________________________________
 *
 * gpl_write_file_version --
 *
 *	Write file version.
 *
 * Arguments:
 *	FILE	[in]	Output file (open for writing).
 *	NAME	[out]	Name part of identifier.
 *	MAJOR	[out]	Major number.
 *	MINOR	[out]	Minor number.
 *
 * Return:
 *	A standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	Stream pointer is advanced by GPL_IDENT_SIZE bytes (the routine
 *	do not move file pointer before writing version).
 *_____________________________________________________________________________
 */

int gpl_write_file_version(FILE *file, const char *name, int major, int minor)
{
  char ident[GPL_IDENT_SIZE];
  int i, c;

  if (major < 0 || minor < 0) {
    return gpl_error("bad version number");
  }
  for (i=0; (c= name[i]) != '\0'; i++) {
    if (c == ' ') {
      return gpl_error("name part of identifier cannot have spaces");
    }
  }
  if (i+ndigits(major)+ndigits(minor) > GPL_IDENT_SIZE-3) {
    return gpl_error("too long identifier: `%s %d.%d'",
		   name, major, minor);
  }
  for (i=0; i<GPL_IDENT_SIZE; i++) {
    ident[i]= '\0';
  }
  sprintf(ident, "%s %d.%d", name, major, minor);
  if (fwrite(ident, GPL_IDENT_SIZE, 1, file) != 1) {
    return gpl_sys_error("cannot write identifier");
  }
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * gpl_read_file_version --
 *
 *	Read file version.
 *
 * Arguments:
 *	FILE	[in]	Input file (open for reading).
 *	NAME	[out]	A character array to store the name part of
 *			identifier (at least GPL_IDENT_SIZE).
 *	MAJOR	[out]	Address where to store major number.
 *	MINOR	[out]	Address where to store minor number.
 *
 * Return:
 *	A standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	Stream pointer is advanced by GPL_IDENT_SIZE bytes (the routine
 *	do not move file pointer before reading version).
 *_____________________________________________________________________________
 */

int gpl_read_file_version(FILE *file, char *name, int *major, int *minor)
{
  char ident[GPL_IDENT_SIZE];
  unsigned int digit;
  int c, i= -1, anyDigits, number;
  
  if (fread(ident, GPL_IDENT_SIZE, 1, file) != 1) {
    return gpl_sys_error("cannot read identifier");
  }
  /* Make sure that ident is zero-terminated. */
  if (ident[GPL_IDENT_SIZE-1] != '\0') {
  badEnd:
    return gpl_error("bad identifier (not padded by zeroes)");
  }
  while ((c= ident[++i]) != '\0') {
    if (c == ' ') {
      name[i]= '\0';
      while ((c= ident[++i]) != '\0' && c == ' ') {
	/* Just skip spaces. */;
      }
      break;
    }
    name[i]= c;
  }
  if (c == '\0') {
    return gpl_error("bad identifier (no version number)");
  }
  number= 0;
  anyDigits= 0;
  while ((c= ident[i]) != '\0' && c != '.') {
    if ((digit= c - '0') > 9) {
    badMajor:
      return gpl_error("bad major number in identifier");
    }
    number= digit+10*number;
    anyDigits= 1;
    i++;
  }
  if (! anyDigits) {
    goto badMajor;
  }
  *major= number;
  if (c == '.') {
    number= 0;
    anyDigits= 0;
    while ((c= ident[++i]) != '\0') {
      if ((digit= c - '0') > 9) {
      badMinor:
	return gpl_error("bad minor number in identifier");
      }
      number= digit+10*number;
      anyDigits= 1;
    }
    if (! anyDigits) {
      goto badMinor;
    }
    *minor= number;
  } else {
    *minor= 0;
  }
  
  while (i < GPL_IDENT_SIZE) {
    if (ident[i++] != '\0') {
      goto badEnd;
    }
  }
  return GPL_OK;
}
/*---------------------------------------------------------------------------*/
