/*
 * gpl_linterp.c --
 *
 *	Linear interpolation and related routines (histogram, ...)
 *	in GPLib.
 */
#ifndef _GPL_LINTERP_C
#define _GPL_LINTERP_C 1
#include "gpl_linterp.h"
#include "gpl_macros.h"

/*---------------------------------------------------------------------------*/
/* Linear interpolation of complex arrays. */

void gpl_linterp_c(gpl_fcomplex_t yp[], const float xp[], int np,
		 const float x[], const gpl_fcomplex_t y[],
		 int stride, int n, int nafter,
		 const gpl_fcomplex_t extra[])
{
  float extra_re[2], extra_im[2];
  if (extra) {
    extra_re[0]= extra[0].re; extra_im[0]= extra[0].im;
    extra_re[1]= extra[1].re; extra_im[1]= extra[1].im;
  }
  gpl_linterp_f((float *)yp, xp, np, x, (const float *)y, stride*2,
	      n, nafter, (extra ? extra_re : NULL));
  gpl_linterp_f((float *)yp+1, xp, np, x, (const float *)y+1, stride*2,
	      n, nafter, (extra ? extra_im : NULL));
}

void gpl_linterp_z(gpl_dcomplex_t yp[], const double xp[], int np,
		 const double x[], const gpl_dcomplex_t y[],
		 int stride, int n, int nafter,
		 const gpl_dcomplex_t extra[])
{
  double extra_re[2], extra_im[2];
  if (extra) {
    extra_re[0]= extra[0].re; extra_im[0]= extra[0].im;
    extra_re[1]= extra[1].re; extra_im[1]= extra[1].im;
  }
  gpl_linterp_d((double *)yp, xp, np, x, (const double *)y, stride*2,
	      n, nafter, (extra ? extra_re : NULL));
  gpl_linterp_d((double *)yp+1, xp, np, x, (const double *)y+1, stride*2,
	      n, nafter, (extra ? extra_im : NULL));
}

/*---------------------------------------------------------------------------*/
/* DOUBLE_IS_FASTER_THAN_FLOAT can be defined to non-zero if double precision
 *                             operations are faster than (or as fast as)
 *                             single ones (beware to penalty due to
 *                             uncontrolled data alignment, e.g. on Pentium 
 *                             processors operations onto double aligned at
 *                             4-byte boundary spend 2.6 times more time).
 */
#ifndef DOUBLE_IS_FASTER_THAN_FLOAT
# define DOUBLE_IS_FASTER_THAN_FLOAT 0
#endif

#define TYPE          INT
#define HISTOGRAM     gpl_histogram_i
#include __FILE__

#define TYPE          LONG
#define HISTOGRAM     gpl_histogram_l
#include __FILE__

#define TYPE          FLOAT
#define DIGITIZE      gpl_digitize_f
#define HISTOGRAM     gpl_histogram_f
#define HISTOX        gpl_histox_f
#define HUNT          gpl_hunt_f
#define LINTERP       gpl_linterp_f
#define LXTR2D        gpl_lxtr2d_f
#include __FILE__

#define TYPE          DOUBLE
#define DIGITIZE      gpl_digitize_d
#define HISTOGRAM     gpl_histogram_d
#define HISTOX        gpl_histox_d
#define HUNT          gpl_hunt_d
#define LINTERP       gpl_linterp_d
#define LXTR2D        gpl_lxtr2d_d
#include __FILE__

#define TYPE          FCOMPLEX
#define HISTOGRAM     gpl_histogram_c
#define HISTOX        gpl_histox_c
#define HUNT          gpl_hunt_f
#define LXTR2D        gpl_lxtr2d_c
#include __FILE__

#define TYPE          DCOMPLEX
#define HISTOGRAM     gpl_histogram_z
#define HISTOX        gpl_histox_z
#define HUNT          gpl_hunt_d
#define LXTR2D        gpl_lxtr2d_z
#include __FILE__

#else /* _GPL_LINTERP_C defined -----------------------------------------------*/

#undef type_t
#undef REAL
#undef real_t
#undef ONE
#undef ZERO
#undef HALF

/* Define some type dependent macros:
     type_t       is the current C-type for the data;
     real_t       is the `best' floating point type (same as real part of
                  type_t for floating point or complex data);
*/

#define type_t     GPL_CTYPE_OF(TYPE)

#if GPL_IS_INTEGER(TYPE)
# if DOUBLE_IS_FASTER_THAN_FLOAT
#  define REAL        FLOAT
# else
#  define REAL        DOUBLE
# endif
#else /* ! GPL_IS_INTEGER(TYPE) */
# define REAL         GPL_REAL_PART_OF(TYPE)
#endif

#if GPL_IS_DOUBLE(TYPE)
# define real_t double
# define ONE    1.0
# define ZERO   0.0
# define HALF   0.5
#else
# define real_t float
# define ONE    1.0F
# define ZERO   0.0F
# define HALF   0.5F
#endif

/*---------------------------------------------------------------------------*/

#if defined(HUNT) && ! GPL_IS_COMPLEX(TYPE)
int HUNT(type_t xp, const type_t x[], int n, int ip)
{
  int ascend= x[n-1]>x[0];
  int jl, ju;
  
  if (ip<=0 || ip>=n) {
    /* Caller has declined to make an initial guess, so fall back to
       garden variety bisection method. */
    if ((xp>=x[n-1]) == ascend) return n;
    if ((xp<x[0]) == ascend) return 0;
    jl= 0;
    ju= n-1;

  } else {
    /* Search from initial guess IP in ever increasing steps to bracket XP. */
    int inc= 1;
    jl= ip;
    if ((xp>=x[ip]) == ascend) {
      /* search toward larger index values */
      if (ip==n-1) return n;
      jl= ip;
      ju= ip+inc;
      while ((xp>=x[ju]) == ascend) {
	jl= ju;
	inc+= inc;
	ju+= inc;
	if (ju>=n) {
	  if ((xp>=x[n-1]) == ascend) return n;
	  ju= n;
	  break;
	}
      }
    } else {
      /* search toward smaller index values */
      if (ip==0) return 0;
      ju= ip;
      jl= ip-inc;
      while ((xp<x[jl]) == ascend) {
	ju= jl;
	inc+= inc;
	jl-= inc;
	if (jl<0) {
	  if ((xp<x[0]) == ascend) return 0;
	  jl= 0;
	  break;
	}
      }
    }
  }

  /* Have X[JL]<=XP<X[JU] for ascend, X[JL]>=XP>X[JU] for !ascend */
  while (ju-jl > 1) {
    ip= (jl+ju)>>1;
    if ((xp>=x[ip]) == ascend) jl= ip;
    else ju= ip;
  }

  return ju;
}
#endif /* HUNT */

/*---------------------------------------------------------------------------*/
#ifdef LINTERP
void LINTERP(type_t yp[], const type_t xp[], int np,
	     const type_t x[], const type_t y[],
	     int stride, int n, int nafter, const type_t extra[])
{
  /* The problem is as follows:
     For each value of the faster indices, and each value of the slower
     indices of Y, and for each value of XP, compute the value of YP
     corresponding to XP by linear interpolation.  Note that for each
     interpolation, the Y vector has stride STRIDE, while the X vector
     always has stride 1. Note also that ALL of the YP for a given XP
     should be computed once XP is found.  */

#if 0  /* Non-optimized version (intentionally left there because the use
	  of strides is easier to follow than in the optimized version).
	  Warning the EXTRA argument is not accounted in this unused part
	  of the code. */
  int ip, j, k;
  int i= 0;  /* hunt takes this as no guess on 1st pass */
  for (ip=0 ; ip<np ; ip++) {
    i= HUNT(xp[ip], x, n, i);
    if (i<1) {
      /* point before first X */
      int ii= 0;
      for (j=0 ; j<nafter ; j++) {
	for (k=0 ; k<stride ; k++) {
	  yp[(ip+np*j)*stride + k]= y[(ii+n*j)*stride + k];
	}
      }
    } else if (i<n) {
      /* point in range */
      type_t c0= (x[i]-xp[ip])/(x[i]-x[i-1]);
      type_t c1= ONE-c0;
      for (j=0 ; j<nafter ; j++) {
	for (k=0 ; k<stride ; k++) {
	  yp[(ip+np*j)*stride + k]=
	    c0 * y[((i-1)+n*j)*stride + k] + c1 * y[(i+n*j)*stride + k];
	}
      }
    } else {
      /* point above last X */
      int ii= n-1;
      for (j=0 ; j<nafter ; j++) {
	for (k=0 ; k<stride ; k++) {
	  yp[(ip+np*j)*stride + k]= y[(ii+n*j)*stride + k];
	}
      }
    }
  }
#else
  int ip, k, j0, j1, jp;
  int i= 0;  /* hunt takes this as no guess on 1st pass */
  if (stride==1) {
    /* Unit stride. */
    int nyp= np*nafter;
    for (ip=0 ; ip<np ; ip++) {
      i= HUNT(xp[ip], x, n, i);
      j1= i;
      j0= j1-1;
      if (i<1) {
	/* point before first X */
	if (extra) {
	  type_t value= extra[0];
	  for (jp=ip*1 ; jp<nyp ; jp+=np) {
	    yp[jp]= value;
	  }
	} else {
	  for (jp=ip*1 ; jp<nyp ; jp+=np, j1+=n) {
	    yp[jp]= y[j1];
	  }
	}
      } else if (i<n) {
	/* point in range */
	type_t c0= (x[i]-xp[ip])/(x[i]-x[i-1]);
	type_t c1= ONE-c0;
	for (jp=ip*1 ; jp<nyp ; jp+=np, j1+=n, j0+=n) {
	  yp[jp]= c0 * y[j0] + c1 * y[j1];
	}
      } else {
	/* point above last X */
	if (extra) {
	  type_t value= extra[1];
	  for (jp=ip*1 ; jp<nyp ; jp+=np) {
	    yp[jp]= value;
	  }
	} else {
	  for (jp=ip*1 ; jp<nyp ; jp+=np, j0+=n) {
	    yp[jp]= y[j0];
	  }
	}
      }
    }
  } else {
    /* Non-unit stride. */
    int nyp= np*stride*nafter;
    int s= stride*n, sp= stride*np;
    for (ip=0 ; ip<np ; ip++) {
      i= HUNT(xp[ip], x, n, i);
      j1= i*stride;
      j0= j1-stride;
      if (i<1) {
	/* point before first X */
	if (extra) {
	  type_t value= extra[0];
	  for (jp=ip*stride ; jp<nyp ; jp+=sp) {
	    for (k=0 ; k<stride ; k++) {
	      yp[jp + k]= value;
	    }
	  }
	} else {
	  for (jp=ip*stride ; jp<nyp ; jp+=sp, j1+=s) {
	    for (k=0 ; k<stride ; k++) {
	      yp[jp + k]= y[j1 + k];
	    }
	  }
	}
      } else if (i<n) {
	/* point in range */
	type_t c0= (x[i]-xp[ip])/(x[i]-x[i-1]);
	type_t c1= ONE-c0;
	for (jp=ip*stride ; jp<nyp ; jp+=sp, j1+=s, j0+=s) {
	  for (k=0 ; k<stride ; k++) {
	    yp[jp + k]= c0 * y[j0 + k] + c1 * y[j1 + k];
	  }
	}
      } else {
	/* point above last X */
	if (extra) {
	  type_t value= extra[1];
	  for (jp=ip*stride ; jp<nyp ; jp+=sp) {
	    for (k=0 ; k<stride ; k++) {
	      yp[jp + k]= value;
	    }
	  }
	} else {
	  for (jp=ip*stride ; jp<nyp ; jp+=sp, j0+=s) {
	    for (k=0 ; k<stride ; k++) {
	      yp[jp + k]= y[j0 + k];
	    }
	  }
	}
      }
    }
  }
#endif
}
#endif /* LINTERP */
/*---------------------------------------------------------------------------*/
#ifdef DIGITIZE
void DIGITIZE(int idx[], const type_t data[], int n,
	      const type_t x[], int nx)
{
  int i, ix= 0; /* hunt takes this as no guess on 1st pass */
  for (i=0 ; i<n ; i++) {
    ix= HUNT(data[i], x, nx, ix);
    idx[i]= ix;
  }
}
#endif /* DIGITIZE */

#ifdef HISTOGRAM
void HISTOGRAM(type_t h[], int n, const int idx[], const type_t wght[], int np)
{
  int i, ip;

  GPL_BZERO(h, n);
  if (! wght) {
    for (ip=0 ; ip<np ; ip++) {
      i= idx[ip];
      if (i>=0 && i<n) {
#if GPL_IS_COMPLEX(TYPE)
	h[i].re++;
#else
	h[i]++;
#endif
      }
    }
  } else {
    for (ip=0 ; ip<np ; ip++) {
      i= idx[ip];
      if (i>=0 && i<n) {
#if GPL_IS_COMPLEX(TYPE)
	h[i].re+= wght[ip].re;
	h[i].im+= wght[ip].im;
#else
	h[i]+= wght[ip];
#endif
      }
    }
  }
}
#endif /* HISTOGRAM */

#ifdef HISTOX
void HISTOX(type_t y[], const real_t x[], int n,
	    const real_t xp[], const type_t wp[], int np,
	    int interpolate)
{
  int i, ip;
  real_t c;

  GPL_BZERO(y, n+1);

  i= 0;  /* hunt takes this as no guess on 1st pass */
  if (! wp) {
    if (! interpolate) {
      for (ip=0 ; ip<np ; ip++) {
	i= HUNT(xp[ip], x, n, i);
#if GPL_IS_COMPLEX(TYPE)
	y[i].re++;
#else
	y[i]++;
#endif
      }
    } else {
      for (ip=0 ; ip<np ; ip++) {
	i= HUNT(xp[ip], x, n, i);
#if GPL_IS_COMPLEX(TYPE)
	if (i<1) {
	  y[0].re++;
	} else if (i<n) {
	  c= (x[i]-xp[ip])/(x[i]-x[i-1]);
	  y[i-1].re+= c;
	  y[i].re+= ONE-c;
	} else {
	  y[n].re++;
	}
#else
	if (i<1) {
	  y[0]++;
	} else if (i<n) {
	  c= (x[i]-xp[ip])/(x[i]-x[i-1]);
	  y[i-1]+= c;
	  y[i]+= ONE-c;
	} else {
	  y[n]++;
	}
#endif
      }
    }
  } else {
    if (! interpolate) {
      for (ip=0 ; ip<np ; ip++) {
	i= HUNT(xp[ip], x, n, i);
#if GPL_IS_COMPLEX(TYPE)
	y[i].re+= wp[ip].re;
	y[i].im+= wp[ip].im;
#else
	y[i]+= wp[ip];
#endif
      }
    } else {
      for (ip=0 ; ip<np ; ip++) {
	i= HUNT(xp[ip], x, n, i);
#if GPL_IS_COMPLEX(TYPE)
	if (i<1) {
	  y[0].re+= wp[ip].re;
	  y[0].im+= wp[ip].im;
	} else if (i<n) {
	  c= (x[i]-xp[ip])/(x[i]-x[i-1]);
	  y[i-1].re+= c*wp[ip].re;
	  y[i-1].im+= c*wp[ip].im;
	  c= ONE-c;
	  y[i].re+= c*wp[ip].re;
	  y[i].im+= c*wp[ip].im;
	} else {
	  y[n].re+= wp[ip].re;
	  y[n].im+= wp[ip].im;
	}
#else
	if (i<1) {
	  y[0]+= wp[ip];
	} else if (i<n) {
	  c= (x[i]-xp[ip])/(x[i]-x[i-1]);
	  y[i-1]+= c*wp[ip];
	  y[i]+= (ONE-c)*wp[i];
	} else {
	  y[n]+= wp[ip];
	}
#endif
      }
    }
  }
}
#endif /* HISTOX */

/*---------------------------------------------------------------------------*/
#ifdef LXTR2D
void LXTR2D(type_t zp[], int n1p, int n2p, const type_t z[], int n1, int n2,
	    const real_t tr[], type_t out)
{
#undef z_
#define z_(a1,a2)     z[(a1)+n1*(a2)]
#undef zp_
#define zp_(a1,a2)    zp[(a1)+n1p*(a2)]
  int i1, i2, i1p, i2p, cs;
  real_t x1, x2;

  for (i2p=0 ; i2p<n2p ; ++i2p) {
    for (i1p=0 ; i1p<n1p ; ++i1p) {
      /* 1st dimension. */
      x1= tr[0] + tr[1]*i1p + tr[2]*i2p;
      if (x1<ZERO) {
	if (x1<=-HALF) {
	outside:
#if GPL_IS_COMPLEX(TYPE)
	  zp_(i1p,i2p).re= out.re;
	  zp_(i1p,i2p).im= out.im;
#else
	  zp_(i1p,i2p)= out;
#endif
	  continue;
	}
	i1= -1;
	cs= 1;
      } else if (x1>=(n1-1)) {
	if (x1>=(n1-HALF)) goto outside;
	i1= n1-1;
	cs= 2;
      } else {
	i1= (int)x1;
	cs= 0;
      }

      /* 2nd dimension. */
      x2= tr[3] + tr[4]*i1p + tr[5]*i2p;
      if (x2<ZERO) {
	if (x2<=-HALF) goto outside;
	i2= -1;
	cs+= 3;
      } else if (x2>=(n2-1)) {
	if (x2>=(n2-HALF)) goto outside;
	i2= n2-1;
	cs+= 6;
      } else {
	i2= (int)x2;
      }
      
      /* Two dimensions together. */
      x1-= i1;
      x2-= i2;
      switch (cs) {
      case 0:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= (ONE-x1)*((ONE-x2)*z_(i1,i2).re + x2*z_(i1,i2+1).re)
	  + x1*((ONE-x2)*z_(i1+1,i2).re + x2*z_(i1+1,i2+1).re);
	zp_(i1p,i2p).im= (ONE-x1)*((ONE-x2)*z_(i1,i2).im + x2*z_(i1,i2+1).im)
	  + x1*((ONE-x2)*z_(i1+1,i2).im + x2*z_(i1+1,i2+1).im);
#else
	zp_(i1p,i2p)= (ONE-x1)*((ONE-x2)*z_(i1,i2) + x2*z_(i1,i2+1))
	  + x1*((ONE-x2)*z_(i1+1,i2) + x2*z_(i1+1,i2+1));
#endif
	break;
      case 1:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= (ONE-x1)*out.re
	  + x1*((ONE-x2)*z_(0,i2).re + x2*z_(0,i2+1).re);
	zp_(i1p,i2p).im= (ONE-x1)*out.im
	  + x1*((ONE-x2)*z_(0,i2).im + x2*z_(0,i2+1).im);
#else
	zp_(i1p,i2p)= (ONE-x1)*out + x1*((ONE-x2)*z_(0,i2) + x2*z_(0,i2+1));
#endif
	break;
      case 2:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= (ONE-x1)*((ONE-x2)*z_(n1-1,i2).re
				   + x2*z_(n1-1,i2+1).re) + x1*out.re;
	zp_(i1p,i2p).im= (ONE-x1)*((ONE-x2)*z_(n1-1,i2).im
				   + x2*z_(n1-1,i2+1).im) + x1*out.im;
#else
	zp_(i1p,i2p)= (ONE-x1)*((ONE-x2)*z_(n1-1,i2) + x2*z_(n1-1,i2+1))
	  + x1*out;
#endif
	break;
      case 3:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= x2*((ONE-x1)*z_(i1,0).re + x1*z_(i1+1,0).re)
	  + (ONE-x2)*out.re;
	zp_(i1p,i2p).im= x2*((ONE-x1)*z_(i1,0).im + x1*z_(i1+1,0).im)
	  + (ONE-x2)*out.im;
#else
	zp_(i1p,i2p)= x2*((ONE-x1)*z_(i1,0) + x1*z_(i1+1,0)) + (ONE-x2)*out;
#endif
	break;
      case 4:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= x1*x2*(z_(0,0).re - out.re) + out.re;
	zp_(i1p,i2p).im= x1*x2*(z_(0,0).im - out.im) + out.im;
#else
	zp_(i1p,i2p)= x1*x2*(z_(0,0) - out) + out;
#endif
	break;
      case 5:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= (ONE-x1)*x2*(z_(n1-1,0).re - out.re) + out.re;
	zp_(i1p,i2p).im= (ONE-x1)*x2*(z_(n1-1,0).im - out.im) + out.im;
#else
	zp_(i1p,i2p)= (ONE-x1)*x2*(z_(n1-1,0) - out) + out;
#endif
	break;
      case 6:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= (ONE-x1)*(ONE-x2)*z_(i1,i2).re
	  + x1*(ONE-x2)*z_(i1+1,i2).re + x2*out.re;
	zp_(i1p,i2p).im= (ONE-x1)*(ONE-x2)*z_(i1,i2).im
	  + x1*(ONE-x2)*z_(i1+1,i2).im + x2*out.im;
#else
	zp_(i1p,i2p)= (ONE-x1)*(ONE-x2)*z_(i1,i2) + x1*(ONE-x2)*z_(i1+1,i2)
	  + x2*out;
#endif
	break;
      case 7:
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= x1*(ONE-x2)*z_(0,i2).re + (ONE-x1*(ONE-x2))*out.re;
	zp_(i1p,i2p).im= x1*(ONE-x2)*z_(0,i2).im + (ONE-x1*(ONE-x2))*out.im;
#else
	zp_(i1p,i2p)= x1*(ONE-x2)*z_(0,i2) + (ONE-x1*(ONE-x2))*out;
#endif
	break;
      default: /*case 8:*/
#if GPL_IS_COMPLEX(TYPE)
	zp_(i1p,i2p).re= (ONE-x1)*(ONE-x2)*z_(n1-1,i2).re
	  + (x1+x2-x1*x2)*out.re;
	zp_(i1p,i2p).im= (ONE-x1)*(ONE-x2)*z_(n1-1,i2).im
	  + (x1+x2-x1*x2)*out.im;
#else
	zp_(i1p,i2p)= (ONE-x1)*(ONE-x2)*z_(n1-1,i2) + (x1+x2-x1*x2)*out;
#endif
	break;
      }
    }
  }
}
#endif /* LXTR2D */

/*---------------------------------------------------------------------------*/
#define GPL_REBIN_CLOSEST
#define GPL_REBIN_INTERP
#define GPL_REBIN_MIN
#define GPL_REBIN_MAX
#define GPL_REBIN_AVG
#define GPL_REBIN_MED

/*
 * coordinate:   -0.5                     N-0.5
 *                 |---+---+- ... -+---+---|
 * pixel number:     0   1          N-2 N-1
 *
 * =>  (XP+0.5)*N = (X+0.5)*NP
 */
#ifdef REBIN2D
void REBIN2D(type_t zp[], ldzp, const type_t z[], int width, int height,
	     int zoom, int method)
{
#undef z_
#define z_(a1,a2)     z[(a1)+width*(a2)]
#undef zp_
#define zp_(a1,a2)    zp[(a1)+ldzp*(a2)]
  real_t x, xp;

  if (zoom>1) {
    /* Enlarge Z. */
    for (y=0 ; y<height ; y++) {
      for (j=0 ; j<zoom ; j++) {
	for (x=0 ; x<width ; x++) {
	  value= z_(x,y);
	  for (i=0 ; i<zoom ; i++) {
	    *z++= value;
	  }
	}
      }
    }
  } else if (zoom<-1) {
    /* Schrink Z. */
  } else if (zoom!=0) {
    if (zp!=z) memcpy(zp, z, width*height*sizeof(type_t));
  }

  for (i2p=0 ; i2p<n2p ; ++i2p) {
    x2= (i2p+HALF)*(real_t)n2/(real_t)n2p - HALF;

    for (i1p=0 ; i1p<n1p ; ++i1p) {
    }
  }
}
#endif

/*---------------------------------------------------------------------------*/
/* Cleanup to allow for re-include of this file. */
#undef TYPE
#undef REAL
#undef ZERO
#undef ONE
#undef HALF
#undef LINTERP
#undef LXTR2D
#undef HUNT
#undef HISTOGRAM
#undef HISTOX
#undef DIGITIZE
#endif /* _GPL_LINTERP_C ------------------------------------------------------*/
