/*
 * gpl_linterp.h --
 *
 *	Definitions for linear interpolation and related routines (histogram,
 *	...) in GPLib library.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *      $Id$
 *      $Log$
 *-----------------------------------------------------------------------------
 */
#ifndef _GPL_LINTERP_H
#define _GPL_LINTERP_H 1
#include "gpl_base.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * gpl_linterp_{f,d,c,z} --
 *
 *   Linear interpolation along one dimension.
 *
 * Description:
 *
 *   For mono-dimensional arrays Y and YP, the routine compute YP such that
 *   (XP[ip], YP[ip]) (ip=0..NP-1) lies on the piecewise linear curve (X[i],
 *   Y[i]) (i=0..N-1).  Points beyond X[0] are set to Y[0] (EXTRA[0] if EXTRA
 *   is not NULL); points beyond X[N-1] are set to Y[N-1] (EXTRA[1] if EXTRA
 *   is not NULL).  The array X must be one dimensional, have N>=2 elements,
 *   and be either monotonically increasing or monotonically decreasing.
 *
 *   The array Y may have more than one dimension, in which case a single
 *   dimension of length N is considered and interpolated to have length NP
 *   in YP.  For complex data it is sufficient to simply consider real data
 *   with an additional first dimension of lenght 2 (actually this is what is
 *   done by the complex routines).
 *
 * Arguments:
 *
 *   YP      [output, NP*STRIDE*NAFTER elements]
 *           Interpolated array along dimension of interest.
 *
 *   XP      [input, NP elements]
 *           Array of interpolation positions.
 *
 *   NP      [input]
 *           Number of interpolation positions.
 *
 *   X       [input, N elements]
 *           Monotone array of coordinates along dimension to interpolate.
 *
 *   Y       [input, N*STRIDE*NAFTER elements]
 *           Array to interpolate (Y has N*STRIDE*NAFTER elements).
 *
 *   STRIDE  [input]
 *           The stride of the dimension of Y to be interpolated (i.e. the
 *           spacing of elements along the dimension of interest), e.g.:
 *               STRIDE=1 to interpolate the first dimension of Y;
 *               STRIDE=LEN1 where LEN1 is the length of the first dimension
 *                   to interpolated the second dimension;
 *               STRIDE=LEN1*LEN2 interpolates the third dimension, and so on.
 *
 *   N       [input, N>=2]
 *           The length of the dimension to interpolate.
 *
 *   NAFTER  [input]
 *           The product of all dimension lengths after the dimension to be
 *           interpolated -- the interpolation of length N will be repeated
 *           a total of STRIDE*NAFTER times.
 *
 *   EXTRA   [input, NULL or 2 elements]
 *           Extrapolation values: points beyond X[0] are set to EXTRA[0] and
 *           points above X[N-1] are set to EXTRA[1].  If EXTRA is NULL,
 *           points beyond X[0] are set to Y[0] and points above X[N-1] are
 *           set to Y[N-1].
 */
extern void gpl_linterp_f(float yp[], const float xp[], int np,
			const float x[], const float y[], int stride,
			int n, int nafter, const float extra[]);
extern void gpl_linterp_d(double yp[], const double xp[], int np,
			const double x[], const double y[], int stride,
			int n, int nafter, const double extra[]);
extern void gpl_linterp_c(gpl_fcomplex_t yp[], const float xp[], int np,
			const float x[], const gpl_fcomplex_t y[],
			int stride, int n, int nafter,
			const gpl_fcomplex_t extra[]);
extern void gpl_linterp_z(gpl_dcomplex_t yp[], const double xp[], int np,
			const double x[], const gpl_dcomplex_t y[],
			int stride, int n, int nafter,
			const gpl_dcomplex_t extra[]);

/*---------------------------------------------------------------------------*/

extern int gpl_hunt_f(float xp, const float x[], int n, int ip);
extern int gpl_hunt_d(double xp, const double x[], int n, int ip);
/*	Locate interval in monotone array X that contains value XP.
	Directly based on the hunt routine from yorick-1.4 by D. Munro
	(ftp://ftp-icf.llnl.gov/pub/Yorick).

	Here, X[N] is a monotone array and, if XP lies in the interval
	from X[0] to X[N-1], then the routine returns index I such that:
	  X[I-1] <= XP < X[I]  (as X is ascending)
	  X[I-1] >= XP > X[I], (as X is descending)
	The value 0 or N will be returned if XP lies outside the interval.

	To improve searchs of nearby values, IP can be set to the previous
	index returned by `hunt'.  Otherwise, IP should be <=0 or >=N */

/*---------------------------------------------------------------------------*/
/* HISTOGRAMS */

extern void gpl_digitize_f(int idx[], const float data[], int n,
			 const float x[], int nx);
extern void gpl_digitize_d(int idx[], const double data[], int n,
			 const double x[], int nx);
/*	Digitize DATA (N elements) into indices to discrete bins:
	    IDX[i]= j  (i=0..N-1)
        and with j such that
            X[j-1] <= DATA[i] < X[j] if X is monotonically increasing, or
            X[j-1] > DATA[i] >= X[j] if X is monotonically decreasing.
        Beyond the bounds of X, store either j=0 or j=NX as appropriate. */

/*
 * gpl_histogram_{i,l,f,d} --
 *
 *   Integrate histogram of digitized data.
 *
 * Arguments:
 *
 *   H     [output, N elements]
 *	   Resulting histogram: H[i]= sum(WP[ip]) such that IDX[ip]=i.
 *
 *   N     [input]
 *	   Number of elements of H.
 *
 *   IDX   [input, NP elements]
 *	   Index of digitized data; IDX[ip]<0 and IDX[ip]>=N
 *         (ip=0..NP-1) are ignored.
 *
 *   WGHT  [optional input, NP elements or NULL]
 *	   Weight to sum, if WP=NULL all weights are taken to be one.
 *
 *   NP    [input]
 *	   Number of elements of IDX and WP.
 *
 * See also:
 *   gpl_histox
 *
 */
extern void gpl_histogram_i(int h[], int n, const int idx[],
			  const int wght[], int np);
extern void gpl_histogram_l(long h[], int n, const int idx[],
			  const long wght[], int np);
extern void gpl_histogram_f(float h[], int n, const int idx[],
			  const float wght[], int np);
extern void gpl_histogram_d(double h[], int n, const int idx[],
			  const double wght[], int np);
extern void gpl_histogram_c(gpl_fcomplex_t h[], int n, const int idx[],
			  const gpl_fcomplex_t wght[], int np);
extern void gpl_histogram_z(gpl_dcomplex_t h[], int n, const int idx[],
			  const gpl_dcomplex_t wght[], int np);



/*
 * gpl_histox_{f,d,c,z} --
 *
 *	Compute complex histogram of data vs. some explicative variable.
 *
 * Arguments:
 *
 *   Y       [output, N+1 elements]
 *           Resulting histogram (see below).
 *
 *   X       [input, N elements]
 *           Monotone array of bin boundaries.
 *
 *   N       [input, N>=2]
 *           Number of values in X.
 *
 *   XP      [input, N>=2]
 *           Explicative variable to compare with bins, has NP elements.
 *
 *   WP      [optional input, NP elements or NULL]
 *           Weights (if NULL same as with all weights equal to one).
 *
 * Warning:
 *   Output array Y has N+1 elements (not N).
 */
extern void gpl_histox_f(float y[], const float x[], int n,
		       const float xp[], const float wp[],
		       int np, int interpolate);
extern void gpl_histox_d(double y[], const double x[], int n,
		       const double xp[], const double wp[],
		       int np, int interpolate);
extern void gpl_histox_c(gpl_fcomplex_t y[], const float x[], int n,
		       const float xp[], const gpl_fcomplex_t wp[],
		       int np, int interpolate);
extern void gpl_histox_z(gpl_dcomplex_t y[], const double x[], int n,
		       const double xp[], const gpl_dcomplex_t wp[],
		       int np, int interpolate);

/*---------------------------------------------------------------------------*/
/* LINEAR 2D TRANSFORM */
extern void gpl_lxtr2d_f(float zp[], int n1p, int n2p,
		       const float z[], int n1, int n2,
		       const float tr[], float out);
extern void gpl_lxtr2d_d(double zp[], int n1p, int n2p,
		       const double z[], int n1, int n2,
		       const double tr[], double out);
extern void gpl_lxtr2d_c(gpl_fcomplex_t zp[], int n1p, int n2p,
		       const gpl_fcomplex_t z[], int n1, int n2,
		       const float tr[], gpl_fcomplex_t out);
extern void gpl_lxtr2d_z(gpl_dcomplex_t zp[], int n1p, int n2p,
		       const gpl_dcomplex_t z[], int n1, int n2,
		       const double tr[], gpl_dcomplex_t out);
/*	Apply a 2D linear transform of coordinates to a 2D array (an image).
	Array TR (6 elements) define the linear coordinate transformation:
	  X= TR[0] + TR[1]*XP + TR[2]*YP;
	  Y= TR[3] + TR[4]*XP + TR[5]*YP;
	where (X,Y) are the pixel coordinates in the input N1-by-N2 array Z
	and (XP,YP) are the pixel coordinates in the output N1P-by-N2P array
	ZP.  To fill ZP, the (XP,YP) coordinates are varied as: XP=0..N1P-1
	and YP=0..N2P-1, then TR is used to localize the corresponding
	(floating point) position (X,Y) in Z and bi-linear interpolation is
	used to compute ZP(XP,YP).  It is assumed that Z(X,Y)=OUT for
	`outside' locations (i.e. X<0 or X>=N1 or Y<0 or Y>=N2). */

/*---------------------------------------------------------------------------*/
#ifdef __cplusplus
} /* C++ */
#endif
#endif /* _GPL_LINTERP_H ------------------------------------------------------*/
