/*
 * gpl_num_svd.c --
 *
 *	Singular Value Decomposition routines in GPLib.
 *	Original FORTRAN routine for SSVDC is from SLATEC and LINPACK.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#ifndef _GPL_NUM_SVD_C
#define _GPL_NUM_SVD_C 1
#include <math.h>
#include "gpl_num.h"

#define GPL_NUM_SINGLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_SINGLE_PRECISION

#define GPL_NUM_DOUBLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_DOUBLE_PRECISION

#else /* _GPL_NUM_SVD_C defined -----------------------------------------------*/

/*
 ****BEGIN PROLOGUE  SSVDC
 ****PURPOSE  Perform the singular value decomposition of a rectangular
 *	     matrix.
 ****LIBRARY   SLATEC (LINPACK)
 ****CATEGORY  D6
 ****TYPE      SINGLE PRECISION (SSVDC-S, DSVDC-D, CSVDC-C)
 ****KEYWORDS  LINEAR ALGEBRA, LINPACK, MATRIX,
 *	      SINGULAR VALUE DECOMPOSITION
 ****AUTHOR  Stewart, G. W., (U. of Maryland)
 ****DESCRIPTION
 *
 *    SSVDC is a subroutine to reduce a real M by N matrix X by orthogonal
 *    transformations U and V to diagonal form.  The elements S(I) are
 *    the singular values of X.  The columns of U are the corresponding
 *    left singular vectors, and the columns of V the right singular
 *    vectors.
 *
 *    On Entry
 *
 *	  X         REAL(LDX,N), where LDX .GE. M.
 *		    X contains the matrix whose singular value
 *		    decomposition is to be computed.  X is
 *		    destroyed by SSVDC.
 *
 *	  LDX       INTEGER
 *		    LDX is the leading dimension of the array X.
 *
 *	  M         INTEGER
 *		    M is the number of rows of the matrix X.
 *
 *	  N         INTEGER
 *		    N is the number of columns of the matrix X.
 *
 *	  LDU       INTEGER
 *		    LDU is the leading dimension of the array U (see below).
 *
 *	  LDV       INTEGER
 *		    LDV is the leading dimension of the array V (see below).
 *
 *	  WORK      REAL(M)
 *		    work is a scratch array.
 *
 *	  JOB       INTEGER
 *		    JOB controls the computation of the singular
 *		    vectors.  It has the decimal expansion AB
 *		    with the following meaning
 *
 *			 A .EQ. 0  Do not compute the left singular vectors.
 *			 A .EQ. 1  Return the M left singular vectors in U.
 *			 A .GE. 2  Return the first MIN(M,N) singular vectors
 *                                 in U.
 *			 B .EQ. 0  Do not compute the right singular vectors.
 *			 B .NE. 0  Return the right singular vectors in V.
 *
 *    On Return
 *
 *	  S         REAL(MN), where MN=MIN(M+1,N).
 *		    The first MIN(M,N) entries of S contain the
 *		    singular values of X arranged in descending
 *		    order of magnitude.
 *
 *	  E         REAL(N).
 *		    E ordinarily contains zeros.  However, see the
 *		    discussion of INFO for exceptions.
 *
 *	  U         REAL(LDU,K), where LDU .GE. M.  If JOBA .EQ. 1, then
 *				    K .EQ. M.  If JOBA .GE. 2 , then
 *				    K .EQ. MIN(M,N).
 *		    U contains the matrix of right singular vectors.
 *		    U is not referenced if JOBA .EQ. 0.  If M .LE. N
 *		    or if JOBA .EQ. 2, then U may be identified with X
 *		    in the subroutine call.
 *
 *	  V         REAL(LDV,N), where LDV .GE. N.
 *		    V contains the matrix of right singular vectors.
 *		    V is not referenced if JOBB .EQ. 0.  If N .LE. M,
 *		    then V may be identified with X in the
 *		    subroutine call.
 *
 *	  INFO      INTEGER.
 *		    the singular values (and their corresponding
 *		    singular vectors) S(INFO+1),S(INFO+2),...,S(M)
 *		    are correct (here M=MIN(M,N)).  Thus if
 *		    INFO .EQ. 0, all the singular values and their
 *		    vectors are correct.  In any event, the matrix
 *		    B = TRANS(U)*X*V is the bidiagonal matrix
 *		    with the elements of S on its diagonal and the
 *		    elements of E on its super-diagonal (TRANS(U)
 *		    is the transpose of U).  Thus the singular
 *		    values of X and B are the same.
 *
 ****REFERENCES  J. J. Dongarra, J. R. Bunch, C. B. Moler, and G. W.
 *		  Stewart, LINPACK Users' Guide, SIAM, 1979.
 ****ROUTINES CALLED  SAXPY, SDOT, SNRM2, SROT, SROTG, SSCAL, SSWAP
 ****REVISION HISTORY  (YYMMDD)
 *  790319  DATE WRITTEN
 *  890531  Changed all specific intrinsics to generic.  (WRB)
 *  890531  REVISION DATE from Version 3.2
 *  891214  Prologue converted to Version 4.0 format.  (BAB)
 *  900326  Removed duplicate information from DESCRIPTION section.
 *	    (WRB)
 *  920501  Reformatted the REFERENCES section.  (WRB)
 ****END PROLOGUE  SSVDC
 */

void SSVDC(real_t *x, integer_t ldx, integer_t m, integer_t n, real_t *s,
	   real_t *e, real_t *u, integer_t ldu, real_t *v, integer_t ldv,
	   real_t *work, integer_t job, integer_t *info)
{
  integer_t kase, iter;
  real_t test1, test2;
  integer_t nctp1;
  real_t b, c;
  integer_t nrtp1;
  real_t f, g;
  integer_t i, j, k, l, p;
  real_t t, scale;
  real_t shift;
  integer_t maxit;
  integer_t wantu, wantv;
  real_t el;
  real_t cs;
  integer_t mm, ls;
  real_t sl;
  integer_t lu;
  real_t sm, sn;
  integer_t lp1, nct, ncu, nrt;
  real_t emm1, smm1;

  /* Parameter adjustments to mimic FORTRAN indexing. */
#define u_(a1,a2) u[(a2)*ldu + (a1)]
#define v_(a1,a2) v[(a2)*ldv + (a1)]
#define x_(a1,a2) x[(a2)*ldx + (a1)]
  x-= 1+ldx;
  u-= 1+ldu;
  v-= 1+ldv;
  --s;
  --e;
  --work;
  
  /* Set the maximum number of iterations. */
  maxit= 40; /* was 30 */
  
  /* Determine what is to be computed. */
  wantv= (job % 10 != 0);
  job= (job % 100) / 10;
  wantu= (job != 0);
  ncu= (job > 1) ? MIN(m,n) : m;

  /* Reduce X to bidiagonal form, storing the diagonal elements
     in S and the super-diagonal elements in E. */
  *info= 0;
  nct= MIN(m-1,n);
  nrt= MAX(0,MIN(n-2,m));
  lu= MAX(nct,nrt);
  if (lu >= 1) {
    for (l=1 ; l<=lu ; ++l) {
      if (l <= nct) {    
	/* Compute the transformation for the L-th
	   column and place the L-th diagonal in S(L). */
	s[l]= SNRM2(m-l+1, &x_(l,l), 1);
	if (s[l] != ZERO) {
	  if (x_(l,l) != ZERO) s[l]= SIGN(s[l], x_(l,l));
	  SSCAL(m-l+1, ONE/s[l], &x_(l,l), 1);
	  x_(l,l)+= ONE;
	  s[l]= -s[l];
	}
      }

      lp1= l+1;
      for (j=lp1 ; j<=n ; ++j) {
	if (l<=nct && s[l]!=ZERO) {
	  /* Apply the transformation. */
	  t= -SDOT(m-l+1, &x_(l,l), 1, &x_(l,j), 1) / x_(l,l);
	  SAXPY(m-l+1, t, &x_(l,l), 1, &x_(l,j), 1);
	}
      
	/* Place the L-th row of X into E for the
	   subsequent calculation of the row transformation. */
	e[j]= x_(l,j);
      }

      if (wantu && l<=nct) {
	/* Place the transformation in U for
	   subsequent back multiplication. */
	for (i=l ; i<=m ; ++i) {
	  u_(i,l)= x_(i,l);
	}
      }

      if (l <= nrt) {
	/* Compute the L-th row transformation and place
	   the L-th super-diagonal in E(L). */
	e[l]= SNRM2(n-l, &e[lp1], 1);
	if (e[l]) {
	  if (e[lp1]) e[l]= SIGN(e[l], e[lp1]);
	  SSCAL(n-l, ONE/e[l], &e[lp1], 1);
	  e[lp1]+= ONE;
	  e[l]= -e[l];
	}
	if (lp1<=m && e[l]!=ZERO) {
	  /* Apply the transformation. */
	  for (i=lp1 ; i<=m ; ++i) work[i]= ZERO;
	  for (j=lp1 ; j<=n ; ++j)
	    SAXPY(m-l, e[j], &x_(lp1,j), 1, &work[lp1], 1);
	  for (j=lp1 ; j<=n ; ++j)
	    SAXPY(m-l, -e[j]/e[lp1], &work[lp1], 1, &x_(lp1,j), 1);
	}

	if (wantv) {
	  /* Place the transformation in V for subsequent
	     back multiplication. */
	  for (i=lp1 ; i<=n ; ++i) v_(i,l) = e[i];
	}
      }
    }
  }

  /* Set up the final bidiagonal matrix or order P. */
  p= MIN(n,m+1);
  nctp1= nct + 1;
  nrtp1= nrt + 1;
  if (nct<n) s[nctp1]= x_(nctp1,nctp1);
  if (m<p) s[p]= ZERO;
  if (nrtp1<p) e[nrtp1]= x_(nrtp1,p);
  e[p]= ZERO;

  /* If required, generate U. */
  if (wantu) {
    for (j=nctp1 ; j<=ncu ; ++j) {
      for (i=1 ; i<=m ; ++i) u_(i,j) = ZERO;
      u_(j,j)= ONE;
    }
    for (l=nct ; l>=1 ; --l) {
      if (s[l] != ZERO) {
	lp1= l+1;
	for (j=lp1 ; j<=ncu ; ++j) {
	  t= -SDOT(m-l+1, &u_(l,l), 1, &u_(l,j), 1) / u_(l,l);
	  SAXPY(m-l+1, t, &u_(l,l), 1, &u_(l,j), 1);
	}
	SSCAL(m-l+1, -ONE, &u_(l, l), 1);
	u_(l,l)+= ONE;
	for (i=1 ; i<l ; ++i) u_(i,l)= ZERO;
      } else {
	for (i=1 ; i<=m ; ++i) u_(i,l)= ZERO;
	u_(l,l)= ONE;
      }
    }
  }

  /* If it is required, generate V. */
  if (wantv) {
    for (l=n ; l>=1 ; --l) {
      if (l<=nrt && e[l]!=ZERO) {
	lp1= l+1;
	for (j=lp1 ; j<=n ; ++j) {
	  t= -SDOT(n-l, &v_(lp1,l), 1, &v_(lp1,j), 1) / v_(lp1,l);
	  SAXPY(n-l, t, &v_(lp1,l), 1, &v_(lp1,j), 1);
	}
      }
      for (i=1 ; i<=n ; ++i) v_(i,l)= ZERO;
      v_(l,l)= ONE;
    }
  }

  /* Main iteration loop for the singular values. */
  mm= p;
  iter= 0;
  for (;;) {

    /* Quit if all the singular values have been found. */
    if (p == 0) return;

    /* If too many iterations have been performed, set flag and return. */
    if (iter >= maxit) {
      *info= p;
      return;
    }

    /* This section of the program inspects for negligible elements in
       the S and E arrays.  On completion the variables KASE and L are
       set as follows:
         KASE= 1     If S(P) and E(L-1) are negligible and L<P
	 KASE= 2     If S(L) is negligible and L<P
	 KASE= 3     If E(L-1) is negligible, L<P, and S(L)..S(P) are
                     not negligible (QR step).
         KASE= 4     If E(P-1) is negligible (convergence). */
    for (l=p-1 ; l>=1 ; --l) {
#if 0
      test1= ABS_OF_EXPR(t, s[l]);
      test1+= ABS_OF_EXPR(t, s[l+1]);
      test2= test + ABS_OF_EXPR(t, e[l]);
      if (test1 == test2) {
	e[l]= ZERO;
	break;
      }
#else
      GPL_NUM_STORE(&test1, FABS(s[l]) + FABS(s[l+1]));
      GPL_NUM_STORE(&test2, test1 + FABS(e[l]));
      if (test1 == test2) {
	e[l]= ZERO;
	break;
      }
#endif
    }
    if (l==p-1) {
      kase= 4;
    } else {
      for (ls=p ; ls>l ; --ls) {
#if 0
	test1= ls==p?(ls-1==l?ZERO:FABS(e[ls-1]))
		    :(ls-1==l?FABS(e[ls]):FABS(e[ls])+FABS(e[ls-1]));
	test2= test1 + FABS(s[ls]);
	if (test1 == test2) {
	  s[ls]= ZERO;
	  break;
	}
#else
	GPL_NUM_STORE(&test1,
		 ls==p?(ls-1==l?ZERO:FABS(e[ls-1]))
		      :(ls-1==l?FABS(e[ls]):FABS(e[ls])+FABS(e[ls-1])));
	GPL_NUM_STORE(&test2, test1 + FABS(s[ls]));
	if (test1 == test2) {
	  s[ls]= ZERO;
	  break;
	}
#endif
      }
      if (ls==l) {
	kase= 3;
      } else if (ls==p) {
	kase= 1;
      } else {
	kase= 2;
	l= ls;
      }
    }
    ++l;

    /* Perform the task indicated by KASE. */
    if (kase==1) {

      /* Deflate negligible S(P). */

      f= e[p-1];
      e[p-1]= ZERO;
      for (k=p-1 ; k>=l ; --k) {
	SROTG(&s[k], &f, &cs, &sn);
	if (k != l) {
	  f= -sn*e[k-1];
	  e[k-1]*= cs;
	}
	if (wantv) SROT(n, &v_(1,k), 1, &v_(1,p), 1, cs, sn);
      }

    } else if (kase==2) {

      /* Split at negligible S(L). */

      f= e[l-1];
      e[l-1]= ZERO;
      for (k=l ; k<=p ; ++k) {
	SROTG(&s[k], &f, &cs, &sn);
	f= -sn*e[k];
	e[k]*= cs;
	if (wantu) SROT(m, &u_(1,k), 1, &u_(1,l-1), 1, cs, sn);
      }

    } else if (kase==3) {

      /* Perform one QR step. */
    
      /* Calculate the shift. */
      scale= ABS_OF_EXPR(t, s[p]);
      if (scale < (t= ABS_OF_EXPR(t, s[p-1]))) scale= t;
      if (scale < (t= ABS_OF_EXPR(t, e[p-1]))) scale= t;
      if (scale < (t= ABS_OF_EXPR(t, s[l]))) scale= t;
      if (scale < (t= ABS_OF_EXPR(t, e[l]))) scale= t;
      sm= s[p] / scale;
      smm1= s[p-1] / scale;
      emm1= e[p-1] / scale;
      sl= s[l] / scale;
      el= e[l] / scale;
      b= ((smm1+sm) * (smm1-sm) + emm1*emm1) / TWO;
      c= sm * emm1;
      c*= c;
      shift= ZERO;
      if (b || c) {
	shift= sqrt(b*b + c);
	if (b < ZERO) shift= -shift;
	shift= c / (b + shift);
      }
      f= (sl+sm) * (sl-sm) - shift;
      g= sl * el;
    
      /* Chase zeros. */
      for (k=l ; k<p ; ++k) {
	SROTG(&f, &g, &cs, &sn);
	if (k!=l) e[k-1]= f;
	f= cs*s[k] + sn*e[k];
	e[k]= cs*e[k] - sn*s[k];
	g= sn*s[k+1];
	s[k+1]*= cs;
	if (wantv) SROT(n, &v_(1, k), 1, &v_(1, k+1), 1, cs, sn);
	SROTG(&f, &g, &cs, &sn);
	s[k]= f;
	f= cs*e[k] + sn*s[k + 1];
	s[k+1]= cs*s[k+1] - sn*e[k];
	g= sn*e[k + 1];
	e[k+1]*= cs;
	if (wantu && k<m) SROT(m, &u_(1,k), 1, &u_(1, k+1), 1, cs, sn);
      }
      e[p-1]= f;
      ++iter;

    } else if (kase==4) {

      /* Convergence. */

      /* Make the singular value positive. */
      if (s[l] < ZERO) {
	s[l]= -s[l];
	if (wantv) SSCAL(n, -ONE, &v_(1,l), 1);
      }
    
      /* Order the singular value. */
      while (l!=mm && s[l]<s[l+1]) {
	SWAP3(t, s[l], s[l+1]);
	if (wantv && l<n) SSWAP(n, &v_(1,l), 1, &v_(1,l+1), 1);
	if (wantu && l<m) SSWAP(m, &u_(1,l), 1, &u_(1,l+1), 1);
	++l;
      }
      iter= 0;
      --p;
    }
  }
}
#undef x_
#undef v_
#undef u_

/*---------------------------------------------------------------------------*/

void SSVCHK(real_t y[], const real_t x[], integer_t m, integer_t n,
	    const real_t s[], const real_t u[], integer_t ldu,
	   const real_t v[], integer_t ldv, real_t ws[])
/*
 * Check SVD factorization.
 *
 * Arguments:
 *     Y     (output) REAL M-vector.
 *           Result of: A*X.
 *
 *     X     (input) REAL N-vector.
 *           Vector to multiply by the SVD factorization of A.
 *
 *     M     (input) INTEGER, M >= 1.
 *           The number of rows of the input matrix A.
 *
 *     N     (input) INTEGER, N >= 1.
 *           The number of columns of the input matrix A.
 *
 *     S     (input) REAL min(M,N)-vector.
 *           Singular values of A: they are real and non-negative, and stored
 *           in descending order.
 *
 *     U     (input) REAL M-by-M or M-by-min(M,N) matrix.
 *           Orthogonal matrix of left singular vectors of A.
 *
 *     LDU   (input) INTEGER, LDU >= M.
 *           The leading dimension of the array U.
 *
 *     V     (input) REAL N-by-N or N-by-min(M,N) matrix.
 *           Orthogonal matrix of right singular vectors of A.
 *
 *     LDV   (input) INTEGER.   LDV >= N.
 *           Leading dimension of array V.
 *
 *     WS    (workspace) REAL min(M,N)-vector.
 *           Workspace area, unused if NULL.  If the singular values are no
 *           longer needed, WS can be the same as S.
 */
{
#define x_(a1)    x[a1]
#define y_(a1)    y[a1]
#define s_(a1)    s[a1]
#define u_(a1,a2) u[ldu*(a2)+a1]
#define v_(a1,a2) v[ldv*(a2)+a1]
#define ws_(a1)   ws[a1]
  integer_t i, j, k;
  integer_t mn= MIN(m, n);
  real_t q;
  double sum; /* In any cases, inner product is done in double precision. */

  if (ws==NULL) {
    for (i=0 ; i<m ; ++i) y_(i)= ZERO;
    for (k=0 ; k<mn ; ++k) {
      /* Form Q = S.V'.X */
      if (! (q= s_(k))) {
	/* Since singular values are sorted in descending order,
	   all other values are negligible. */
	break;
      }
      for (sum=0.0, j=0 ; j<n ; ++j) sum+= v_(j,k) * x_(j);
      q*= sum;

      /* Accumulate U.(S.V'.X) */
      for (i=0 ; i<m ; ++i) y_(i)+= q * u_(i,k);
    }
  } else {
    /* Store S.V'.X in WS */
    for (k=0 ; k<mn ; ++k) {
      if (! (q= s_(k))) {
	/* Since singular values are sorted in descending order,
	   all other values are negligible. */
	mn= k;
	break;
      }
      for (sum=0.0, j=0 ; j<n ; ++j) sum+= v_(j,k) * x_(j);
      ws_(k)= sum * q;
    }
    /* Form U.(S.V'.X) */
    for (i=0 ; i<m ; ++i) {
      for (sum=0.0, k=0 ; k<mn ; ++k) sum+= u_(i,k) * ws_(k);
      y_(i)= sum;
    }
  }
}
#undef x_
#undef y_
#undef s_
#undef u_
#undef v_
#undef ws_

/*---------------------------------------------------------------------------*/

integer_t SSVSLV(real_t x[], const real_t b[], integer_t m, integer_t n,
		 const real_t s[], const real_t u[], integer_t ldu,
		 const real_t v[], integer_t ldv, real_t r, real_t ws[])
/*
 * Given the Singular Value Decomposition of a matrix A, solve the
 * regularized problem: A*X=B and return the rank of A.
 *
 * Arguments:
 *     X     (output) REAL N-vector.
 *           Solution of: A*X = B.
 *
 *     B     (input) REAL M-vector.
 *           Right hand side vector.
 *
 *     M     (input) INTEGER, M >= 1.
 *           The number of rows of the input matrix A.
 *
 *     N     (input) INTEGER, N >= 1.
 *           The number of columns of the input matrix A.
 *
 *     S     (input) REAL min(M,N)-vector.
 *           Singular values of A: they are real and non-negative, and stored
 *           in descending order.
 *
 *     U     (input) REAL M-by-M or M-by-min(M,N) matrix.
 *           Orthogonal matrix of left singular vectors of A.
 *
 *     LDU   (input) INTEGER, LDU >= M.
 *           The leading dimension of the array U.
 *
 *     V     (input) REAL N-by-N or N-by-min(M,N) matrix.
 *           Orthogonal matrix of right singular vectors of A.
 *
 *     LDV   (input) INTEGER.   LDV >= N.
 *           Leading dimension of array V.
 *
 *     R     (input) REAL.
 *           Regularization level: R=0.0 for no regularization; otherwise,
 *           the singular value cutoff is about R * max(S).  Suggested value:
 *           not less than sqrt(machine precision) (SIC).
 *
 *     WS    (workspace) REAL min(M,N)-vector.
 *           Workspace area, unused if NULL.  If the singular values are no
 *           longer needed, WS can be the same as S.
 *
 * Return:
 *     The estimated rank of A (number of singular values above cutoff).
 */
{
#define x_(a1)    x[a1]
#define b_(a1)    b[a1]
#define s_(a1)    s[a1]
#define u_(a1,a2) u[ldu*(a2)+a1]
#define v_(a1,a2) v[ldv*(a2)+a1]
#define ws_(a1)   ws[a1]
  double sum;  /* In any cases, inner product is done in double precision. */
  real_t r2, q;
  integer_t i, j, k;
  integer_t mn= MIN(m, n);
  integer_t rank= 0;
  
  /* Initialization. */
  if (r<ZERO) r= ZERO;
  else if (r>ZERO) r*= s_(0); /* 1st singular value has highest value */
  r2= r*r;
  
  if (ws==NULL) {
    /* Perform regularized inversion without temporary vector. */
    for (j=0 ; j<n ; ++j) x_(j)= ZERO;
    for (k=0 ; k<mn ; ++k) {
      if (! (q= s_(k))) {
	/* Since singular values are sorted in descending order,
	   all other values are negligible. */
	break;
      }
      /* Form (U'.B)/SIGMA */
      for (sum=0.0, i=0 ; i<m ; ++i) sum+= u_(i,k) * b_(i);
      if (q < r) {
	q*= sum/(q*q+r2);
      } else {
	rank+= 1;
	q= sum/(q+r2/q);
      }

      /* Accumulate V.((U'.B)/SIGMA)) */
      for (j=0 ; j<n ; ++j) x_(j)+= q * v_(j,k);
    }
  } else {
    /* Perform regularized inversion with a temporary vector. */
    /* Compute WS= (1/S)*tranpose(U)*B. */
    for (k=0 ; k<mn ; ++k) {
      if (! (q= s_(k))) {
	/* Since singular values are sorted in descending order,
	   all other values are negligible. */
	mn= k;
	break;
      }
      for (sum=0.0, i=0 ; i<m ; ++i) sum+= u_(i,k) * b_(i);
      if (q < r) {
	ws_(k)= sum*q/(q*q+r2);
      } else {
	rank+= 1;
	ws_(k)= sum/(q+r2/q);
      }
    }

    /* Compute: X= tranpose(V)*WS. */
    for (j=0 ; j<n ; ++j) {
      for (sum=0.0, k=0 ; k<mn ; ++k) sum+= v_(j,k) * ws_(k);
      x_(j)= sum;
    }
  }
  return rank;
}
#undef x_
#undef b_
#undef s_
#undef u_
#undef v_
#undef ws_

#endif /* _GPL_NUM_SVD_C ------------------------------------------------------*/
