/*
 * gpl_ops1.c --
 *
 *	Vectorized operations for GPLib library.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */

#ifndef GPL_OPS1_C
#define GPL_OPS1_C 1

#include <math.h>
#include "gpl_base.h"
#include "gpl_macros.h"
#include "gpl_ops1.h"

#define NAME(name, abbrev) GPL_JOIN4(gpl_, name, _, abbrev)
#define TRUE_R(x)         ((x) != 0)
#define FALSE_R(x)        (! (x))
#define TRUE_R_I(x, i)    ((x)[i] != 0)
#define FALSE_R_I(x, i)   (! (x)[i])
#define TRUE_Z(re, im)    ((re) || (im))
#define FALSE_Z(re, im)   (! (re) && ! (im))
#define TRUE_Z_I(x, i)    TRUE_Z((x)[i], (x)[(i)+1])
#define FALSE_Z_I(x, i)   FALSE_Z((x)[i], (x)[(i)+1])
#define AND_R_I(x, y, i)  ((x)[i] && (y)[i])
#define OR_R_I(x, y, i)   ((x)[i] || (y)[i])
#define AND_Z_I(x, y, i)  (TRUE_Z_I(x, i) && TRUE_Z_I(y, i))
#define OR_Z_I(x, y, i)   (TRUE_Z_I(x, i) || TRUE_Z_I(y, i))

#define UNARY_PROTO(dst_type, name, abbrev) dst_type *NAME(name, abbrev) \
  (dst_type *dst, const CTYPE *src, int number)

#define BINARY_PROTO(dst_type, name, abbrev) dst_type *NAME(name, abbrev) \
  (dst_type *dst, const CTYPE *left, const CTYPE *right, int number, int flags)

#define SIMPLE_BINARY(OP)						      \
  int i;								      \
  if (flags == 1) /* left operand is a scalar */ {			      \
    RTYPE _left= left[0];						      \
    for (i=0 ; i<number ; i++) { dst[i]= _left OP right[i]; }		      \
  } else if (flags == 2) /* right operand is a scalar */ {		      \
    RTYPE _right= right[0];						      \
    for (i=0 ; i<number ; i++) { dst[i]= left[i] OP _right; }		      \
  } else /* assume both operands have same length */ {			      \
    for (i=0 ; i<number ; i++) { dst[i]= left[i] OP right[i]; }		      \
  }									      \
  return dst

#define SIMPLE_BINARY_FUNC(FUNC)					      \
  int i;								      \
  if (flags == 1) /* left operand is a scalar */ {			      \
    double _left= left[0];						      \
    for (i=0 ; i<number ; i++) { dst[i]= FUNC(_left, right[i]); }	      \
  } else if (flags == 2) /* right operand is a scalar */ {		      \
    double _right= right[0];						      \
    for (i=0 ; i<number ; i++) { dst[i]= FUNC(left[i], _right); }	      \
  } else /* assume both operands have same length */ {			      \
    for (i=0 ; i<number ; i++) { dst[i]= FUNC(left[i], right[i]); }	      \
  }									      \
  return dst

#define CHECK_DIVIDE(OP)						      \
  int i;								      \
  RTYPE _left, _right;							      \
  if (flags == 1) /* left operand is a scalar */ {			      \
    _left= left[0];							      \
    for (i=0 ; i<number ; i++) {					      \
  	if (! (_right= right[i])) gpl_integer_division_by_zero();		      \
  	dst[i]= _left OP _right;					      \
    }									      \
  } else if (flags == 2) /* right operand is a scalar */ {		      \
    if (! (_right= right[0])) gpl_integer_division_by_zero();		      \
    for (i=0 ; i<number ; i++) { dst[i]= left[i] OP _right; }		      \
  } else /* assume both operands have same length */ {			      \
    for (i=0 ; i<number ; i++) {					      \
  	if (! (_right= right[i])) gpl_integer_division_by_zero();		      \
  	dst[i]= left[i] OP _right;					      \
    }									      \
  }									      \
  return dst

/*---------------------------------------------------------------------------*/
/* The trick here is to self-include this file with a different definition of
   the macro TYPE. */

#undef TYPE
#define TYPE BYTE
#include __FILE__

#undef TYPE
#define TYPE SHORT
#include __FILE__

#undef TYPE
#define TYPE INT
#include __FILE__

#undef TYPE
#define TYPE LONG
#include __FILE__

#undef TYPE
#define TYPE FLOAT
#include __FILE__

#undef TYPE
#define TYPE DOUBLE
#include __FILE__

#undef TYPE
#define TYPE FCOMPLEX
#include __FILE__

#undef TYPE
#define TYPE DCOMPLEX
#include __FILE__

#else /* GPL_OPS1_C defined ---------------------------------------------------*/

#ifndef TYPE
# error "must be compiled with -DTYPE=BYTE/SHORT/INT/LONG/FLOAT/..."
#endif

#undef CTYPE
#define CTYPE              GPL_CTYPE_OF(TYPE)
#undef RTYPE
#define RTYPE              GPL_CTYPE_OF(GPL_REAL_PART_OF(TYPE))
#undef ABBREV
#define ABBREV             GPL_ABBREV(TYPE)
#undef IS_INTEGER
#define IS_INTEGER         GPL_IS_INTEGER(TYPE)
#undef IS_COMPLEX
#define IS_COMPLEX         GPL_IS_COMPLEX(TYPE)

/*---------------------------------------------------------------------------*/
/* COPY */

extern UNARY_PROTO(CTYPE, copy, ABBREV);

UNARY_PROTO(CTYPE, copy, ABBREV)
{
  if (dst!=src) {
    int i;
#if IS_COMPLEX
    RTYPE *_dst= (RTYPE*)dst, *_src= (RTYPE *)src;
    number*= 2;
    for (i=0 ; i<number ; i++) { _dst[i]= _src[i]; }
#else
    for (i=0 ; i<number ; i++) { dst[i]= src[i]; }
#endif
  }
  return dst;
}

/*---------------------------------------------------------------------------*/
/* NEGATE */

extern UNARY_PROTO(CTYPE, negate, ABBREV);

UNARY_PROTO(CTYPE, negate, ABBREV)
{
  int i;
#if IS_COMPLEX
  RTYPE *_dst= (RTYPE*)dst, *_src= (RTYPE *)src;
  number*= 2;
  for (i=0 ; i<number ; i++) { _dst[i]= -_src[i]; }
#else
  for (i=0 ; i<number ; i++) { dst[i]= -src[i]; }
#endif
  return dst;
}

/*---------------------------------------------------------------------------*/
/* COMPLEMENT */

#if IS_INTEGER
extern UNARY_PROTO(CTYPE, complement, ABBREV);

UNARY_PROTO(CTYPE, complement, ABBREV)
{
  int i;
  for (i=0 ; i<number ; i++) { dst[i]= ~src[i]; }
  return dst;
}
#endif /* IS_INTEGER */

/*---------------------------------------------------------------------------*/
/* NOT */

extern UNARY_PROTO(int, not, ABBREV);

UNARY_PROTO(int, not, ABBREV)
{
  int i;
#if IS_COMPLEX
  RTYPE *_src= (RTYPE *)src;
  for (i=0 ; i<number ; i++) { dst[i]= FALSE_Z(_src[2*i], _src[2*i+1]); }
#else
  for (i=0 ; i<number ; i++) { dst[i]= FALSE_R(src[i]); }
#endif
  return dst;
}

/*---------------------------------------------------------------------------*/
/* TRUE */

extern UNARY_PROTO(int, true, ABBREV);

UNARY_PROTO(int, true, ABBREV)
{
  int i;
#if IS_COMPLEX
  RTYPE *_src= (RTYPE *)src;
  for (i=0 ; i<number ; i++) { dst[i]= TRUE_Z(_src[2*i], _src[2*i+1]); }
#else
  for (i=0 ; i<number ; i++) { dst[i]= TRUE_R(src[i]); }
#endif
  return dst;
}

/*---------------------------------------------------------------------------*/
/* SQUARE */

extern UNARY_PROTO(CTYPE, square, ABBREV);

UNARY_PROTO(CTYPE, square, ABBREV)
{
  int i;
#if IS_COMPLEX
  RTYPE *_dst= (RTYPE *)dst, *_src= (RTYPE *)src, re, im;
  number*= 2;
  for (i=0 ; i<number ; i++) {
    re= _src[i]; im= _src[i+1];
    _dst[i]= re*re - im*im;
    _dst[i+1]= 2*re*im;
  }
#else
  for (i=0 ; i<number ; i++) { dst[i]= src[i] * src[i]; }
#endif
  return dst;
}

/*---------------------------------------------------------------------------*/
/* CUBE */

extern UNARY_PROTO(CTYPE, cube, ABBREV);

UNARY_PROTO(CTYPE, cube, ABBREV)
{
  int i;
#if IS_COMPLEX
  RTYPE *_dst= (RTYPE *)dst, *_src= (RTYPE *)src, re, im;
  number*= 2;
  for (i=0 ; i<number ; i++) {
    re= _src[i]; im= _src[i+1];
    _dst[i]= re*re*re - 3*re*im*im;
    _dst[i+1]= 3*re*re*im - im*im*im;
  }
#else
  for (i=0 ; i<number ; i++) { dst[i]= src[i] * src[i] * src[i]; }
#endif
  return dst;
}

/*---------------------------------------------------------------------------*/
/* TO_OTHER */

#undef FUNCTION
#if IS_COMPLEX
# define FUNCTION(ctype, name, abbrev)					      \
extern ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number);   \
ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number)	      \
{									      \
  int i;								      \
  RTYPE *_src= (RTYPE *)src;						      \
  for (i=0 ; i<number ; i++) { dst[i]= _src[2*i]; }			      \
  return dst;								      \
}
#else /* not IS_COMPLEX */
# define FUNCTION(ctype, name, abbrev)					      \
extern ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number);   \
ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number)	      \
{									      \
  int i;								      \
  for (i=0 ; i<number ; i++) { dst[i]= src[i]; }			      \
  return dst;								      \
}
#endif /* not IS_COMPLEX */

FUNCTION(gpl_byte_t, to_byte, ABBREV)
FUNCTION(short, to_short, ABBREV)
FUNCTION(int, to_int, ABBREV)
FUNCTION(long, to_long, ABBREV)
FUNCTION(float, to_float, ABBREV)
FUNCTION(double, to_double, ABBREV)

#undef FUNCTION
#if IS_COMPLEX
# define FUNCTION(ctype, rtype, name, abbrev)				      \
extern ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number);   \
ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number)	      \
{									      \
  int i;								      \
  rtype *_dst= (rtype *)dst;						      \
  RTYPE *_src= (RTYPE *)src;						      \
  number*= 2;								      \
  for (i=0 ; i<number ; i++) { _dst[i]= _src[i]; }			      \
  return dst;								      \
}
#else /* not IS_COMPLEX */
# define FUNCTION(ctype, rtype, name, abbrev)				      \
extern ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number);   \
ctype *NAME(name, abbrev)(ctype *dst, const CTYPE *src, int number)	      \
{									      \
  int i;								      \
  rtype *_dst= (rtype *)dst;						      \
  for (i=0 ; i<number ; i++) { _dst[2*i]= src[i]; _dst[2*i+1]= 0; }	      \
  return dst;								      \
}
#endif /* not IS_COMPLEX */

FUNCTION(gpl_fcomplex_t, float, to_fcomplex, ABBREV)
FUNCTION(gpl_dcomplex_t, double, to_dcomplex, ABBREV)

/*---------------------------------------------------------------------------*/
/* ADD */

extern BINARY_PROTO(CTYPE, add, ABBREV);

BINARY_PROTO(CTYPE, add, ABBREV)
{
#if ! IS_COMPLEX
  SIMPLE_BINARY(+);
#else /* IS_COMPLEX */
  int i;
  RTYPE *_dst= (RTYPE *)dst, *_left= (RTYPE *)left, *_right= (RTYPE *)right;
  number*= 2;
  if (flags == 1) /* left operand is a scalar */ {
    RTYPE lr= _left[0], li= _left[1];
    for (i=0 ; i<number ; i+=2) {
      _dst[i]=   GPL_ADD_RE(lr, li, _right[i], _right[i+1]);
      _dst[i+1]= GPL_ADD_IM(lr, li, _right[i], _right[i+1]);
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    RTYPE rr= _right[0], ri= _right[1];
    for (i=0 ; i<number ; i+=2) {
      _dst[i]=   GPL_ADD_RE(_left[i], _left[i+1], rr, ri);
      _dst[i+1]= GPL_ADD_IM(_left[i], _left[i+1], rr, ri);
    }
  } else /* assume both operands have same length */ {
    for (i=0 ; i<number ; i+=2) {
      _dst[i]=   GPL_ADD_RE(_left[i], _left[i+1], _right[i], _right[i+1]);
      _dst[i+1]= GPL_ADD_IM(_left[i], _left[i+1], _right[i], _right[i+1]);
    }
  }
  return dst;
#endif /* IS_COMPLEX */
}

/*---------------------------------------------------------------------------*/
/* SUBTRACT */

extern BINARY_PROTO(CTYPE, subtract, ABBREV);

BINARY_PROTO(CTYPE, subtract, ABBREV)
{
#if ! IS_COMPLEX
  SIMPLE_BINARY(-);
#else /* IS_COMPLEX */
  int i;
  RTYPE *_dst= (RTYPE *)dst, *_left= (RTYPE *)left, *_right= (RTYPE *)right;
  number*= 2;
  if (flags == 1) /* left operand is a scalar */ {
    RTYPE lr= _left[0], li= _left[1];
    for (i=0 ; i<number ; i+=2) {
      _dst[i]=   GPL_SUBTRACT_RE(lr, li, _right[i], _right[i+1]);
      _dst[i+1]= GPL_SUBTRACT_IM(lr, li, _right[i], _right[i+1]);
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    RTYPE rr= _right[0], ri= _right[1];
    for (i=0 ; i<number ; i+=2) {
      _dst[i]=   GPL_SUBTRACT_RE(_left[i], _left[i+1], rr, ri);
      _dst[i+1]= GPL_SUBTRACT_IM(_left[i], _left[i+1], rr, ri);
    }
  } else /* assume both operands have same length */ {
    for (i=0 ; i<number ; i+=2) {
      _dst[i]=   GPL_SUBTRACT_RE(_left[i], _left[i+1], _right[i], _right[i+1]);
      _dst[i+1]= GPL_SUBTRACT_IM(_left[i], _left[i+1], _right[i], _right[i+1]);
    }
  }
  return dst;
#endif /* IS_COMPLEX */
}

/*---------------------------------------------------------------------------*/
/* MULTIPLY */

extern BINARY_PROTO(CTYPE, multiply, ABBREV);

BINARY_PROTO(CTYPE, multiply, ABBREV)
{
#if ! IS_COMPLEX
  SIMPLE_BINARY(*);
#else /* IS_COMPLEX */
  int i;
  RTYPE *_dst= (RTYPE *)dst, *_left= (RTYPE *)left, *_right= (RTYPE *)right;
  RTYPE lr, li, rr, ri;
  number*= 2;
  if (flags == 1) /* left operand is a scalar */ {
    lr= _left[0]; li= _left[1];
    for (i=0 ; i<number ; i+=2) {
      rr= _right[i]; ri= _right[i+1];
      _dst[i]=   GPL_MULTIPLY_RE(lr, li, rr, ri);
      _dst[i+1]= GPL_MULTIPLY_IM(lr, li, rr, ri);
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    rr= _right[0]; ri= _right[1];
    for (i=0 ; i<number ; i+=2) {
      lr= _left[i]; li= _left[i+1];
      _dst[i]=   GPL_MULTIPLY_RE(lr, li, rr, ri);
      _dst[i+1]= GPL_MULTIPLY_IM(lr, li, rr, ri);
    }
  } else /* assume both operands have same length */ {
    for (i=0 ; i<number ; i+=2) {
      lr= _left[i]; li= _left[i+1];
      rr= _right[i]; ri= _right[i+1];
      _dst[i]=   GPL_MULTIPLY_RE(lr, li, rr, ri);
      _dst[i+1]= GPL_MULTIPLY_IM(lr, li, rr, ri);
    }
  }
  return dst;
#endif /* IS_COMPLEX */
}

/*---------------------------------------------------------------------------*/
/* DIVIDE */

extern BINARY_PROTO(CTYPE, divide, ABBREV);

BINARY_PROTO(CTYPE, divide, ABBREV)
{
#if ! IS_COMPLEX
  SIMPLE_BINARY(/);
#else /* IS_COMPLEX */
  int i;
  RTYPE *_dst= (RTYPE *)dst, *_left= (RTYPE *)left, *_right= (RTYPE *)right;
  RTYPE lr, li, rr, ri;
  number*= 2;
  if (flags == 1) /* left operand is a scalar */ {
    lr= _left[0]; li= _left[1];
    for (i=0 ; i<number ; i+=2) {
      rr= _right[i]; ri= _right[i+1];
      /* avoid overflows */
      if (GPL_ABS(rr) > GPL_ABS(ri)) {
	ri/= rr;
	rr*= (1.0 + ri*ri);
	_dst[i]= (lr + ri*li) / rr;
	_dst[i+1]= (li - ri*lr) / rr;
      } else {
	rr/= ri;
	ri*= (1.0 + rr*rr);
	_dst[i]= (rr*lr + li) / ri;
	_dst[i+1]= (rr*li - lr) / ri;
      }
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    rr= _right[0]; ri= _right[1];
    /* avoid overflows */
    if (GPL_ABS(rr) > GPL_ABS(ri)) {
      ri/= rr;
      rr*= (1.0 + ri*ri);
      for (i=0 ; i<number ; i+=2) {
	lr= _left[i]; li= _left[i+1];
	_dst[i]= (lr + ri*li) / rr;
	_dst[i+1]= (li - ri*lr) / rr;
      }
    } else {
      rr/= ri;
      ri*= (1.0 + rr*rr);
      for (i=0 ; i<number ; i+=2) {
	lr= _left[i]; li= _left[i+1];
	_dst[i]= (rr*lr + li) / ri;
	_dst[i+1]= (rr*li - lr) / ri;
      }
    }
  } else /* assume both operands have same length */ {
    for (i=0 ; i<number ; i+=2) {
      lr= _left[i]; li= _left[i+1];
      rr= _right[i]; ri= _right[i+1];
      if (GPL_ABS(rr) > GPL_ABS(ri)) {
	ri/= rr;
	rr*= (1.0 + ri*ri);
	_dst[i]= (lr + ri*li) / rr;
	_dst[i+1]= (li - ri*lr) / rr;
      } else {
	rr/= ri;
	ri*= (1.0 + rr*rr);
	_dst[i]= (rr*lr + li) / ri;
	_dst[i+1]= (rr*li - lr) / ri;
      }
    }
  }
  return dst;
#endif /* IS_COMPLEX */
}

/*---------------------------------------------------------------------------*/
/* MODULO */

#if ! IS_COMPLEX

extern BINARY_PROTO(CTYPE, modulo, ABBREV);

BINARY_PROTO(CTYPE, modulo, ABBREV)
{
# if IS_INTEGER
  SIMPLE_BINARY(%);
# else /* IS_INTEGER */
  SIMPLE_BINARY_FUNC(fmod);
# endif  /* IS_INTEGER */
}

#endif /* IS_COMPLEX */

/*---------------------------------------------------------------------------*/
/* POWER */

extern BINARY_PROTO(CTYPE, pow, ABBREV);

BINARY_PROTO(CTYPE, pow, ABBREV)
{
#if ! IS_COMPLEX
  SIMPLE_BINARY_FUNC(pow);
#else /* IS_COMPLEX */
  return NULL;
#endif /* IS_COMPLEX */
}

/*---------------------------------------------------------------------------*/
/* COMPARISON */

#if IS_COMPLEX

int *NAME(equal,ABBREV)(int *dst, const CTYPE *left,
			const CTYPE *right, int number, int flags)
{
  int i;
  if (flags == 1) /* left operand is a scalar */ {
    RTYPE lr= left->re, li= left->im, *_right= (RTYPE *)right;
    for (i=0 ; i<number ; i++) {
      dst[i]= (lr==_right[2*i] && li==_right[2*i+1]);
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    RTYPE *_left= (RTYPE*)left, rr= right->re, ri= right->im;
    for (i=0 ; i<number ; i++) {
      dst[i]= (_left[2*i]==rr && _left[2*i+1]==ri);
    }
  } else /* assume both operands have same length */ {
    RTYPE *_left= (RTYPE*)left, *_right= (RTYPE *)right;
    for (i=0 ; i<number ; i++) {
      dst[i]= (_left[2*i]==_right[2*i] && _left[2*i+1]==_right[2*i+1]);
    }
  }
  return dst;
}

int *NAME(not_equal,ABBREV)(int *dst, const CTYPE *left,
			    const CTYPE *right, int number, int flags)
{
  int i;
  if (flags == 1) /* left operand is a scalar */ {
    RTYPE lr= left->re, li= left->im, *_right= (RTYPE *)right;
    for (i=0 ; i<number ; i++) {
      dst[i]= (lr!=_right[2*i] || li!=_right[2*i+1]);
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    RTYPE *_left= (RTYPE*)left, rr= right->re, ri= right->im;
    for (i=0 ; i<number ; i++) {
      dst[i]= (_left[2*i]!=rr || _left[2*i+1]!=ri);
    }
  } else /* assume both operands have same length */ {
    RTYPE *_left= (RTYPE*)left, *_right= (RTYPE *)right;
    for (i=0 ; i<number ; i++) {
      dst[i]= (_left[2*i]!=_right[2*i] || _left[2*i+1]!=_right[2*i+1]);
    }
  }
  return dst;
}

#else /* not IS_COMPLEX */

extern BINARY_PROTO(int, equal, ABBREV);
BINARY_PROTO(int, equal, ABBREV) { SIMPLE_BINARY(==); }

extern BINARY_PROTO(int, not_equal, ABBREV);
BINARY_PROTO(int, not_equal, ABBREV) { SIMPLE_BINARY(!=); }

extern BINARY_PROTO(int, greater, ABBREV);
BINARY_PROTO(int, greater, ABBREV) { SIMPLE_BINARY(>); }

extern BINARY_PROTO(int, greater_eq, ABBREV);
BINARY_PROTO(int, greater_eq, ABBREV) { SIMPLE_BINARY(>=); }

extern BINARY_PROTO(int, less, ABBREV);
BINARY_PROTO(int, less, ABBREV) { SIMPLE_BINARY(<); }

extern BINARY_PROTO(int, less_eq, ABBREV);
BINARY_PROTO(int, less_eq, ABBREV) { SIMPLE_BINARY(<=); }

#endif /* not IS_COMPLEX */

#if IS_INTEGER

extern BINARY_PROTO(CTYPE, shift_left, ABBREV);
BINARY_PROTO(CTYPE, shift_left, ABBREV) { SIMPLE_BINARY(<<); }

extern BINARY_PROTO(CTYPE, shift_right, ABBREV);
BINARY_PROTO(CTYPE, shift_right, ABBREV) { SIMPLE_BINARY(>>); }

extern BINARY_PROTO(CTYPE, bitwise_or, ABBREV);
BINARY_PROTO(CTYPE, bitwise_or, ABBREV) { SIMPLE_BINARY(|); }

extern BINARY_PROTO(CTYPE, bitwise_and, ABBREV);
BINARY_PROTO(CTYPE, bitwise_and, ABBREV) { SIMPLE_BINARY(&); }

extern BINARY_PROTO(CTYPE, bitwise_xor, ABBREV);
BINARY_PROTO(CTYPE, bitwise_xor, ABBREV) { SIMPLE_BINARY(^); }

#endif /* IS_INTEGER */

/*---------------------------------------------------------------------------*/
/* LOGICAL_AND/OR */

#undef LEFT
#undef RIGHT
#undef FALSE_I
#undef TRUE_I
#undef AND_I
#undef OR_I

#if IS_COMPLEX
# define FALSE_I(x, i)  FALSE_Z_I(x, 2*i)
# define TRUE_I(x, i)   TRUE_Z_I(x, 2*i)
# define AND_I(x, y, i) AND_Z_I(x, y, 2*i);
# define OR_I(x, y, i)  OR_Z_I(x, y, 2*i);
# define RIGHT _right
# define LEFT  _left
#else
# define FALSE_I(x, i)  FALSE_R_I(x, i)
# define TRUE_I(x, i)   TRUE_R_I(x, i)
# define AND_I(x, y, i) AND_R_I(x, y, i);
# define OR_I(x, y, i)  OR_R_I(x, y, i);
# define RIGHT right
# define LEFT  left
#endif

extern BINARY_PROTO(int, logical_and, ABBREV);

BINARY_PROTO(int, logical_and, ABBREV)
{
#if IS_COMPLEX
  RTYPE *_left= (RTYPE *)left, *_right= (RTYPE *)right;
#endif
  int i;
  if (flags == 1) /* left operand is a scalar */ {
    if (FALSE_I(LEFT, 0)) {
      for (i=0 ; i<number ; i++) { dst[i]= 0; }
    } else {
      for (i=0 ; i<number ; i++) { dst[i]= TRUE_I(RIGHT, i); }
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    if (FALSE_I(RIGHT, 0)) {
      for (i=0 ; i<number ; i++) { dst[i]= 0; }
    } else {
      for (i=0 ; i<number ; i++) { dst[i]= TRUE_I(LEFT, i); }
    }
  } else /* assume both operands have same length */ {
    for (i=0 ; i<number ; i++) { dst[i]= AND_I(LEFT, RIGHT, i); }
  }
  return dst;
}

extern BINARY_PROTO(int, logical_or, ABBREV);

BINARY_PROTO(int, logical_or, ABBREV)
{
#if IS_COMPLEX
  RTYPE *_left= (RTYPE *)left, *_right= (RTYPE *)right;
#endif
  int i;
  if (flags == 1) /* left operand is a scalar */ {
    if (TRUE_I(LEFT, 0)) {
      for (i=0 ; i<number ; i++) { dst[i]= 1; }
    } else {
      for (i=0 ; i<number ; i++) { dst[i]= TRUE_I(RIGHT, i); }
    }
  } else if (flags == 2) /* right operand is a scalar */ {
    if (TRUE_I(RIGHT, 0)) {
      for (i=0 ; i<number ; i++) { dst[i]= 1; }
    } else {
      for (i=0 ; i<number ; i++) { dst[i]= TRUE_I(LEFT, i); }
    }
  } else /* assume both operands have same length */ {
    for (i=0 ; i<number ; i++) { dst[i]= OR_I(LEFT, RIGHT, i); }
  }
  return dst;
}

/*---------------------------------------------------------------------------*/
/* MIN/MAX */

#if ! IS_COMPLEX
BINARY_PROTO(RTYPE, min2, ABBREV);
BINARY_PROTO(RTYPE, min2, ABBREV)
{
  SIMPLE_BINARY_FUNC(GPL_MIN);
}

BINARY_PROTO(RTYPE, max2, ABBREV);
BINARY_PROTO(RTYPE, max2, ABBREV)
{
  SIMPLE_BINARY_FUNC(GPL_MAX);
}
#endif /* IS_COMPLEX */

#endif /* GPL_OPS1_C defined --------------------------------------------------*/
