/*
 * gpl_option.h --
 *
 *	Definitions for parsing of command line options in GPLib library.
 *
 * Copyright and warranties:
 *	Copyright (c) 1999 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef GPL_OPTION_H
#define GPL_OPTION_H 1

#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------------*/
  
/* PARSING OF COMMAND LINE OPTIONS */

typedef struct _gpl_opt_data gpl_opt_data_t;
typedef struct _gpl_opt_def  gpl_opt_def_t;

/* The `gpl_opt_def_t' structure describes one option (see below for details). */
struct _gpl_opt_def {
  const char *name;    /* Long option name (NULL if none). */
  const char *abbrev;  /* String of possible option abbreviations
			  (NULL if none). */
  int         has_arg; /* Is there an argument for this option?  One
			  of the following values: */
#define GPL_OPT_ARG_NONE      0
#define GPL_OPT_ARG_REQUIRED  1
#define GPL_OPT_ARG_OPTIONAL  2
  int         value;   /* User defined value or client data. */
};

/* The `gpl_opt_data_t' structure is initialized by `gpl_opt_init()' and managed
   by `gpl_opt_next()'.  Its contents should be considered as "read-only" by the
   caller. */
struct _gpl_opt_data {
  int                argc;	/* Argument count. */
  const char       **argv;	/* Argument array. */
  const gpl_opt_def_t *options;	/* Array of recognized options. */
  int                mode;	/* Mode for parsing arguments.  One of: */
#define GPL_OPT_PERMUTE_ARGS    1	/* Permute arguments in `argv' so that all
				   options get parsed in order and, when
				   `gpl_opt_next()' returns
				   GPL_OPT_END_OF_OPTIONS, all other (i.e.
				   non-option) arguments are stored in
				   `argv[arg_index...argc-1]'. */
#define GPL_OPT_POSIXLY_CORRECT 2	/* First non-option argument marks the
				   end of command line options. */
#define GPL_OPT_RETURN_IN_ORDER 3	/* Etl_GetOption returns each time a non-option
				   argument is encountered and stores its
				   textual value in optionArg. */
  int               first_arg;	/* Index of first non-option argument meet so
				   far (only used in GPL_OPT_PERMUTE_ARGS
				   mode). */
  int               arg_index;	/* Index in `argv' of next argument to process.
				   In case of error (i.e. when
				   `gpl_opt_next()' returns ':' or '?'), this
				   is the index of the invalid argument.  When
				   all options are parsed (i.e. when
				   `gpl_opt_next()' returns
				   GPL_OPT_END_OF_OPTIONS), this is the index of
				   the first non-option argument of `argv'. */
  const char       *next_opt;	/* Pointer to next character in current
				   argument, if this is a short option; NULL,
				   otherwise. */
  const char       *opt_arg;	/* Argument of last parsed option (NULL if
				   option has no argument). */
  int               opt_index;	/* Index of last parsed option (-1 if last
				   parsed argument was not an option). */
  int               opt_char;	/* Character of last abbreviated option (0 if
				   last parsed argument was not an abbreviated
				   option; may be used for error reporting). */
};

extern void gpl_opt_init(gpl_opt_data_t *opt, int argc, const char *argv[],
		       const gpl_opt_def_t options[], int mode);
/*	Instanciates the structure pointed by OPT for further parsing by
	`gpl_opt_next()'.  ARGC and ARGV are the argument count and array as
	passed to the main() function on program invocation.  OPTIONS is an
	array (can be NULL) of structures describing the recognized options.

	Recognized long option names are given by OPTIONS[N].name and the
	possible abreviations for that Nth option, if any, are the characters
	of the string OPTIONS[N].abbrev (characters `:' or `?' are forbidden).
	In the ARGV array, long option names are elements that are prefixed by
	`--' and abbreviated options are elements that prefixed by `-'.  Short
	option switches consist in single characters just after the initial
	`-' and can be grouped in a single argument.
	
	For options that have OPTIONS[N].has_arg set to GPL_OPT_ARG_OPTIONAL, an
	optional option argument can be specified just after the short option
	character or after an `=' in a long option and consists in all the
	text of the ARGV element after the option character or the `='.
	Ortherwise the optional argument is assumed to be missing.
	
	For options that have OPTIONS[N].has_arg set to GPL_OPT_ARG_REQUIRED,
	the required option argument can be specified like an optional one or,
	if missing, is taken to be the next ARGV element.
   
	OPTIONS[N].value is the value returned `gpl_opt_next()' when this option
	is encountered.

	The end of the option array is marked by an `gpl_opt_def_t' structure
	with both `name' and `abbrev' set to NULL.  */

extern int gpl_opt_next(gpl_opt_data_t *opt);
/*	Parses next element of OPT->argv and returns a value that describes
	the index/type of the parsed element.

	Each element of ARGV that starts with `-' or `--' (and that is not
	exactly `-' or `--') is a short (abbreviated) or a long option switch
	respectively.  Short option switches consist in single characters just
	after the initial `-' and can be grouped in a single argument.

	If `gpl_opt_next()' find that next argument is an option, it stores its
	option index in OPT->opt_index and returns the option value
	OPT->options[OPT->opt_index].value.  If that option has an argument it
	is stored in OPT->opt_arg; otherwise OPT->opt_arg is set to NULL.
	When all options get parsed, `gpl_opt_next()' returns
	GPL_OPT_END_OF_OPTIONS.  In case of error, `gpl_opt_next()' returns
	GPL_OPT_MISSING_ARG, if a required option argument is missing, or
	ETL_UNKNOWN_OPTION for an unknown option and stores related error
	message that can be retrieved by calling gpl_error_info().

	If the parsing mode is GPL_OPT_RETURN_IN_ORDER, `gpl_opt_next()' returns
	GPL_OPT_NOT_AN_OPTION for each non-option argument and store the
	argument in OPT->opt_arg.  If the parsing mode is
	GPL_OPT_POSIXLY_CORRECT, the parsing of options stops as soon as a
	non-option argument is encountered.  If the parsing mode is
	GPL_OPT_PERMUTE_ARGS, `gpl_opt_next()' permutes options (and their
	argument if any) so that, when `gpl_opt_next()' returns
	GPL_OPT_END_OF_OPTIONS, all non-option argument fill the end of ARGV and
	all options have been parsed.  In any parsing mode, `--' marks the end
	of the options unless it is the argument of an option.
   
	Option syntax summary:
	  --           means end-of-options
	  -            is a valid argument

	Long option syntax:
	  --NAME       is a long name option without argument
	  --NAME VALUE is a long name option with a required argument
	  --NAME=VALUE is a long name option with a required/optional argument

	Short option syntax: the required argument of a short option consists
	in the text after the option character or the next argument as in:
	  -abcd        can be: options `a', `b' and `c'
	                       option `a' with argument `bc'
	                       option `a' and option `b' with argument `c'
   
	`gpl_opt_next()' returns the index in the option array OPT->options of
	the parsed option, or one of the following values: */
#define GPL_OPT_NOT_AN_OPTION        -1 /* for a non option argument */
#define GPL_OPT_END_OF_OPTIONS       -2 /* when end of option is encountered */
#define GPL_OPT_MISSING_ARG          -3 /* for a missing required argument */
#define GPL_OPT_SPURIOUS_ARG         -4 /* when an argument was supplied for an
					 option that has no argument */
#define GPL_OPT_UNKNOWN_SHORT_OPTION -5 /* for an unknown short option */
#define GPL_OPT_UNKNOWN_LONG_OPTION  -6 /* for an unknown long option */

/*---------------------------------------------------------------------------*/
#ifdef __cplusplus
} /* C++ */
#endif
#endif /* GPL_OPTION_H --------------------------------------------------------*/
