/*
 * gpl_shm.c --
 *
 *	Management of shared memory in E-library.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include "gpl_base.h"

#ifndef HAVE_SHM
# define HAVE_SHM 0
#endif

#if HAVE_SHM
# include <sys/ipc.h>
# include <sys/shm.h>
#endif


/*---------------------------------------------------------------------------*/
/* SHARED MEMORY
   =============

   E-library provides mangement of memory segment that can be shared by
   different processes (on the same machine).

   The owner of the shared memory segment allocate memory with gpl_shm_alloc();
   other processes can attach the shared memory to their data segment.
   Eventually, the shared memory segment is detached with gpl_shm_detach() by
   the owner and the other processes.  The memory is not destroyed until the
   last detach.  The management of shared memory in E-library guarantees that
   allocated segments will be destroyed after the last detach (either explicit
   or because of process exit).

   In order to make the following routines similar to the standard library
   malloc/free; E-library reserves some bytes at the bottom of the shared
   memory segment to store some information (actually the shmid and a magic
   number for quick sanity check); this `private header' is below the address
   PTR of the public area which is the only address used by the E-library
   shared memory routines.  This limit the usage of such shared memory to
   E-library based applications. */

/* Structure to store shared memory information. */
typedef struct _gpl_shm_info gpl_shm_info_t;
struct _gpl_shm_info {
    int shm_id;                      /* shared memory identifier */
#if 0 /* IPC key is private */
    int ipc_key;                     /* IPC key */
#endif
    unsigned short mode;             /* lower 9 bits of access modes */
    unsigned short owner_uid;        /* effective user ID of owner */
    unsigned short owner_gid;        /* effective group ID of owner */
    unsigned short creator_uid;      /* effective user ID of creator */
    unsigned short creator_gid;      /* effective group ID of creator */
    int	creator_pid;                 /* process ID of creator */
    int size;                        /* size of the public area */
    int references;                  /* no. of current attaches */
    int	last_operator_pid;           /* process ID of last operator */
    unsigned long last_attach_time;  /* last attach time */
    unsigned long last_detach_time;  /* last detach time */
    unsigned long last_change_time;  /* last change time */
};

extern void *gpl_shm_alloc(int size, int perms);
/* Return the address of the public area of a shared memory segment with
   access permissions PERMS.  The public area can hold at least (since shared
   memory segment are subject to rounding up to PAGE_SIZE bytes), SIZE bytes.
   In case of error, NULL is returned and a message is left that can be
   retrieved by gpl_error_info. */

extern void *gpl_shm_attach(int id, int flags);
/* Attach a shared memory segment identified by ID to the data segment of the
   calling process; the FLAGS can be 0 or SHM_RDONLY (read-only).  Return the
   address of the public area of the shared memory segment or NULL in case of
   error (error message can be recovered with gpl_error_info).  When no longer
   needed, the shared memory segment should be freed (detached) by
   gpl_shm_detach().
   
   RESTRICTIONS: ID must identify a shared memory segment that have been
     allocated with gpl_shm_alloc (because the management of shared memory in
     E-library is specific). */

extern int gpl_shm_detach(void *ptr);
/* Detach (free) shared memory segment at PTR.  Return standard result: GPL_OK
   or GPL_ERROR.  If PTR is NULL, nothing is done. */

extern int gpl_shm_check(void *ptr);
/* Check validity of address PTR of public area of shared memory managed by
   E-library.  Return standard result: GPL_OK or GPL_ERROR. */

extern int gpl_shm_query_info(void *ptr, gpl_shm_info_t *info);
/* Check shared memory at PTR and store its descriptor in INFO (see man page
   of shmctl for more information).  Return standard result: GPL_OK or
   GPL_ERROR. */

extern int gpl_shm_config(void *ptr, int perms);
/* Configure shared memory at PTR so that the segment has access permissions
   PERMS (only the lowest 9 bits of PERMS are used).  Return standard result:
   GPL_OK or GPL_ERROR. */

/*---------------------------------------------------------------------------*/

/* SHARED_MEMORY_OFFSET is the size of private shared memory header (this
   private area is used to store the shared memory identifier and a magic
   number for quick integrity check of the shared memory); must be a multiple
   of a double and must be large enough to store 2 int's. */
#if 0
#define SHARED_MEMORY_OFFSET 8
#else
#undef MAX
#define MAX(a,b)  ((a)>=(b)?(a):(b))
static const int SHARED_MEMORY_OFFSET= MAX(sizeof(double), 2*sizeof(int));
#endif

/* Magic number for shared memory sanity check. */
#define SHARED_MEMORY_MAGIC ((unsigned int)0xDEADC0DE)

/* Macros to access the private shared memory header (PTR is the address of
   the public part of the shared memory). */
#define ShmIdent(ptr) (((int *)(ptr))[-2])
#define ShmMagic(ptr) (((unsigned int *)(ptr))[-1])
#define ShmAddr(ptr)  (((char *)(ptr))-SHARED_MEMORY_OFFSET)

/* Macros to check integrity of shared memory (PTR is the address of the
   public part of the shared memory). */
#define BadJustification(ptr) (((int)(ptr))%SHARED_MEMORY_OFFSET!=0)
#define BadMagic(ptr)         (ShmMagic(ptr)!=SHARED_MEMORY_MAGIC)
#define BadIdent(ptr)         (ShmIdent(ptr)<0)


#if HAVE_SHM
static int get_info(void *ptr, struct shmid_ds *info_ptr);
/*	Low-level worker routine for gpl_shm_config and gpl_shm_query_info that
	checks address PTR of public shared memory area and retrieve system
	information.  Returns GPL_OK or GPL_ERROR (with a suitable error
	message). */
#else
static int shm_not_implemented PROTO((void));
#endif

/*---------------------------------------------------------------------------*/

int gpl_shm_detach(void *ptr)
{
#if HAVE_SHM
    if (ptr!=NULL) {
	if (BadJustification(ptr) || BadMagic(ptr) || BadIdent(ptr)) {
	    return gpl_error("corrupted shared memory");
	}
	if (shmdt(ShmAddr(ptr)) != 0) {
	    return gpl_sys_error("unable to detach shared memory");
	}
    }
    return GPL_OK;
#else
    return shm_not_implemented();
#endif
}

void *gpl_shm_attach(int id, int flags)
{
#if HAVE_SHM
    char *ptr;

    if ((ptr= shmat(id, NULL, flags)) == (char *)-1) {
	gpl_sys_error("unable to attach shared memory");
	return NULL;
    }
    ptr+= SHARED_MEMORY_OFFSET;
    if (BadJustification(ptr) || ShmIdent(ptr)!=id || BadMagic(ptr)) {
	(void)shmdt(ShmAddr(ptr));
	gpl_error("corrupted shared memory or segment not managed by GPLib");
	return NULL;
    }
    return ptr;
#else
    shm_not_implemented();
    return NULL;
#endif
}

/*-----------------------------------------------------------------------------
  To allocate shared memory,  proceeds as follow:
   (1) create new segment of shared memory
   (2) attach the new segment
   (3) mark the created segment so that is is removed at last detach.
 */

#undef ROUND_UP
#define ROUND_UP(x,y) ((((x)+((y)-1))/(y))*(y))

void *gpl_shm_alloc(int size, int perms)
{
#if HAVE_SHM
    int id;
    char *ptr;
    
    /* Round-up size with enough space for the private header. */
    if (size <= 0) {
	gpl_error("bad shared memory size");
	return NULL;
    }
    size= ROUND_UP(size+SHARED_MEMORY_OFFSET, SHARED_MEMORY_OFFSET);

    /* Create and attach shared memory segment; also mark the created segment
       so that is is removed at last detach. */
    if ((id= shmget(IPC_PRIVATE, size, (perms&0777)|IPC_CREAT)) == -1) {
	gpl_sys_error("cannot allocate shared memory");
	return NULL;
    }
    if ((ptr= (char *)shmat(id, NULL, 0)) == (char *)-1
	    || shmctl(id, IPC_RMID, NULL) != 0) {
	gpl_error("unexpected shared memory error");
	return NULL;
    }

    /* Setup private header and return address of public area. */
    ShmIdent(ptr)= id;
    ShmMagic(ptr)= SHARED_MEMORY_MAGIC;
    return ptr+SHARED_MEMORY_OFFSET;
#else
    shm_not_implemented();
    return NULL;
#endif
}

int gpl_shm_config(void *ptr, int perms)
{
#if HAVE_SHM
    struct shmid_ds shmds;

    if (get_info(ptr, &shmds) != GPL_OK) {
	return GPL_ERROR;
    }
    perms&= 0777; /* Only the lowest 9 bits of perms are used. */
    if ((shmds.shm_perm.mode & 0777) != perms) {
	shmds.shm_perm.mode= perms;
    	if (shmctl(ShmIdent(ptr), IPC_SET, &shmds) != 0) {
	    return gpl_sys_error("unable to get shared memory permissions");
	}
    }
    return GPL_OK;
#else
    return shm_not_implemented();
#endif
}

int gpl_shm_query_info(void *ptr, gpl_shm_info_t *info_ptr)
{
#if HAVE_SHM
    struct shmid_ds ds;
    
    if (get_info(ptr, &ds) != GPL_OK) {
	return GPL_ERROR;
    }
    if (info_ptr!=NULL) {
	info_ptr->shm_id=            ShmIdent(ptr);
#if 0 /* do not copy this private(?) IPC key */ 
	info_ptr->ipc_key=           ds.shm_perm.key;
	info_ptr->ipc_key=           ds.shm_perm.__key; /* Linux */
#endif
	info_ptr->mode=              ds.shm_perm.mode;
	info_ptr->owner_uid=         ds.shm_perm.uid;
	info_ptr->owner_gid=         ds.shm_perm.gid;
	info_ptr->creator_uid=       ds.shm_perm.cuid;
	info_ptr->creator_gid=       ds.shm_perm.cuid;
	info_ptr->creator_pid=       ds.shm_cpid;
	info_ptr->size=              ds.shm_segsz - SHARED_MEMORY_OFFSET;
	info_ptr->references=        ds.shm_nattch;
	info_ptr->last_operator_pid= ds.shm_lpid;
	info_ptr->last_attach_time=  ds.shm_atime;
	info_ptr->last_detach_time=  ds.shm_dtime;
	info_ptr->last_change_time=  ds.shm_ctime;
    }
    return GPL_OK;
#else
    return shm_not_implemented();
#endif
}

int gpl_shm_check(void *ptr)
{
#if HAVE_SHM
    struct shmid_ds shmds;
    return get_info(ptr, &shmds);
#else
    return shm_not_implemented();
#endif
}

#if HAVE_SHM
static int get_info(void *ptr, struct shmid_ds *info_ptr)
{
    if (ptr==NULL || BadJustification(ptr) || BadMagic(ptr) || BadIdent(ptr)) {
badAddress:
	return gpl_error("invalid/corrupted shared memory address");
    }
    if (shmctl(ShmIdent(ptr), IPC_STAT, info_ptr) != 0) {
	return gpl_sys_error("unable to shared memory information");
    }
    if (info_ptr->shm_segsz <= SHARED_MEMORY_OFFSET) {
	goto badAddress;
    }
    return GPL_OK;
}
#else
static int shm_not_implemented()
{
    return gpl_error("no support for shared memory on this machine");
}
#endif
