/*
 * gpl_workspace.c --
 *
 *	Portable routines to manage temporary workspaces that can be
 *	automatically deallocated.  These routines are intended to be robust
 *	against interrupts.
 *
 * Copyright:
 *	Eric THIEBAUT, 1998.
 *
 *	The handling of the stack direction and the function i00afunc() are
 *	borrowed from GNU `alloca.c' code by D. A. Gwyn and is therefore
 *	subject to the GPL (GNU Public License).
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include <stdio.h>
/*#include <sys/types.h>*/
#include "gpl_workspace.h"

#ifndef NULL 
# define NULL 0
#endif

#if defined(__STDC__) || defined(__cplusplus)
# define ANSI_C 1
#else
# define ANSI_C 0
#endif

#if ANSI_C
extern void *malloc(size_t size);
extern void  free(void *ptr);
extern void *memcpy(void *dst, const void *src, size_t size);
#else
extern char *malloc();
extern void  free();
extern char *memcpy();
#endif

/*---------------------------------------------------------------------------*/
/* If your stack is a linked list of frames, you have to provide an "address
   metric" ADDRESS_FUNCTION macro.  */

#if defined (CRAY) && defined (CRAY_STACKSEG_END)
# if ANSI_C
static long i00afunc(void *addr);
#else
static long i00afunc();
#endif
# define ADDRESS_FUNCTION(arg) ((char *)i00afunc(&(arg)))
#else
# define ADDRESS_FUNCTION(arg) ((char *)&(arg))
#endif

/* Define STACK_DIRECTION if you know the direction of stack growth for your
   system; otherwise it will be automatically deduced at run-time.

   STACK_DIRECTION > 0 => grows toward higher addresses
   STACK_DIRECTION < 0 => grows toward lower addresses
   STACK_DIRECTION = 0 => direction of growth unknown  */
#ifndef STACK_DIRECTION
# define STACK_DIRECTION 0  /* Stack direction unknown.  */
#endif

#if STACK_DIRECTION != 0    /* Stack direction known at compile-time. */

# define STACK_DIR STACK_DIRECTION

#else /* STACK_DIRECTION == 0; need run-time code.  */

# define STACK_DIR stack_direction

static int stack_direction= 0;		/* 1 or -1 once known.  */
static void find_stack_direction(void);
static void find_stack_direction(void)
{
  static char *addr= NULL;	/* Address of first `dummy', once known.  */
  auto char probe;		/* To get stack address.  */

  if (addr == NULL) {
    /* Initial entry.  */
    addr= ADDRESS_FUNCTION(probe);
    find_stack_direction();	/* Recurse once.  */
  } else {
    /* Second entry.  */
    if (ADDRESS_FUNCTION(probe) > addr) {
      stack_direction= 1;		/* Stack grew upward.  */
    } else {
      stack_direction= -1;		/* Stack grew downward.  */
    }
  }
}

#endif /* STACK_DIRECTION == 0 */

/*---------------------------------------------------------------------------*/
/* Structure to store workspace information. */
typedef struct workspace workspace_t;
struct workspace {
    void          *data;
    gpl_ws_delete_t *delete;
    char          *depth;
};

/* Private data. */
static workspace_t   *ws_list= NULL;  /* workspace list */
static int            ws_size= 0;     /* number of slots in ws_list */
static int            ws_count= 0;    /* number of occupied slots in ws_list */

/*---------------------------------------------------------------------------*/

#if ANSI_C
char *gpl_ws_cleanup(register char *depth)
#else
char *gpl_ws_cleanup(depth)
     register char *depth;
#endif
{
  gpl_ws_delete_t *delete;
  int i= ws_count;
  workspace_t *ws= ws_list;

  if (depth==NULL) {
    /* Probes stack depth. */
    auto char probe;
    depth= ADDRESS_FUNCTION(probe);
  }
#undef CLEANUP
#define CLEANUP(CMP)							      \
  while (--i >= 0 && ws[i].depth CMP depth) {				      \
    if ((delete= ws[i].delete) != NULL) {				      \
      ws[i].delete= NULL;						      \
      delete(ws[i].data);						      \
    }									      \
  }

  /* Cleanup: free all workspaces attached at or deeper than depth. */
#if STACK_DIRECTION==0
  if (stack_direction==0) {
    /* Unknown growth direction. */
    find_stack_direction();
  }
  if (stack_direction > 0) {
    CLEANUP(>=)
  } else {
    CLEANUP(<=)
  }
#else /* STACK_DIRECTION != 0 */
# if STACK_DIRECTION>0
  CLEANUP(>=)
# else
  CLEANUP(<=)
# endif
#endif /* STACK_DIRECTION != 0 */

  ws_count= i+1;
  return depth;
}

/*---------------------------------------------------------------------------*/

/* Augment workspace list (only called by gpl_ws_attach but kept separately to
   all for inline with optimization -finline-functions). */
static int grow_list(void);
static int grow_list(void)
{
#undef MIN_SIZE
#define MIN_SIZE 32
  unsigned int new_size= (ws_size>MIN_SIZE) ? 2*ws_size : MIN_SIZE;
  workspace_t *new_list= (workspace_t *)malloc(new_size*sizeof(workspace_t));
  if (new_list==NULL) {
    gpl_error("not enough memory to augment workspace list");
    return 1;
  }
  if (ws_list!=NULL) {
    workspace_t *old_list= ws_list;
    memcpy(new_list, old_list, sizeof(workspace_t)*ws_count);
    ws_list= new_list;
    ws_size= new_size;
    free(old_list);
  } else {
    ws_list= new_list;
    ws_size= new_size;
  }
  return 0;
}

#if ANSI_C
int gpl_ws_attach(char *depth, void *data, gpl_ws_delete_t *delete)
#else
int gpl_ws_attach(depth, data, Free)
     char          *depth;
     char          *data;
     gpl_ws_delete_t *delete;
#endif
{
  /* Cowardly refuse to attach workspace at a depth that is not at least
     as deep as the last workspace in list. */
  if (ws_count>=1) {
#if STACK_DIRECTION==0
    if (stack_direction==0) {
      /* Unknown growth direction. */
      find_stack_direction();
    }
#endif
    if (STACK_DIR * (depth - ws_list[ws_count-1].depth) < 0) {
      return gpl_error("bad stack depth");
    }
  }
  
  /* Augment list size if needed (do not use realloc in case it is
     not robust against interrupts). */
  if (ws_count>=ws_size && grow_list()) return GPL_ERROR;
  
  /* Register the workspace in the new slot (update count at the end). */
  ws_list[ws_count].depth= depth;
  ws_list[ws_count].data= data;
  ws_list[ws_count++].delete= delete;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

#if ANSI_C
int gpl_ws_detach(void *data)
#else
int gpl_ws_detach(data)
     char *data;
#endif
{
  int i= ws_count-1;

  /* Fast operation if DATA matches the last attached one (also reduce the
     number of used slots). */
  if (ws_list[i].data == data) {
    ws_list[i].delete= NULL;
    ws_count-= 1;
    return GPL_OK;
  }
  
  /* Scan rest of the list until DATA matches. */
  while (--i>=0) {
    if (ws_list[i].data == data) {
      ws_list[i].delete= NULL;
      return GPL_OK;
    }
  }
  return gpl_error("cannot detach workspace: unregistered data");
}

/*---------------------------------------------------------------------------*/

#if ANSI_C
int gpl_ws_delete(void *data)
#else
int gpl_ws_delete(data)
    char *data;
#endif
{
  gpl_ws_delete_t *delete;
  int i= ws_count-1;
  
  /* Fast operation if DATA matches the last attached one (also reduce the
     number of used slots). */
  if (ws_list[i].data == data) {
    if ((delete= ws_list[i].delete) != NULL) {
      /* Better to never delete than delete twice. */
      ws_list[i].delete= NULL;
      (*delete)(ws_list[i].data);
    }
    ws_count-= 1;
    return GPL_OK;
  }

  /* Scan rest of the list until DATA matches. */
  while (--i>=0) {
    if (ws_list[i].data == data) {
      if ((delete= ws_list[i].delete) != NULL) {
	/* Better to never delete than delete twice. */
	ws_list[i].delete= NULL;
	(*delete)(ws_list[i].data);
      }
      return GPL_OK;
    }
  }
  return gpl_error("cannot delete workspace: unregistered data");
}

/*---------------------------------------------------------------------------*/

#if ANSI_C
void *gpl_ws_create(char *depth, size_t size)
#else
char *gpl_ws_create(depth, size)
     char *depth;
     size_t size)
#endif
{
  void *ptr;

  if (size!=0 && (ptr= malloc(size)) != NULL) {
    if (gpl_ws_attach(depth, ptr, free) == GPL_OK) {
      return ptr;
    }
    free(ptr);
    return NULL;
  }
  gpl_error("bad size or not enough memory for workspace");
  return NULL;
}

/*---------------------------------------------------------------------------*/

#if GPL_WS_NO_ELIB
static GPL_WS_CONST char *ws_error= NULL; /* last error message */

#if ANSI_C
const char *gpl_error_info(int clear)
#else
char *gpl_error_info(clear)
    int clear;
#endif
{
  if (ws_error) {
    GPL_WS_CONST char *info= ws_error;
    if (clear) ws_error= NULL;
    return info;
  }
  return "no error message from workspace routines";
}

#if ANSI_C
int gpl_error(const char *message)
#else
int gpl_error(message)
     char *message;
#endif
{
  ws_error= message;
}

#endif /* GPL_WS_NO_ELIB */

/*---------------------------------------------------------------------------*/

void gpl_ws_info()
{
  int i;
  
  if (ws_count==0) {
    fprintf(stderr, "*** empty workspace list ***\n");
    return;
  }
  
  if (ws_count<0) {
    fprintf(stderr, "*** ERROR: WORKSPACE LIST CORRUPTED ***\n");
    ws_count= 0;
    return;
  }
  
  fprintf(stderr, "*** stack direction= %d\n", STACK_DIR);
  fprintf(stderr, "*** index   depth    data     free\n");
  fprintf(stderr, "*** ----- -------- -------- --------\n");
  for (i=ws_count-1 ; i>=0 ; i--) {
    fprintf(stderr, "*** %5d %8u 0x%08x 0x%08x\n", i,
	    (unsigned)ws_list[i].depth,
	    (unsigned)ws_list[i].data,
	    (unsigned)ws_list[i].delete);
  }
}

/*---------------------------------------------------------------------------*/

#if defined (CRAY) && defined (CRAY_STACKSEG_END)

#ifdef DEBUG_I00AFUNC
#include <stdio.h>
#endif

#ifndef CRAY_STACK
#define CRAY_STACK
#ifndef CRAY2
/* Stack structures for CRAY-1, CRAY X-MP, and CRAY Y-MP */
struct stack_control_header {
    long shgrow:32;		/* Number of times stack has grown.  */
    long shaseg:32;		/* Size of increments to stack.  */
    long shhwm:32;		/* High water mark of stack.  */
    long shsize:32;		/* Current size of stack (all segments).  */
};

/* The stack segment linkage control information occurs at
   the high-address end of a stack segment.  (The stack
   grows from low addresses to high addresses.)  The initial
   part of the stack segment linkage control information is
   0200 (octal) words.  This provides for register storage
   for the routine which overflows the stack.  */

struct stack_segment_linkage {
    long ss[0200];		/* 0200 overflow words.  */
    long sssize:32;		/* Number of words in this segment.  */
    long ssbase:32;		/* Offset to stack base.  */
    long:32;
    long sspseg:32;		/* Offset to linkage control of previous
				   segment of stack.  */
    long:32;
    long sstcpt:32;		/* Pointer to task common address block.  */
    long sscsnm;		/* Private control structure number for
				   microtasking.  */
    long ssusr1;		/* Reserved for user.  */
    long ssusr2;		/* Reserved for user.  */
    long sstpid;		/* Process ID for pid based multi-tasking.  */
    long ssgvup;		/* Pointer to multitasking thread giveup.  */
    long sscray[7];		/* Reserved for Cray Research.  */
    long ssa0;
    long ssa1;
    long ssa2;
    long ssa3;
    long ssa4;
    long ssa5;
    long ssa6;
    long ssa7;
    long sss0;
    long sss1;
    long sss2;
    long sss3;
    long sss4;
    long sss5;
    long sss6;
    long sss7;
};

#else /* CRAY2 */
/* The following structure defines the vector of words
   returned by the STKSTAT library routine.  */
struct stk_stat {
    long now;			/* Current total stack size.  */
    long maxc;			/* Amount of contiguous space which would
				   be required to satisfy the maximum
				   stack demand to date.  */
    long high_water;		/* Stack high-water mark.  */
    long overflows;		/* Number of stack overflow ($STKOFEN) calls.  */
    long hits;			/* Number of internal buffer hits.  */
    long extends;		/* Number of block extensions.  */
    long stko_mallocs;		/* Block allocations by $STKOFEN.  */
    long underflows;		/* Number of stack underflow calls ($STKRETN).  */
    long stko_free;		/* Number of deallocations by $STKRETN.  */
    long stkm_free;		/* Number of deallocations by $STKMRET.  */
    long segments;		/* Current number of stack segments.  */
    long maxs;			/* Maximum number of stack segments so far.  */
    long pad_size;		/* Stack pad size.  */
    long current_address;	/* Current stack segment address.  */
    long current_size;		/* Current stack segment size.  This
				   number is actually corrupted by STKSTAT to
				   include the fifteen word trailer area.  */
    long initial_address;	/* Address of initial segment.  */
    long initial_size;		/* Size of initial segment.  */
};

/* The following structure describes the data structure which trails
   any stack segment.  I think that the description in 'asdef' is
   out of date.  I only describe the parts that I am sure about.  */

struct stk_trailer {
    long this_address;		/* Address of this block.  */
    long this_size;		/* Size of this block (does not include
				   this trailer).  */
    long unknown2;
    long unknown3;
    long link;			/* Address of trailer block of previous
				   segment.  */
    long unknown5;
    long unknown6;
    long unknown7;
    long unknown8;
    long unknown9;
    long unknown10;
    long unknown11;
    long unknown12;
    long unknown13;
    long unknown14;
};

#endif /* CRAY2 */
#endif /* not CRAY_STACK */

#ifdef CRAY2
/* Determine a "stack measure" for an arbitrary ADDRESS.
   I doubt that "lint" will like this much. */

static long
i00afunc (long *address)
{
  struct stk_stat status;
  struct stk_trailer *trailer;
  long *block, size;
  long result = 0;

  /* We want to iterate through all of the segments.  The first
     step is to get the stack status structure.  We could do this
     more quickly and more directly, perhaps, by referencing the
     $LM00 common block, but I know that this works.  */

  STKSTAT (&status);

  /* Set up the iteration.  */

  trailer = (struct stk_trailer *) (status.current_address
				    + status.current_size
				    - 15);

  /* There must be at least one stack segment.  Therefore it is
     a fatal error if "trailer" is null.  */

  if (trailer == 0)
    abort ();

  /* Discard segments that do not contain our argument address.  */

  while (trailer != 0)
    {
      block = (long *) trailer->this_address;
      size = trailer->this_size;
      if (block == 0 || size == 0)
	abort ();
      trailer = (struct stk_trailer *) trailer->link;
      if ((block <= address) && (address < (block + size)))
	break;
    }

  /* Set the result to the offset in this segment and add the sizes
     of all predecessor segments.  */

  result = address - block;

  if (trailer == 0)
    {
      return result;
    }

  do
    {
      if (trailer->this_size <= 0)
	abort ();
      result += trailer->this_size;
      trailer = (struct stk_trailer *) trailer->link;
    }
  while (trailer != 0);

  /* We are done.  Note that if you present a bogus address (one
     not in any segment), you will get a different number back, formed
     from subtracting the address of the first block.  This is probably
     not what you want.  */

  return (result);
}

#else /* not CRAY2 */
/* Stack address function for a CRAY-1, CRAY X-MP, or CRAY Y-MP.
   Determine the number of the cell within the stack,
   given the address of the cell.  The purpose of this
   routine is to linearize, in some sense, stack addresses
   for alloca.  */

static long
i00afunc (long address)
{
  long stkl = 0;

  long size, pseg, this_segment, stack;
  long result = 0;

  struct stack_segment_linkage *ssptr;

  /* Register B67 contains the address of the end of the
     current stack segment.  If you (as a subprogram) store
     your registers on the stack and find that you are past
     the contents of B67, you have overflowed the segment.

     B67 also points to the stack segment linkage control
     area, which is what we are really interested in.  */

  stkl = CRAY_STACKSEG_END ();
  ssptr = (struct stack_segment_linkage *) stkl;

  /* If one subtracts 'size' from the end of the segment,
     one has the address of the first word of the segment.

     If this is not the first segment, 'pseg' will be
     nonzero.  */

  pseg = ssptr->sspseg;
  size = ssptr->sssize;

  this_segment = stkl - size;

  /* It is possible that calling this routine itself caused
     a stack overflow.  Discard stack segments which do not
     contain the target address.  */

  while (!(this_segment <= address && address <= stkl))
    {
#ifdef DEBUG_I00AFUNC
      fprintf (stderr, "%011o %011o %011o\n", this_segment, address, stkl);
#endif
      if (pseg == 0)
	break;
      stkl = stkl - pseg;
      ssptr = (struct stack_segment_linkage *) stkl;
      size = ssptr->sssize;
      pseg = ssptr->sspseg;
      this_segment = stkl - size;
    }

  result = address - this_segment;

  /* If you subtract pseg from the current end of the stack,
     you get the address of the previous stack segment's end.
     This seems a little convoluted to me, but I'll bet you save
     a cycle somewhere.  */

  while (pseg != 0)
    {
#ifdef DEBUG_I00AFUNC
      fprintf (stderr, "%011o %011o\n", pseg, size);
#endif
      stkl = stkl - pseg;
      ssptr = (struct stack_segment_linkage *) stkl;
      size = ssptr->sssize;
      pseg = ssptr->sspseg;
      result += size;
    }
  return (result);
}

#endif /* not CRAY2 */
#endif /* CRAY */

/*---------------------------------------------------------------------------*/
