/*
 * gpl_xdr.c --
 *
 *	XDR (eXternal Data Representation) input/output routines for GPLib.
 *	This package can be compiled alone (i.e. out of GPLib).
 *
 * Warning:
 *	This package may not be portable since it assumes that at most byte
 *	swapping is needed for your machine data layout to match XDR (works
 *	on most IEEE machines: PC's, Sun, IBM or HP workstations, ...).
 *	You may also have to modify some typedefs...
 *
 * Copyright and warranty:
 *	Copyright (c) 1995-2001 Eric THIEBAUT.
 *
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef GPL_XDR_C
#define GPL_XDR_C 1
#include "gpl_xdr.h"

/* Figure out whether byte swapping is needed or not. */
#undef SWAP_BYTES
#if (GPL_BYTE_ORDER == GPL_BIG_ENDIAN)
# define SWAP_BYTES 0
#else
# if (GPL_BYTE_ORDER == GPL_LITTLE_ENDIAN)
#  define SWAP_BYTES 1
# else
#  error "unsupported byte order"
# endif
#endif

/*---------------------------------------------------------------------------*/

#undef gpl_xdr_read_byte
int gpl_xdr_read_byte(void *ptr, int number, FILE *stream)
{
  return fread(ptr, 1, number, stream);
}

#undef gpl_xdr_write_byte
int gpl_xdr_write_byte(const void *ptr, int number, FILE *stream)
{
  return fwrite(ptr, 1, number, stream);
}

/*---------------------------------------------------------------------------*/

#if SWAP_BYTES
#undef SWAP_BUF_SIZE
#define SWAP_BUF_SIZE 1024
static char swapBuf[SWAP_BUF_SIZE];
#endif /* SWAP_BYTES */

#undef XDR_SIZE
#undef XDR_READ
#undef XDR_WRITE

#define XDR_SIZE  2
#define XDR_READ  gpl_xdr_read_short
#define XDR_WRITE gpl_xdr_write_short
#include __FILE__

#define XDR_SIZE  4
#define XDR_READ  gpl_xdr_read_long
#define XDR_WRITE gpl_xdr_write_long
#include __FILE__

#define XDR_SIZE  4
#define XDR_READ  gpl_xdr_read_float
#define XDR_WRITE gpl_xdr_write_float
#include __FILE__

#define XDR_SIZE  8
#define XDR_READ  gpl_xdr_read_double
#define XDR_WRITE gpl_xdr_write_double
#include __FILE__

/*---------------------------------------------------------------------------*/
#else /* GPL_XDR_C defined */

#undef SWAP
#define SWAP(i,j) c=data[i];data[i]=data[j];data[j]=c

int XDR_READ(void *ptr, int number, FILE *stream)
{
#if SWAP_BYTES
  if ((number= fread(ptr, XDR_SIZE, number, stream)) > 0) {
    register int i, n= number*XDR_SIZE;
    register char c, *data= (char *)ptr;	
    for (i=0 ; i<n ; i+=XDR_SIZE) {
#if XDR_SIZE==2
      SWAP(i, i+1);
#endif
#if XDR_SIZE==4
      SWAP(i,   i+3);
      SWAP(i+1, i+2);
#endif
#if XDR_SIZE==8
      SWAP(i,   i+7);
      SWAP(i+1, i+6);
      SWAP(i+2, i+5);
      SWAP(i+3, i+4);
#endif
    }
  }
  return number;
#else
  return fread(ptr, XDR_SIZE, number, stream);
#endif
}

int XDR_WRITE(const void *ptr, int number, FILE *stream)
{
#if SWAP_BYTES
  char       *dst;
  const char *src;
  int         n, nwrite, nbuf, nleft;
  
  /* May be no swap needed. */
  if (number==0) return 0;
  
  /* Swap bytes into static buffer, then write bytes. */
  nbuf= (SWAP_BUF_SIZE / XDR_SIZE);
  src= ptr;
  for (nleft= number; nleft > 0; nleft -= nwrite) {
    dst= swapBuf;
    nwrite= nleft < nbuf ? nleft : nbuf;
    for (n=0; n<nwrite; n++, src+=XDR_SIZE, dst+=XDR_SIZE) {
#if XDR_SIZE==2
      dst[0]= src[1];
      dst[1]= src[0];
#endif
#if XDR_SIZE==4
      dst[0]= src[3];
      dst[1]= src[2];
      dst[2]= src[1];
      dst[3]= src[0];
#endif
#if XDR_SIZE==8
      dst[0]= src[7];
      dst[1]= src[6];
      dst[2]= src[5];
      dst[3]= src[4];
      dst[4]= src[3];
      dst[5]= src[2];
      dst[6]= src[1];
      dst[7]= src[0];
#endif
    }
    if ((n= fwrite(swapBuf, XDR_SIZE, nwrite, stream)) != nwrite) {
      nleft -= n > 0 ? n : 0;
      break;
    }
  }
  return number - nleft;
#else
  return fwrite(ptr, XDR_SIZE, number, stream);
#endif
}

#undef XDR_WRITE
#undef XDR_READ
#undef XDR_SIZE
/*---------------------------------------------------------------------------*/
#endif /* GPL_XDR_C */
