/*
 * tclsrv.h --
 *
 *	Definitions for named socket, pipes for TCL (unix).
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2002 Eric THIEBAUT.
 *
 *	This file is part of TclSrv.
 *
 *	TclSrv is free  software; you can redistribute it  and/or modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	TclSrv  is distributed  in the  hope that  it will  be  useful, but
 *	WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 *	MERCHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU
 *	General Public License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with TclSrv (file "COPYING"  in the top source directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330, Boston, MA 02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

/* This  library implements  a simple  way to establish  a connection  to a
   Tcl/Tk command  server.  Once a  client has established a  connection to
   the server (tclsrv_create optionnaly followed by tclsrv_connect), it can
   send scripts  to be evaluated  by the server  and receive the  result of
   that evaluation  (with tclsrv_send).  Named  (local) socket are  used to
   allow for easy control of who is authorized to connect to the server.

   To make  things easy, the  send/receive operations are done  in blocking
   mode or  with a given timeout.   If the timeout  expired, the connection
   get closed.

   The proper way to check an existing connection is to send a no-op script
   (a NULL string or empty "" string).  */

#ifndef _TCLSRV_H
#define _TCLSRV_H 1

/* Definitions allowing this header file to be used either with or without
   ANSI C features like function prototypes.  */
#if defined(__STDC__) || defined(__cplusplus)
#  define TCLSRV_ARGS(x) x
#  define TCLSRV_CONST   const
#else
#  define TCLSRV_ARGS(x)
#  define TCLSRV_CONST
#endif
#ifdef __cplusplus
#  define _TCLSRV_BEGIN_DECLS extern "C" {
#  define _TCLSRV_END_DECLS   }
#else
#  define _TCLSRV_BEGIN_DECLS
#  define _TCLSRV_END_DECLS
#endif

_TCLSRV_BEGIN_DECLS
typedef struct _tclsrv_client_t tclsrv_client_t;

extern tclsrv_client_t *tclsrv_create TCLSRV_ARGS((const char *path));
/* Create a new un-connected client for Tcl command server listening on
   named socket PATH.  NULL is returned if PATH is not valid (too short or
   too long) or if there is not enough memory for the client instance.  If
   the return value is non-NULL, connection can be established with
   tclsrv_connect or the first time tclsrv_send is called.  The instance is
   destroyed with tclsrv_destroy. */

extern void tclsrv_destroy TCLSRV_ARGS((tclsrv_client_t *this));
/* Close connection to Tcl command server, if any, and delete instance;
   THIS no longer exists. */

extern int tclsrv_connect TCLSRV_ARGS((tclsrv_client_t *this, long ms));
/* Connect to Tcl command server listening on named socket PATH and return
   a new structure that can be used to send scripts to the server or NULL
   in case of error.  MS can be used to set the timeout (in milliseconds)
   for establishing the connection :
     if MS < 0, wait forever until connection accepted or refused
     if MS = 0, asynchronous connection: connection will be completed
                later (prior to the first successful tclsrv_send)
     if MS > 0, maximum number of milliseconds to wait for connection
                to complete.
   BUGS: asynchronous connection (MS=0) is not yet implemented.
   The returned value is TCLSRV_OK or TCLSRV_ERROR -- error message
   can be retrieved by tclsrv_result(THIS). */

extern int tclsrv_disconnect TCLSRV_ARGS((tclsrv_client_t *this));
/* Close connection to Tcl command server; THIS can be reconnected (to the
   same server) by tclsrv_connect. The returned value is TCLSRV_OK or
   TCLSRV_ERROR -- error message can be retrieved by tclsrv_result(THIS).*/

extern TCLSRV_CONST char *tclsrv_path TCLSRV_ARGS((tclsrv_client_t *this));
/* Get the name of the local socket connected to this. */

extern int tclsrv_connected TCLSRV_ARGS((tclsrv_client_t *this));
/* Check if THIS is connected.  Useful, if you don't like the re-connect
   feature of tclsrv_send. */

extern int tclsrv_status TCLSRV_ARGS((tclsrv_client_t *this));
/* Get the return code of the last script evaluated by tclsrv_send (same
   value as the one returned by tclsrv_send). */

extern TCLSRV_CONST char *tclsrv_result TCLSRV_ARGS((tclsrv_client_t *this));
/* Get the result of the last script, possibly an error message, evaluated by
   tclsrv_send. */

extern int tclsrv_send TCLSRV_ARGS((tclsrv_client_t *this, long ms,
				    const char *script));
/* Evaluate SCRIPT at global level with the Tcl server connected to client
 * THIS and store the result in in internal buffer of THIS.  If THIS is not
 * connected, re-connection is attempted prior to sending the script.  The
 * caller can set a timeout (in milliseconds) for the result to be received
 * with MS>=0; if MS<0, tclsrv_send will block until the result is received
 * or until an error occurs (e.g. server closed connection) in which case
 * the connection get closed to avoid synchronization problems. The return
 * value is one of:
 *   TCLSRV_OK           - The script completed normally; the result of
 *                         the script is stored in internal buffer of THIS.
 *   TCLSRV_ERROR        - The script couldn't be completed successfully;
 *                         the client's result describes what went wrong (note:
 *                         there are two possible sources of errors: either
 *                         the script itself results in an error, or the
 *                         connection to the server is broken).
 *   TCLSRV_RETURN       - The script requests that the current procedure
 *                         return; the client's result contains the
 *                         procedure's return value.
 *   TCLSRV_BREAK        - The command requests that the innermost loop
 *                         be exited; the client's result is meaningless.
 *   TCLSRV_CONTINUE     - Go on to the next iteration of the current
 *                         loop; the client's result is meaningless.
 *   TCLSRV_INTERRUPTED  - Some interruption occured during communication;
 *                         this will prevent any further exchange (i.e. you'll
 *                         have to re-connect to the server or call
 *                         tclsrv_send again).  */
#define TCLSRV_OK           0
#define TCLSRV_ERROR        1
#define TCLSRV_RETURN       2
#define TCLSRV_BREAK        3
#define TCLSRV_CONTINUE     4
#define TCLSRV_INTERRUPTED -1

#ifdef _TCL

#define TCLSRV_MODE_OWNER    1  /* access granted for user */
#define TCLSRV_MODE_GROUP    2  /* access granted for group */
#define TCLSRV_MODE_OTHER    4  /* access granted for others (DANGEROUS!)*/
#define TCLSRV_MODE_SINGLE   8  /* only one connection per server */
#define TCLSRV_MODE_PRIVATE  (TCLSRV_MODE_OWNER|TCLSRV_MODE_SINGLE)
extern int Tclsrv_CreateCommandServer _ANSI_ARGS_((Tcl_Interp *interp,
						   CONST char *path,
						   int mode));
extern int Tclsrv_ShutdownCommandServer _ANSI_ARGS_((Tcl_Interp *interp,
						     const char *path));

extern int Tclsrv_Init _ANSI_ARGS_((Tcl_Interp *interp));
extern Tcl_ObjCmdProc Tclsrv_PipeObjCmd;
extern Tcl_ObjCmdProc Tclsrv_CommandServerObjCmd;
#endif /* _TCL */
_TCLSRV_END_DECLS

#endif /* _TCLSRV_H ---------------------------------------------------------*/
