/*
 * tclsrv_lib.c --
 *
 *	Implements client part of Tcl command server (TclSrv).
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2002 Eric THIEBAUT.
 *
 *	This file is part of TclSrv.
 *
 *	TclSrv is free  software; you can redistribute it  and/or modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	TclSrv  is distributed  in the  hope that  it will  be  useful, but
 *	WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 *	MERCHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU
 *	General Public License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with TclSrv (file "COPYING"  in the top source directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330, Boston, MA 02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <tcl.h>
#include "tclsrv_private.h"

#ifndef DEBUG
#  define DEBUG 0
#endif
#if DEBUG
#  define DEBUG_MSG(arglist) fprintf arglist
#else
#  define DEBUG_MSG(arglist)
#endif

struct _tclsrv_client_t {
  int           fd;      /* File descriptor of connection socket. */
  int           status;  /* Status of last evaluation. */
  char         *buffer;  /* Buffer to store result or error message. */
  size_t        bufsize; /* Size of buffer. */
  char          path[1]; /* Local socket name (MUST BE THE LAST MEMBER). */
};

static int alloc_buffer TCLSRV_ARGS((tclsrv_client_t *this, size_t minsize));
static int system_error TCLSRV_ARGS((tclsrv_client_t *this,
				     TCLSRV_CONST char *before, int code,
				     TCLSRV_CONST char *after));
static int my_error TCLSRV_ARGS((tclsrv_client_t *this,
				 TCLSRV_CONST char *errmsg));
static int clear_error TCLSRV_ARGS((tclsrv_client_t *this));
static int connection_closed_by_server TCLSRV_ARGS((tclsrv_client_t *this));
static int interrupt_error TCLSRV_ARGS((tclsrv_client_t *this));
static int read_data TCLSRV_ARGS((tclsrv_client_t *this, size_t pending,
				  long ms, struct timeval *timeleft));

/* Define a small minimum buffer size (at least sufficient
   to read message header + 1 byte = 9 bytes). */
#define MIN_BUFSIZE 32

void tclsrv_destroy(this)
     tclsrv_client_t *this;
{
  tclsrv_disconnect(this);
  if (this->buffer) free(this->buffer);
  free(this);
}

int tclsrv_disconnect(this)
     tclsrv_client_t *this;
{
  int fd = this->fd;
  this->fd = -1;
  if (fd >= 0 && close(fd)) return system_error(this, "close socket failed (",
						errno, ")");
  return clear_error(this);
}

TCLSRV_CONST char *tclsrv_result(this)
     tclsrv_client_t *this;
{
  if (this->status == TCLSRV_INTERRUPTED) (void)interrupt_error(this);
  this->buffer[this->bufsize - 1] = 0;
  return this->buffer;
}

TCLSRV_CONST char *tclsrv_path(this)
     tclsrv_client_t *this;
{
  return this->path;
}

int tclsrv_connected(this)
     tclsrv_client_t *this;
{
  if (this->status == TCLSRV_INTERRUPTED) (void)interrupt_error(this);
  return (this->fd >= 0);
}

int tclsrv_status(this)
     tclsrv_client_t *this;
{
  return (this->status == TCLSRV_INTERRUPTED)
    ? interrupt_error(this)
    : this->status;
}
tclsrv_client_t *tclsrv_create(path)
     TCLSRV_CONST char *path;
{
  size_t size, length;
  tclsrv_client_t *this;
  struct sockaddr_un sa;

  if (! path || (length = strlen(path)) <= 0 ||
      length >= sizeof(sa.sun_path)) {
    return NULL; /* bad path */
  }
  size = STRUCT_OFFSET(tclsrv_client_t, path) + length + 1;
  if ((this = malloc(size)) != NULL) {
    memset(this, 0, size);
    if (length) memcpy(this->path, path, length);
    this->path[length] = 0;
    this->fd = -1;
    if (alloc_buffer(this, MIN_BUFSIZE) != TCLSRV_OK) {
      tclsrv_destroy(this);
      return NULL;
    }
    (void)my_error(this, "not connected");
  }
  return this;
}

/*
 * ESTABLISHING THE CONNECTION:
 *   - open client socket
 *   - attempt to connect
 *   - send a first no-op script to check that the connection was accepted
 *     by the server
 *
 *  MS - Timeout for connection (in milliseconds)
 *       if MS < 0, wait forever until connection accepted or refused
 *       if MS = 0, asynchronous connection: connection will be completed
 *                  later (prior to the first successful tclsrv_send)
 *       if MS > 0, maximum number of milliseconds to wait for connection
 *                  to complete
 */
int tclsrv_connect(this, ms)
     tclsrv_client_t *this;
     long             ms;
{
  struct sockaddr_un sa;
  if (this->status == TCLSRV_INTERRUPTED) (void)interrupt_error(this);
  if (this->fd >= 0) {
    /* Already connected. */
    return clear_error(this);
  }

  /* Create client socket. */
  if ((this->fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
    return system_error(this, "create socket failed (", errno, ")");
  }

  /* Set socket i/o buffers. */
# if defined(SOCKET_BUFSIZE) && SOCKET_BUFSIZE > 0
  {
    int value;
    socklen_t value_length;
    value_length = sizeof(value);
    (void)getsockopt(this->fd, SOL_SOCKET, SO_SNDBUF, &value, &value_length);
    if (value < SOCKET_BUFSIZE) {
      value = SOCKET_BUFSIZE;
      (void)setsockopt(this->fd, SOL_SOCKET, SO_SNDBUF, &value, sizeof(value));
    }
    value_length = sizeof(value);
    (void)getsockopt(this->fd, SOL_SOCKET, SO_RCVBUF, &value, &value_length);
    if (value < SOCKET_BUFSIZE) {
      value = SOCKET_BUFSIZE;
      (void)setsockopt(this->fd, SOL_SOCKET, SO_RCVBUF, &value, sizeof(value));
    }
  }
# endif /* SOCKET_BUFSIZE */

  /* Setup socket variables (address family; name of server socket) and
     connect to the server. */
  sa.sun_family = AF_LOCAL;
  strcpy(sa.sun_path, this->path);
  if (connect(this->fd, (struct sockaddr *)&sa, sizeof(sa))) {
    /* Connection refused? */
    tclsrv_disconnect(this);
    return system_error(this, "connection failed (", errno, ")");
  }

  /* Send no-op script to check the connection. */
  if (tclsrv_send(this, ms, "") != TCLSRV_OK && this->status != TCLSRV_ERROR) {
    tclsrv_disconnect(this);
    my_error(this, "evaluation of no-op script failed");
  }
  return this->status;
}

/* SENDING DATA
     For a sockect/pipe channel, the SIGPIPE signal will be received if the
     peer has closed the reading end of the channel and an attempt is made
     to write data to the peer.  For that reason I choose to set the
     INTERRUPTED status prior to sending the data and to restore it only if
     everuthing went ok.  Another possibility would be to ignore the signal
     which causes the write operation to returns an error with errno set to
     EPIPE.  We therefore disable the SIGPIPE signal prior to writing and
     restore previous signal handler after the write.  This is illustrated
     by the following example: */
#if 0
static void sample()
{
   void (*handler) TCLSRV_ARGS((int));
   handler = signal(SIGPIPE, SIG_IGN);
   nbytes = write(this->sock, this->buffer + this->offset, this->pending);
   if (handler != SIG_IGN) signal(SIGPIPE, handler);
}
#endif

int tclsrv_send(this, ms, script)
     tclsrv_client_t *this;
     long ms;
     TCLSRV_CONST char *script;
{
  unsigned char *hdr;
  char *address;
  char *errmsg=NULL;
  tclsrv_uint_t length, pending;
  ssize_t nbytes;
  int flags, status;
  struct timeval timeleft;

  if (this->status == TCLSRV_INTERRUPTED || this->fd < 0) {
    /* Attempt to re-connect. */
    if (tclsrv_connect(this, ms) != TCLSRV_OK) return this->status;
  }

  /* Prepare message to send.  Copy header and script in internal buffer
     to allow send in one call. */
  length = script ? strlen(script) : 0;
  pending = length + 8; /* script length + header */
  if (pending > this->bufsize && alloc_buffer(this, pending) != TCLSRV_OK) {
    return system_error(this, "insufficient memory to send message ",
			errno, ")");
  }
  hdr = (unsigned char *)this->buffer;
  SET_MAGIC(hdr);
  SET_LENGTH(hdr, length);
  if (length) memcpy(this->buffer + 8, script, length);

  /* Set blocking/non-blocking mode depending on timeout. */
  flags = fcntl(this->fd, F_GETFL);
  if (ms >= 0) {
    timeleft.tv_sec = (ms / 1000);
    timeleft.tv_usec = (ms % 1000)*1000;
    if (flags == -1 || (!(flags & (O_NONBLOCK)) && 
			fcntl(this->fd, F_SETFL, (flags & (O_NONBLOCK))))) {
      return system_error(this, "set non-blocking mode failed (", errno, ")");
    }
  } else {
    if (flags == -1 || ((flags & (O_NONBLOCK)) && 
			fcntl(this->fd, F_SETFL, (flags & ~(O_NONBLOCK))))) {
      return system_error(this, "set blocking mode failed (", errno, ")");
    }
  }

  /* Set interrupted status in case a non-ignored SIGPIPE or another signal
     interrupt the complete send/receive operation. */
  this->status = TCLSRV_INTERRUPTED;

  /* Send request. */
  address = this->buffer;
  for (;;) {
    /* Wait for connection to be writable if timeout given. */
    if (ms >= 0) {
      fd_set set;
      int status;
      FD_ZERO(&set);
      FD_SET(this->fd, &set);
      status = select(this->fd + 1, NULL, &set, NULL, &timeleft);
      if (status != 1) {
	char msg[80];
	tclsrv_disconnect(this);
	if (status == 0) {
	  sprintf(msg, "timeout of %ld milliseconds expired", ms);
	  return my_error(this, msg);
	}
	strcpy(msg, "error while waiting for connection to be writable");
	if (status == -1) {
	  strcat(msg, " (");
	  return system_error(this, msg, errno, ")");
	}
	return my_error(this, msg);
      }
      DEBUG_MSG(("time left = %.3f / %.3f seconds\n",
		timeleft.tv_sec + 1e-6*timeleft.tv_usec, 1e-3*ms));
    }

    /* Send message and check number of bytes written. */
    nbytes = write(this->fd, address, pending);
#   if DEBUG
    if (nbytes >= 0) DEBUG_MSG(("sent %u pending bytes / %u\n",
				nbytes, 8+length));
#   endif
    if (nbytes == pending) break;
    if (nbytes > 0 && nbytes < pending) {
      /* Partial write: update number of pending bytes and buffer offset. */
      pending -= nbytes;
      address += nbytes;
    } else if (nbytes == -1) {
      /* An error has occured. */
      if (errno == EAGAIN && ms >= 0) continue;
      if (errno == EPIPE) return connection_closed_by_server(this);
      (void)tclsrv_disconnect(this);
      return system_error(this, "write error (", errno, ")");
    } else {
      /* Unexpected return value! */
      tclsrv_disconnect(this);
      return my_error(this, "unexpected write error");
    }
  }


  /* Receive result.  The read must be performed in two steps:
       1/ read message header + 1 byte (Tcl result code) and compute
          length of result string;
       2/ read result string if any. */
  if (read_data(this, 9, ms, &timeleft) != TCLSRV_OK) return this->status;
  hdr = (unsigned char *)this->buffer; /* internal buffer may have changed */
  DEBUG_MSG(("received 9 bytes header: 0x%02x%02x%02x%02x %lu %u\n",
	     hdr[0], hdr[1], hdr[2], hdr[3],
	     (unsigned long)GET_LENGTH(hdr),
	     (unsigned int)hdr[8]));
  if (BAD_MAGIC(hdr)) {
    errmsg = "bad magic number";
    goto disconnect;
  }
  switch (hdr[8]) {
  case TCL_OK:       status = TCLSRV_OK;       break;
  case TCL_ERROR:    status = TCLSRV_ERROR;    break;
  case TCL_RETURN:   status = TCLSRV_RETURN;   break;
  case TCL_BREAK:    status = TCLSRV_BREAK;    break;
  case TCL_CONTINUE: status = TCLSRV_CONTINUE; break;
  default:
    errmsg = "unexpected result code";
    goto disconnect;
  }
  length = GET_LENGTH(hdr);
  if (length < 1) {
    errmsg = "bad message length";
    goto disconnect;
  }
  --length; /* We already have read the first byte of the message contents. */
  if (length > 0 && read_data(this, length, ms, &timeleft) != TCLSRV_OK) {
    return this->status;
  }
  this->buffer[length] = 0; /* read_data has allocated at least 1 more byte */
  return (this->status = status);

  /* Disconnect on error. */
 disconnect:
  tclsrv_disconnect(this);
  return my_error(this, errmsg);
}

/* Read data from THIS.
   Side effect: internal buffer get re-allocated as needed. */
static int read_data(this, pending, ms, timeleft)
     tclsrv_client_t  *this;
     size_t          pending;
     long            ms;
     struct timeval *timeleft;
{
  ssize_t nbytes;
  char *address;
  int inp = this->fd;

  /* Make sure the internal buffer is large enough to read the data (plus
     one extra byte for the final null, in case we are reading the result
     string). */
  if (pending >= this->bufsize &&
      alloc_buffer(this, pending + 1) != TCLSRV_OK) {
    return system_error(this, "insufficient memory to receive message ",
			errno, ")");
  }

  /* Read until all expected bytes are received or until error or timeout. */
  address = this->buffer;
  for (;;) {
    /* Wait for connection to be readable if timeout given. */
    if (ms >= 0) {
      fd_set set;
      int status;
      FD_ZERO(&set);
      FD_SET(inp, &set);
      status = select(inp + 1, &set, NULL, NULL, timeleft);
      if (status != 1) {
	char msg[80];
	tclsrv_disconnect(this);
	if (status == 0) {
	  sprintf(msg, "timeout of %ld milliseconds expired", ms);
	  return my_error(this, msg);
	}
	strcpy(msg, "error while waiting for connection to be readable");
	if (status == -1) {
	  strcat(msg, " (");
	  return system_error(this, msg, errno, ")");
	}
	return my_error(this, msg);
      }
    }

    /* Attemp to read data. */
    nbytes = read(inp, address, pending);
    if (nbytes == pending) return TCLSRV_OK;
    if (nbytes > 0 && nbytes < pending) {
      /* Partial read: update number of pending bytes and buffer offset. */
      pending -= nbytes;
      address += nbytes;
    } else if (nbytes == 0) {
      /* The standard is that read() returns zero byte on end-of-file, which
	 only occurs when other side of socket was closed. */
      return connection_closed_by_server(this);
    } else if (nbytes == -1) {
      /* An error has occured. */
      if (errno == EAGAIN && ms >= 0) continue;
      tclsrv_disconnect(this);
      return system_error(this, "read error (", errno, ")");
    } else {
      /* Unexpected return value! */
      tclsrv_disconnect(this);
      return my_error(this, "unexpected read result");
    }
  }
}

static int alloc_buffer(this, minsize)
     tclsrv_client_t *this;
     size_t         minsize;
{
  char *new, *old = this->buffer;
  size_t size = MIN_BUFSIZE;
  while (size < minsize) size <<= 1;
  if (! (new = malloc(size))) return TCLSRV_ERROR;
  this->buffer = new;
  this->bufsize = size;
  if (old) free(old);
  return TCLSRV_OK;
}

static int clear_error(this)
     tclsrv_client_t *this;
{
  if (this->buffer) this->buffer[0] = 0;
  return (this->status = TCLSRV_OK);
}

static int connection_closed_by_server(this)
  tclsrv_client_t *this;
{
  tclsrv_disconnect(this);
  return my_error(this, "connection closed by server");
}

static int interrupt_error(this)
  tclsrv_client_t *this;
{
  tclsrv_disconnect(this);
  return my_error(this, "interrupted while sending/receiving data");
}

static int system_error(this, before, code, after)
   tclsrv_client_t *this;
   TCLSRV_CONST char *before;
   int code;
   TCLSRV_CONST char *after;
{
  int i;
  TCLSRV_CONST char *str[3];
  size_t length, len[3];

  str[0] = before;
  str[1] = strerror(code);
  str[2] = after;
  length = 0;
  for (i=0 ; i<3 ; ++i) {
    len[i] = str[i] ? strlen(str[i]) : 0;
    length += len[i];
  }
  if (this->bufsize <= length) alloc_buffer(this, length + 1);
  if (this->bufsize) {
    size_t offset=0;
    for (i=0 ; i<3 ; ++i) {
      length = len[i];
      if (offset + length >= this->bufsize) {
	length = this->bufsize - 1 - offset;
      }
      if (length) {
	char *msg = this->buffer + offset;
	memcpy(msg, str[i], length);
	if (i == 1) {
	  /* Little hack to un-capitalize system error message. */
	  if (msg[0] >= 'A' && msg[0] <= 'Z' &&
	      msg[1] >= 'a' && msg[1] <= 'z') {
	    msg[0] += 'a' - 'A';
	  }
	}
      }
      offset += length;
    }
    this->buffer[offset + length] = 0;
  }
  return (this->status = TCLSRV_ERROR);
}

static int my_error(this, errmsg)
   tclsrv_client_t *this;
   TCLSRV_CONST char *errmsg;
{
  size_t length = strlen(errmsg);
  if (this->bufsize <= length) alloc_buffer(this, length + 1);
  if (this->bufsize) {
    if (length >= this->bufsize) length = this->bufsize - 1;
    if (length) memcpy(this->buffer, errmsg, length);
    this->buffer[length] = 0;
  }
  return (this->status = TCLSRV_ERROR);
}

/*---------------------------------------------------------------------------*/
