/*
 * yeti_math.c --
 *
 *	Additional math builtin functions and generalized matrix-vector
 *	multiplication for Yorick.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *	
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *	
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *	
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_math.c,v 1.1 2005/05/24 13:24:51 eric Exp $
 *	$Log: yeti_math.c,v $
 *	Revision 1.1  2005/05/24 13:24:51  eric
 *	Initial revision
 *
 */

#include <math.h>
#include "yeti.h"
#include "bcast.h"
#include "yio.h"

/* Some constants. */
#define PI              3.141592653589793238462643383279502884197
#define TWO_PI          6.283185307179586476925286766559005768394
#define ONE_OVER_TWO_PI 0.1591549430918953357688837633725143620345

/*
 * Utility macros: STRINGIFY takes an argument and wraps it in "" (double
 * quotation marks), JOIN joins two arguments.  Both are capable of
 * performing macro expansion of their arguments.
 */
#define VERBATIM(x) x
#if defined(__STDC__) || defined(__cplusplus) || defined(c_plusplus)
# define STRINGIFY(x)   STRINGIFY1(x)
# define STRINGIFY1(x)  #x
# define JOIN(a,b)      JOIN1(a,b)
# define JOIN1(a,b)     a##b
#else
# define STRINGIFY(x)   "x"
# define JOIN(a,b)      VERBATIM(a)/**/VERBATIM(b)
#endif


extern BuiltIn Y_sinc;

extern void *BuildResultU(Operand *op, StructDef *base);
extern void *BuildResult2(Operand *l, Operand *r);
extern void PopToI(Symbol *s);
extern void PopToL(Symbol *s);
extern void PopToD(Symbol *s);

/* ANSI standard math.h functions */
extern double sin(double);
extern double cos(double);
extern double tan(double);
extern double asin(double);
extern double acos(double);
extern double atan(double);
extern double atan2(double, double);
extern double sinh(double);
extern double cosh(double);
extern double tanh(double);
extern double exp(double);
extern double log(double);
extern double log10(double);
extern double sqrt(double);
extern double ceil(double);
extern double floor(double);

/* function either present in math library or implemented in nonc.c */
extern double hypot(double, double);

/* UnaryTemplate and Looper definitions are stolen from Yorick/std0.c */

typedef void Looper(double *dst, const double *src, long n);

static void UnaryTemplate(int nArgs, Looper *DLooper, Looper *ZLooper);

static void UnaryTemplate(int nArgs, Looper *DLooper, Looper *ZLooper)
{
  Operand op;
  int promoteID;
  if (nArgs!=1) YError("expecting exactly one argument");
  if (!sp->ops) YError("unexpected keyword");
  sp->ops->FormOperand(sp, &op);
  promoteID= op.ops->promoteID;
  if (promoteID<=T_DOUBLE) {
    if (promoteID<T_DOUBLE) op.ops->ToDouble(&op);
    DLooper(BuildResultU(&op, &doubleStruct), op.value, op.type.number);
    PopToD(sp-2);
  } else {
    if (promoteID>T_COMPLEX) YError("expecting numeric argument");
    ZLooper(BuildResultU(&op, &complexStruct), op.value, 2*op.type.number);
    PopTo(sp-2);
  }
  Drop(1);
}

/* ----- sinc(x) = sin(PI*x)/PI/x ----- */

static void sincDLoop(double *dst, const double *src, long n);
static void sincZLoop(double *dst, const double *src, long n) ;

void Y_sinc(int nArgs) { UnaryTemplate(nArgs, &sincDLoop, &sincZLoop); }

static void sincDLoop(double *dst, const double *src, long n)
{
  const double pi = PI;
  long i;
  double x;
  
  for (i=0 ; i<n ; i++) {
    x = src[i];
    if (x) {
      x *= pi;
      dst[i] = sin(x)/x;
    } else {
      dst[i] = 1.0;
    }
  }
}

static void sincZLoop(double *dst, const double *src, long n) 
{
  long i;
  double lr, li, rr, ri;
    
  for (i=0 ; i<n ; i+=2) {
    rr = src[i];
    ri = src[i+1];
    if (rr || ri) {
      rr *= PI;
      ri *= PI;
      lr = sin(rr) * cosh(ri);
      li = cos(rr) * sinh(ri);
      /* Take care of overflows (this piece of code should be faster than
	 DivideZ() in Yorick/ops2.c).  We already know that (rr,ri) != (0,0),
	 nevertheless, Yorick catch divisions by zero. */
      if (fabs(rr) > fabs(ri)) {
	ri /= rr;
	rr = 1.0/((1.0 + ri*ri)*rr);
	dst[i]   = (lr + li*ri)*rr;
	dst[i+1] = (li - lr*ri)*rr;
      } else {
	rr /= ri;
	ri = 1.0/((1.0 + rr*rr)*ri);
	dst[i]   = (lr*rr + li)*ri;
	dst[i+1] = (li*rr - lr)*ri;
      }
    } else {
      dst[i] = 1.0;
      dst[i+1] = 0.0;	/* Not needed? */
    }
  }
}

/*---------------------------------------------------------------------------*/

extern BuiltIn Y_round;

/* Note: you can use "floor(X+0.5)" if you don't have the round() function. */
extern double round(double x);
extern float  roundf(float x);

void Y_round(int nArgs)
{
  Operand op;
  int promoteID;
  long number, i;
  if (nArgs != 1) YError("round takes exactly one argument");
  if (! sp->ops) YError("unexpected keyword");
  sp->ops->FormOperand(sp, &op);
  promoteID = op.ops->promoteID;
  if (promoteID == T_DOUBLE) {
    double *x, *y;
    x = op.value;
    y = BuildResultU(&op, &doubleStruct);
    number = op.type.number;
    for (i=0 ; i<number ; ++i) y[i] = round(x[i]);
    PopToD(sp-2);
  } else if (promoteID <= T_FLOAT) {
    float *x, *y;
    if (promoteID != T_FLOAT) op.ops->ToFloat(&op);
    x = op.value;
    y = BuildResultU(&op, &floatStruct);
    number = op.type.number;
    for (i=0 ; i<number ; ++i) y[i] = roundf(x[i]);
    PopTo(sp-2);
  } else {
    YError("expecting non-complex numeric argument");
  }
  Drop(1);
}

extern BuiltIn Y_arc;

void Y_arc(int nArgs)
{
  Operand op;
  int promoteID;
  long number, i;
  if (nArgs != 1) YError("arc takes exactly one argument");
  if (! sp->ops) YError("unexpected keyword");
  sp->ops->FormOperand(sp, &op);
  promoteID = op.ops->promoteID;
  if (promoteID == T_DOUBLE) {
    double *x, *y;
    const double rad = TWO_PI;
    const double scl = ONE_OVER_TWO_PI;
    x = op.value;
    y = BuildResultU(&op, &doubleStruct);
    number = op.type.number;
    for (i=0 ; i<number ; ++i) y[i] = x[i] - rad*round(scl*x[i]);
    PopToD(sp-2);
  } else if (promoteID <= T_FLOAT) {
    float *x, *y;
    const float rad = JOIN(TWO_PI,F);
    const float scl = JOIN(ONE_OVER_TWO_PI,F);
    if (promoteID != T_FLOAT) op.ops->ToFloat(&op);
    x = op.value;
    y = BuildResultU(&op, &floatStruct);
    number = op.type.number;
    for (i=0 ; i<number ; ++i) y[i] = x[i] - rad*roundf(scl*x[i]);
    PopTo(sp-2);
  } else {
    YError("expecting non-complex numeric argument");
  }
  Drop(1);
}

/*---------------------------------------------------------------------------*/

extern BuiltIn Y__mvmult;

#define MAX_DIMS 10	/* Maximum number of dimensions allowed in
			   matrix-vector multiplication. */

void Y__mvmult(int argc)
{
  int transpose;
  long i, j, k=0, l, m=1, n=1;
  double *a, *x, *y, s;
  long na, da[MAX_DIMS], nx, dx[MAX_DIMS], ny, *dy= dx;
  long *index = NULL;
  Dimension *dims;
  Symbol *stack = sp;
  const double zero = 0.0;
    
  transpose = (argc==3 || argc==5) ? yeti_get_boolean(stack--) : 0;
  if (argc==2 || argc==3) {
    
    /* Get `X' vector. */
    x = YGet_D(stack--, 0, &dims);
    nx = yeti_get_dims(dims, dx, NULL, MAX_DIMS);
    
    /* Get `A' matrix. */
    a = YGet_D(stack--, 0, &dims);
    na = yeti_get_dims(dims, da, NULL, MAX_DIMS);
    
  } else if (argc==4 || argc==5) {
    
    long *ptr;
    
    /* Get `X' vector. */
    x = YGet_D(stack--, 0, &dims);
    nx = yeti_get_dims(dims, dx, NULL, MAX_DIMS);
    
    /* Get `dimsof(A)' vector. */
    ptr = YGet_L(stack--, 0, &dims);
    na = *ptr++; /* Rank of A. */
    if (dims==NULL || dims->next!=NULL || dims->number!=na+1)
      goto bad_dim_list;
    if (na > MAX_DIMS)
      goto too_many_dims;
    for (i=0 ; i<na ; ++i) {
      if ((da[i] = ptr[i]) <= 0)
	goto bad_dim_list;
    }
    
    /* Get index `where(A!=0)' vector. */
    index = YGet_L(stack--, 0, &dims);
    k = TotalNumber(dims);
    
    /* Get `A' matrix. */
    a = YGet_D(stack--, 0, &dims);
    if (TotalNumber(dims) != k) YError("bad sparse data");
    
  } else {
    YError("bad number of arguments for `_mvmult`");
  }

  /* Prepare result. */
  if (nx <= 0 || na <= nx)
    goto bad_dims;
  if ((ny= na-nx) > MAX_DIMS)
    goto too_many_dims;
  if (transpose) {
    /* Check that dimensions of input match FIRST dimensions of kernel and
       copy LAST dimensions of kernel into dimension list of result.  */
    for (i=0 ; i<nx ; ++i) {
      if ((j = dx[i]) != da[i]) goto bad_dims;
      n *= j;
    }
    for (i=0 ; i<ny ; ++i) {
      m *= (dy[i] = da[nx+i]);
    }
  } else {
    /* Check that dimensions of input match LAST dimensions of kernel and copy
       FIRST dimensions of kernel into dimension list of result.  */
    for (i=0 ; i<nx ; ++i) {
      if ((j = dx[i]) != da[ny+i]) goto bad_dims;
      n *= j;
    }
    for (i=0 ; i<ny ; ++i) {
      m *= (dy[i]= da[i]);
    }
  }
  y = yeti_push_new_double_array(yeti_make_dims(dy, NULL, ny))->value.d;
  
  /* Perfom matrix-vector multiplication. */
  if (index==NULL) {
    /* Full `matrix'. */
    if (transpose) {
      for (i=0 ; i<m ; ++i, a+=n) {
	for (s=zero, j=0 ; j<n ; ++j) s += a[j] * x[j];
	y[i]= s;
      }
    } else {
      for (i=0 ; i<m ; ++i) y[i]= zero;
      for (j=0 ; j<n ; ++j, a+=m) {
	if ((s = x[j]) != zero) {
	  for (i=0 ; i<m ; ++i) y[i] += a[i] * s;
	}
      }
    }
  } else {
    /* Sparse `matrix'. */
    j = m*n; /* Boundary for index. */
    for (i=0 ; i<m ; ++i) y[i] = zero;
    if (transpose) {
      for (l=0 ; l<k ; ++l) {
	/* Get and check index of matrix element (indices start at 1 in
	   Yorick). */
	if ((i = index[l] - 1) < 0 || i >= j) goto bad_index;
	y[i/n] += a[l] * x[i%n];
      }
    } else {
      for (l=0 ; l<k ; ++l) {
	/* Get and check index of matrix element (indices start at 1 in
	   Yorick). */
	if ((i = index[l] - 1) < 0 || i >= j) goto bad_index;
	y[i%m] += a[l] * x[i/m];
      }
    }
  }
  return;
 bad_dim_list:
  YError("bad dimension list vector");
 bad_index:
  YError("bad index");
 bad_dims:
  YError("incompatible dimensions");
 too_many_dims:
  YError("too many dimensions");
}


#if 0
/* The offsetof macro may be defined in <stddef.h> (it is ANSI standard).  */
/* #define offsetof(structure, member)  ((long)&(((structure*)0)->member))
   (does not work on Crays) */
#ifndef offsetof
#define offsetof(structure, member) \
  ((long)((char *)&(((structure*)0)->member) - (char *)0))
#endif

static Array *dereference(void *ptr)
{
  if (ptr) return (Array *)((char *)ptr - offsetof(Array, value));
  return (Array *)0;
}

static int mvmult_get_dimlist(void *ptr, long len[], int maxdims);
static int mvmult_get_dimlist(void *ptr, long len[], int maxdims)
{
  int i, ndims;
  if (ptr) {
    Array *a = (Array *)((char *)ptr - offsetof(Array, value));
    Dimension *dims = a->type.dims;
    switch (a->ops->typeID) {
#define GET_DIMLIST(type)						     \
      {									     \
	type *v = a->value.c;						     \
	if ((ndims = v[0]) != a->type.number - 1) goto bad_dimlist;	     \
	if (ndims > maxdims) goto too_many_dims;			     \
	for (i=0 ; i<ndims ; ++i) if ((len[i]=v[i+1]) < 1) goto bad_dimlist; \
      } break
    case T_CHAR:  GET_DIMLIST(char);
    case T_SHORT: GET_DIMLIST(short);
    case T_INT:   GET_DIMLIST(int);
    case T_LONG:  GET_DIMLIST(long);
#undef GET_DIMLIST
    default:
      YError("bad data type for dimension list");
    }
    if (! dims || dims->next) YError("dimension list must be a vector");
    return ndims;
  }
 bad_dimlist:
  YError("invalid dimension list");
 too_many_dims:
  YError("too many dimensions");
  return -1; /* avoid compiler warnings */
}

/* DOCUMENT y = _mvmult(a, x, 0/1);
       -or- y = _mvmult(aa, ai, ad, x, 0/1);
       -or- y = _mvmult(aa, ai1, ad1, ai2, ad2, x, 0/1);

     Return  result  of  (generalized)  matrix-vector  multiplication.   If
     optional last argument is  non-zero, the matrix operator is transposed
     (default is  no transpose).  In the  first form, A is  the full matrix
     array.  The second  form is designed for sparse operator  A: AA is the
     vector  of non-zero  elements of  A, AI  is the  vector of  indices of
     non-zero elements (starting at 1) and DA is the vector dimsof(A).  The
     third form is  for sparse operator A with separate  index lists: AA is
     the vector of  non-zero elements of A; AI1 and AI2  are the indices of
     the non-zero elements along the  pseudo-rows/columns of A; AD1 and AD2
     are the dimension lists along the pseudo-rows/columns of A.  Note that
     AA, AI1 and AI2 must have the same number of elements.

   SEE ALSO: mvmult. */
void Y_mvmult2(int argc)
{
  Operand op;
  void *ptr;
  long *index1=NULL, *index2=NULL;

  void **pointer;

  int transpose;
  long i, j, k=0, l, m=1, n=1;
  double *a, *x, *y, s;
  long na, da[MAX_DIMS], nx, dx[MAX_DIMS], ny, *dy = dx;
  Dimension *dims;
  Symbol *stack = sp;
  const double zero = 0.0;

  /* Get transpose flag. */
  if (argc == 3) {
    transpose = yeti_get_boolean(stack);
    --stack;
  } else if (argc == 2) {
    transpose = 0;
  } else {
    YError("bad number of arguments for mvmult2");
    transpose = 0; /*a voids compiler warnings */
  }

  /* Get "vector" array. */
  x = YGet_D(stack, 0, &dims);
  nx = yeti_get_dims(dims, dx, NULL, MAX_DIMS);
  --stack;

  /* Get matrix coefficients. */
  if (! stack->ops) YError("unexpected keyword");
  stack->ops->FormOperand(stack, &op);
  switch (op.ops->typeID) {
  case T_CHAR:
  case T_SHORT:
  case T_INT:
  case T_LONG:
  case T_FLOAT:
    op.ops->ToDouble(&op);
  case T_DOUBLE:
    na = yeti_get_dims(op.type.dims, da, NULL, MAX_DIMS);

    adims = op.type.dims;
    if (transpose) {
    /* Check that dimensions of input match FIRST dimensions of kernel and
       copy LAST dimensions of kernel into dimension list of result.  */
      const Dimension *p1, *p2;
      p1 = xdims;
      p2 = adims;
      
    } else {
    }

    break;
  case T_COMPLEX:
    goto bad_complex;
  case T_POINTER:
    pointer = op.value;
    if (! op.type.dims || op.type.dims->next ||
	((nptr = op.type.dims->number) != 3 && nptr != 5)) {
      YError("sparse matrix must be a vector of 3 or 5 pointers");
      return;
    }
    if (nptr == 3) {
      /* mono-index sparse matrix style:
	 [&DATA, &INDEX, &DIMS] */
      a = get_ptr_d(ptr[0], &na);
      index1 = get_ptr_l(ptr[1], &n1);
      kk = get_ptr_dims(ptr[2], da, MAX_DIMS);
    } else {
      /* 2-index sparse matrix style:
	 [&DATA, &INDEX1, &DIMS1, &INDEX2, &DIMS2] */
      a = get_ptr_d(ptr[0], &na);
      index1 = get_ptr_l(ptr[1], &n1);
      kk = get_ptr_dims(ptr[2], da, MAX_DIMS);
      index2 = get_ptr_l(ptr[3], &n2);
      kk += get_ptr_dims(ptr[4], da+kk, MAX_DIMS-kk);
     }
      
    break;
  default:
    YError("bad data type for matrix");
    return;
  }
  YETI_SOLVE_REF(stack);

  ...;


  if (argc==2 || argc==3) {
    
    /* Get `X' vector. */
    x = YGet_D(stack--, 0, &dims);
    nx = yeti_get_dims(dims, dx, NULL, MAX_DIMS);
    
    /* Get `A' matrix. */
    a = YGet_D(stack--, 0, &dims);
    na = yeti_get_dims(dims, da, NULL, MAX_DIMS);
    
  } else if (argc==4 || argc==5) {
    
    long *ptr;
    
    /* Get `X' vector. */
    x = YGet_D(stack--, 0, &dims);
    nx = yeti_get_dims(dims, dx, NULL, MAX_DIMS);
    
    /* Get `dimsof(A)' vector. */
    ptr = YGet_L(stack--, 0, &dims);
    na = *ptr++; /* Rank of A. */
    if (dims==NULL || dims->next!=NULL || dims->number!=na+1)
      goto bad_dim_list;
    if (na > MAX_DIMS)
      goto too_many_dims;
    for (i=0 ; i<na ; ++i) {
      if ((da[i] = ptr[i]) <= 0)
	goto bad_dim_list;
    }
    
    /* Get index `where(A!=0)' vector. */
    index = YGet_L(stack--, 0, &dims);
    k = TotalNumber(dims);
    
    /* Get `A' matrix. */
    a = YGet_D(stack--, 0, &dims);
    if (TotalNumber(dims) != k) YError("bad sparse data");
    
  } else {
    YError("bad number of arguments for `_mvmult`");
  }

  /* Prepare result. */
  if (nx <= 0 || na <= nx) goto bad_dims;
  if ((ny = na - nx) > MAX_DIMS) goto too_many_dims;
  if (transpose) {
    /* Check that dimensions of input match FIRST dimensions of kernel and
       copy LAST dimensions of kernel into dimension list of result.  */
    for (i=0 ; i<nx ; ++i) {
      if ((j = dx[i]) != da[i]) goto bad_dims;
      n *= j;
    }
    for (i=0 ; i<ny ; ++i) {
      m *= (dy[i] = da[nx+i]);
    }
  } else {
    /* Check that dimensions of input match LAST dimensions of kernel and copy
       FIRST dimensions of kernel into dimension list of result.  */
    for (i=0 ; i<nx ; ++i) {
      if ((j = dx[i]) != da[ny+i]) goto bad_dims;
      n *= j;
    }
    for (i=0 ; i<ny ; ++i) {
      m *= (dy[i]= da[i]);
    }
  }
  y = yeti_push_new_double_array(yeti_make_dims(dy, NULL, ny))->value.d;
  
  /* Perfom matrix-vector multiplication. */
  if (! index1) {
    /* Non-sparse matrix. */
    if (transpose) {
      for (i=0 ; i<m ; ++i, a+=n) {
	for (s=zero, j=0 ; j<n ; ++j) s += a[j]*x[j];
	y[i] = s;
      }
    } else {
      for (i=0 ; i<m ; ++i) y[i] = zero;
      for (j=0 ; j<n ; ++j, a+=m) {
	if ((s = x[j]) != zero) {
	  for (i=0 ; i<m ; ++i) y[i] += a[i] * s;
	}
      }
    }
  } else if (! index2) {
    /* Sparse matrix with single index list. */
    j = m*n; /* Boundary for index. */
    for (i=0 ; i<m ; ++i) y[i] = zero;
#warning "filling with 0 not needed?"
#warning "use memset to fill with 0"
#warning "use unsigned indices for faster test on bounds"
    if (transpose) {
      for (l=0 ; l<k ; ++l) {
	/* Get and check index of matrix element (indices start at 1 in
	   Yorick). */
	if ((i = index1[l] - 1) < 0 || i >= j) goto bad_index;
	y[i/n] += a[l] * x[i%n];
      }
    } else {
      for (l=0 ; l<k ; ++l) {
	/* Get and check index of matrix element (indices start at 1 in
	   Yorick). */
	if ((i = index1[l] - 1) < 0 || i >= j) goto bad_index;
	y[i%m] += a[l] * x[i/m];
      }
    }
  } else {
    /* Sparse matrix with 2 index lists. */
    unsigned int i1, i2;
    --y; /* Yorick's indices start at 1 */
    --x; /* Yorick's indices start at 1 */
    if (transpose) {
      for (l=0 ; l<k ; ++l) {
	if ((i1 = index1[l]) > max1 ||
	    (i2 = index2[l]) > max2) goto bad_index;
	y[i2] += a[l]*x[i1];
      }
    } else {
      for (l=0 ; l<k ; ++l) {
	if ((i1 = index1[l]) > max1 ||
	    (i2 = index2[l]) > max2) goto bad_index;
	y[i1] += a[l]*x[i2];
      }
    }
  }
  return;
 bad_dim_list:
  YError("bad dimension list vector");
 bad_index:
  YError("bad index");
 bad_dims:
  YError("incompatible dimensions");
 too_many_dims:
  YError("too many dimensions");
}
#endif
/*---------------------------------------------------------------------------*/
